/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2010   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  FILE: haralicktest.cc
 *
 */


#include "image3d.h"
#include "descriptors/haralick.h"

using namespace std;
using namespace i3d;


template <class T> void AnalyzeTexture(const Image3d<T> &img, T ql = 0)
{
#ifdef I3D_DEBUG
	 cout << "Making cooccurrence matrix" << endl;
#endif

	 HaralickFeatures<T> *hf;

	 if (img.GetSizeZ() == 1)
		  hf = new HaralickFeatures<T>(img, Haralick2D_neighbourhood, 8, ql);
	 else
		  hf = new HaralickFeatures<T>(img, Tesar3D_neighbourhood, 6, ql);

#ifdef I3D_DEBUG
	 cout << "\n Show generated matrices" << endl;
	 hf->ShowMatrices();

	 cout << "\n Show used vectors" << endl;
	 hf->ShowSeparationVectors();

	 cout << "Evaluating the individual mean Haralick features: \n" << endl;

	 
	 cout << "Entropy: " << 
				hf->GetMeanFeature(Entropy) << endl;
	 cout << "Texture Contrast: " << 
				hf->GetMeanFeature(TextureContrast) << endl;
	 cout << "Texture Correlation: " << 
				hf->GetMeanFeature(TextureCorrelation) << endl;
	 cout << "Texture Homogeneity: " << 
				hf->GetMeanFeature(TextureHomogeneity) << endl;
	 cout << "Inverse Difference Moment: " << 
				hf->GetMeanFeature(InverseDifferenceMoment) << endl;
	 cout << "Maximum Probability: " << 
				hf->GetMeanFeature(MaximumProbability) << endl;
	 cout << "Uniformity Of Energy: " << 
				hf->GetMeanFeature(UniformityOfEnergy) << endl;
	 cout << "Contrast: " << 
				hf->GetMeanFeature(Contrast) << endl;
	 cout << "Sum Average: " << 
				hf->GetMeanFeature(SumAverage) << endl;
	 cout << "Sum Entropy: " << 
				hf->GetMeanFeature(SumEntropy) << endl;
	 cout << "Difference Entropy: " << 
				hf->GetMeanFeature(DifferenceEntropy) << endl;
	 cout << "Variance: " << 
				hf->GetMeanFeature(Variance) << endl;
	 cout << "Information Measure Of Correlation A: " <<
				hf->GetMeanFeature(InfMeasureOfCorrelationA) << endl;
	 cout << "Information Measure Of Correlation B: " <<
				hf->GetMeanFeature(InfMeasureOfCorrelationB) << endl;
	 cout << "Maximal Correlation Coefficient: " <<
				hf->GetMeanFeature(MaximalCorrelationCoefficient) << endl;
#else

	 cout << 
				hf->GetMeanFeature(Entropy) << "	";
	 cout << 
				hf->GetMeanFeature(TextureContrast) << "	";
	 cout << 
				hf->GetMeanFeature(TextureCorrelation) << "	";
	 cout << 
				hf->GetMeanFeature(TextureHomogeneity) << "	";
	 cout << 
				hf->GetMeanFeature(InverseDifferenceMoment) << "	";
	 cout << 
				hf->GetMeanFeature(MaximumProbability) << "	";
	 cout << 
				hf->GetMeanFeature(UniformityOfEnergy) << "	";
	 cout << 
				hf->GetMeanFeature(Contrast) << "	";
	 cout << 
				hf->GetMeanFeature(SumAverage) << "	";
	 cout << 
				hf->GetMeanFeature(SumEntropy) << "	";
	 cout << 
				hf->GetMeanFeature(DifferenceEntropy) << "	";
	 cout << 
				hf->GetMeanFeature(Variance) << "	";
	 cout << 
				hf->GetMeanFeature(InfMeasureOfCorrelationA) << "	";
	 cout << 
				hf->GetMeanFeature(InfMeasureOfCorrelationB) << "	";
	 cout << 
				hf->GetMeanFeature(MaximalCorrelationCoefficient) << endl;
#endif

	 delete hf;
}

int main(int argc, char* argv[])
{

  if (argc < 2 ) {
   cerr << argv[0] << " ... evaluation of Haralick texture features \n\n"
	"Copyright (C) 2010 Centre for Biomedical Image Analysis \n"
	"This is free software; see the source for copying conditions. "
	"There is NO\nwarranty; not even for MERCHANTABILITY or FITNESS FOR "
	"A PARTICULAR PURPOSE.\n\n";
    cerr << "Usage: "<< 
				argv[0] << " <input image> [<quantization constant>]" << endl;
	 cerr << " - input image: the analyzed texture" << endl;
	 cerr << " - quantization constant: into how many intensity level \n " <<
				"   should be the analyzed image quantized" << endl << endl;
   exit(1);
  }

  try  
  {
		ImgVoxelType vt = ReadImageType(argv[1]);

		switch (vt)
		{
			 case Gray8Voxel: 
						{
							 Image3d<GRAY8> img(argv[1]);

							 if (argc == 3)
								 AnalyzeTexture(img, (GRAY8) atoi(argv[2]));
							 else
								 AnalyzeTexture(img);

							 break;
						}
			 case Gray16Voxel: 
						{
							 Image3d<GRAY16> img(argv[1]);

							 if (argc == 3)
								  AnalyzeTexture(img, (GRAY16) atoi(argv[2]));
							 else
								  AnalyzeTexture(img);
							 break;
						}
			 default: cerr << "wrong image type" << endl;
		}
  }
  catch (IOException& e) {
    cerr << e <<endl;
    exit(1);
  }
  catch (InternalException& e) {
    cerr << e <<endl;
    exit(1);
  }
  catch (bad_alloc&) {
    cerr << "Not enough memory." << endl;
    exit(1);
  }
  catch (...) {
    cerr << "haralicktest: Unknown exception." << endl;
    exit(1);
  }

  return 0;
}
