/*
* i3dlib - image manipulation library
*
* Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Library General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Library General Public License for more details.
*
* You should have received a copy of the GNU Library General Public
* License along with this library; if not, write to the Free
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


/*
* FILE: medianfilter.cc
*
* Image filtering using median filters
*
* Radoslav Stefanik <256711@mail.muni.cz> 2010
*/

#include <iostream>
#include "image3d.h"
#include "rankFilterByHistogram.h"
#include "neighbours.h"
#include "rank_filters.h"
#include <cstring>
using namespace std;
using namespace i3d;

int main(int argc, char* argv[]) {
    if (argc < 6) {
        cerr << argv[0] << " ... median filter\n"
        "Copyright (C) 2006 Centre for Biomedical Image Analysis \n"
        "This is free software; see the source for copying conditions. "
        "There is NO\nwarranty; not even for MERCHANTABILITY or FITNESS FOR "
        "A PARTICULAR PURPOSE.\n\n";
        cerr << "Usage: "<< argv[0] << " input output function_type SE_type SE_size" << endl;
        cerr << "  input  - input image to load can be specified by" << endl;
        cerr << "           b) file name in i3d format: <file_name>.i3d" << endl;
        cerr << "  output - name of output file " << endl;
        cerr << "  function_type - used function for median " << endl;
        cerr << "           a) classic "<< endl;
        cerr << "           b) histogram  - implementation with moving istogram" << endl;
        cerr << "  SE_type - shape of used structural element" << endl;
        cerr << "           a) line" << endl;
        cerr << "           b) rectangle" << endl;
        cerr << "           c) cube" << endl;
        cerr << "  SE_size - nuber of elements in SE, for rectangle and cube is the length of side, should be odd"  << endl;
        exit(1);
    }

    try  {
        cerr << "Reading image Type" << endl;
        ImgVoxelType vt = ReadImageType(argv[1]);
        cerr << "Voxel=" << VoxelTypeToString(vt) << endl;
        int usedHistogram = strcmp( "classic", argv[3]);
        int sizeOfSE = atoi(argv[5]);
        if(sizeOfSE%2==0) sizeOfSE++;
        int Numofelements;
        int* pole;
        //we prepare input data for function if SE is line
        if( strcmp( "line", argv[4])==0 ) {
            pole = new int[sizeOfSE*3];
            Numofelements = sizeOfSE;
            for(int i=-(sizeOfSE/2), k=0; i<sizeOfSE/2 +1; i++,k+=3) {
                pole[k]=i, pole[k+1]=0, pole[k+2]=0;
            }

        //in case of rectangle
        } else if(strcmp("rectangle", argv[4])==0) {
            pole = new int[sizeOfSE*sizeOfSE*3];
            Numofelements = sizeOfSE* sizeOfSE;
            for(int i=-(sizeOfSE/2), k=0; i<sizeOfSE/2 +1; i++) {
                for(int j=-(sizeOfSE/2); j<sizeOfSE/2+1; j++,k+=3) {
                    pole[k]=i, pole[k+1]=j, pole[k+2]=0;
                }
            }

        //and cube
        } else {
            pole = new int[sizeOfSE*sizeOfSE*sizeOfSE*3];
            Numofelements = sizeOfSE* sizeOfSE*sizeOfSE;
            for(int i=-(sizeOfSE/2), k=0; i<sizeOfSE/2 +1; i++) {
                for(int j=-(sizeOfSE/2); j<sizeOfSE/2+1; j++) {
                    for(int m=-(sizeOfSE/2); m<sizeOfSE/2+1; m++,k+=3) {
                        pole[k]=i, pole[k+1]=j, pole[k+2]=m;
                    }
                }
            }
        }
        Neighbourhood neighbor(pole, Numofelements);
        switch (vt) {
        case Gray8Voxel: {
            Image3d<GRAY8> imgin(argv[1]);
            Image3d<GRAY8> imgout;
            if(usedHistogram!=0) {

                rankFilterByHistogram<GRAY8>( imgin, imgout, neighbor, neighbor.offset.size()/2+1);
            } else {
                MedianFilter<GRAY8>(imgin, imgout, neighbor);
            }
            imgout.SaveImage(argv[2], IMG_JPEG,0);
            break;
        }
        case Gray16Voxel: {
            Image3d<GRAY16> imgin(argv[1]);
            Image3d<GRAY16> imgout;
            if(usedHistogram!=0) {
                rankFilterByHistogram<GRAY16>( imgin, imgout, neighbor, neighbor.offset.size()/2+1);
            } else {
                MedianFilter<GRAY16>(imgin, imgout, neighbor);
            }
            imgout.SaveImage(argv[2]);

            break;
        }
        default:
            cerr << "Other voxel types not supported" << endl;
        }

    } catch (IOException& e) {
        cerr << e <<endl;
        exit(1);
    } catch (InternalException& e) {
        cerr << e <<endl;
        exit(1);
    } catch (bad_alloc&) {
        cerr << "Not enough memory." << endl;
        exit(1);
    } catch (...) {
        cerr << "Unknown exception." << endl;
        exit(1);
    }

    return 0;
}
