/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2012   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * FILE: imgHDF5.cc
 *
 * I/O routines for HDF5 images
 *
 * author: Maros Gonda (no contact email) 2011
 * modified by: David Svoboda <svoboda@fi.muni.cz> 2012 
 *
 */

#include <cstdio>
#include <string.h>

#include "i3d_config.h"

#ifdef WITH_HDF5

#include "imgHDF5.h"

#include "H5LTpublic.h"
#include "H5DSpublic.h"


#ifndef DIM_X
	#define DIM_X 0
#endif

#ifndef DIM_Y
	#define DIM_Y 1
#endif

#ifndef DIM_Z
	#define DIM_Z 2
#endif

#ifndef DIM_C 
	#define DIM_C 3
#endif

#ifndef DIM_T 
	#define DIM_T 4
#endif

/// Conversion macros
#define METERS2MICRONS(x) ((x)*1000000)
#define MICRONS2METERS(x) ((x)/1000000)


namespace i3d {

using namespace std;

/***********************************************************************************
		GetAttribute
    checks wheater selected dataset owns attribute, on possitive loads its content

        note: 	HDF5Lite API includes function H5LTget_atttribute
		which requires also dataset name but this private class function
		offers higher adaptation possibilities

 ***********************************************************************************/
template <class A> A* HDF5Reader::GetAttribute(hid_t dataSet, const char* attributeName, H5T_class_t checkTypeClass, A* attributeData, hsize_t *attributeSize, hsize_t *storageSize) 
{	
	bool isOK = true;
	
	/* First check if the attribute is present. 
	   We do not want to open non-existing attribute. 
	   The program won't fail but generates warngings.
	   */
	if (H5Aexists(dataSet, attributeName) == 0)
		return NULL;
	 
	// check if attribute with specified name exists and open it
	hid_t attributeID = H5Aopen(dataSet, attributeName, H5P_DEFAULT);
	if (attributeID<0)
		return NULL;
		
	hid_t attributeType = H5Aget_type(attributeID);
	H5T_class_t type_class = H5Tget_class(attributeType);
	hsize_t storeSize = H5Aget_storage_size(attributeID);
	
	if (type_class == checkTypeClass) 
	{
		// if attribute type is equal to required type continue loading data
		
		
		// attribute space, size and size
		hid_t nativeType = H5Tget_native_type(attributeType, H5T_DIR_ASCEND);
		
		hid_t attributeSpace = H5Aget_space(attributeID);
		int attributeDimensions = H5Sget_simple_extent_ndims(attributeSpace);
		hsize_t *attributeDims = new hsize_t[attributeDimensions];
		
		if (H5Sget_simple_extent_dims(attributeSpace, attributeDims, NULL) < 0)
		{
			isOK = false;
		}
		else 
		{
			hsize_t dataSize = 1;

			for (int i=0; i<attributeDimensions; i++)
			{
				dataSize*= attributeDims[i];
			}
			
			// allocate memory of specified size according to specified data type and its size
			attributeData = new A[ static_cast<size_t>(sizeof(A)==storeSize ? dataSize : dataSize*storeSize) ];
			
			
			if (H5Aread(attributeID, nativeType, attributeData) < 0) 
			{
				isOK = false;
				delete[] attributeData;
			} 
			else 
			{				
				// if parametres for attribute or dataspace sizes are set, return their values
				if (attributeSize!=NULL)
					*attributeSize = dataSize;

				if (storageSize!=NULL)
					*storageSize = storeSize;
			}
		}
		
		delete[] attributeDims;
		H5Sclose(attributeSpace);
		H5Tclose(nativeType);
	} 
	else 
	{
#ifdef I3D_DEBUG
		cout << " HDF5: Wrong type of attribute " << attributeName << endl;
#endif
		isOK = false;
	}

	// close opened objects
	H5Aclose(attributeID);
	H5Tclose(attributeType);
		
	// if there was no error, return allocated and loaded 
	// data pointer, otherwise return NULL pointer
	return (isOK ? attributeData : NULL);
}

/****************************************************************************
		IsImage
	has selected dataset Image class attribute?
 ****************************************************************************/
bool HDF5Reader::IsImage(hid_t dataset) 
{
	char classname[] = "IMAGE";
	char *attributeClass = NULL;
	bool isImage = false;
	hsize_t size=0;
	
	// load CLASS attribute and check if it is image
	attributeClass = GetAttribute(dataset, "CLASS", H5T_STRING, attributeClass, NULL, &size);
	isImage = ((attributeClass) && (strcmp(classname, attributeClass)==0));
	
	delete[] attributeClass;
	return isImage;
}


/****************************************************************************
	DimensionScaleIterate
	help function for iterating throught dataset dimension scales
   Do not change function parametres - they are required by H5DSiterate!
****************************************************************************/
herr_t HDF5Reader::DimensionScaleIterate(hid_t did, unsigned dimension, hid_t dsid, void *object)
{	
	if (object) 
	{
		HDF5Reader *reader = static_cast<HDF5Reader*>(object);
		return reader->DimensionScaleCheck(did, dimension, dsid);
	} else {
#ifdef I3D_DEBUG
		cout << "\tObject reference was not set, can't iterate through dimensions" << endl;
#endif
		return -1;
	}	
}


/****************************************************************************
		DimensionScaleCheck
	Load and check data of specified dataset of specified dimension 
****************************************************************************/
herr_t HDF5Reader::DimensionScaleCheck(hid_t mainDataSet, int dimension, hid_t dataset) 
{ 	
	float spacing;
	
	// check if referenced dataset is image being loaded
	if (mainDataSet != dataSetID)
		return -1;
		
	// load data from specified dataset
	try 
	{
		spacing = LoadObjectData(dataset); 
	} 
	catch(...) 
	{
		// if error occured during loading data,
		// stop iterating and send error value
		return -1;
	}
	
	// ---------------------------------------------------
	//     assign resolution to selected dimension
	// ---------------------------------------------------
	if (dimension==dimsPosition[DIM_X]) 
	{
		header.resolution->x = 1.0f/METERS2MICRONS(spacing);
#ifdef I3D_DEBUG
		cout << "    - setting X-axis resolution" << endl;
#endif
	} 
	else if (dimension==dimsPosition[DIM_Y]) 
	{
		header.resolution->y = 1.0f/METERS2MICRONS(spacing);
#ifdef I3D_DEBUG
		cout << "    - setting Y-axis resolution" << endl;
#endif
	} 
	else if (dimension==dimsPosition[DIM_Z]) 
	{
		header.resolution->z = 1.0f/METERS2MICRONS(spacing);
#ifdef I3D_DEBUG
		cout << "    - setting Z-axis resolution" << endl;
#endif
	}
		
	/*
#	ifdef I3D_DEBUG
	cout << "    - Dimension[" << dimension << "] checked" << endl;
#	endif
	*/

	return 0;
}


/****************************************************************************
		LoadObjectData
  Loads numeric data from specified dataset and returns float value of first
 field - used for loading image offset, resolution and other microscope data
 ****************************************************************************/
float HDF5Reader::LoadObjectData(hid_t dataset) 
{
	float buffer; 
	int dataSize=1;
	
	/*
#	ifdef I3D_DEBUG
		cout << "    - checking data type of dataset" << endl;
#	endif
	*/
	
	hid_t dataType = H5Dget_type(dataset);
	H5T_class_t dataClassType = H5Tget_class(dataType);
	size_t dataTypeSize  = H5Tget_size(dataType);	
	
	hid_t datafileSpace = H5Dget_space(dataset);
	int dataDimensions = H5Sget_simple_extent_ndims(datafileSpace);
	
	/*
#	ifdef I3D_DEBUG
		cout << "    - checking correct size of dimensions" << endl;
#	endif
	*/
	
	// find out size of dimensions and prepare correct space for loading data
	hsize_t *dataDims = new hsize_t[dataDimensions];
	if (H5Sget_simple_extent_dims(datafileSpace, dataDims, NULL) < 0)
	{
		delete[] dataDims;
		throw IOException("Unable to recieve dataset dimension sizes");
	} 
	else
	{
		for (int i=0; i<dataDimensions; i++) dataSize*= static_cast<int>(dataDims[i]);
	}
	
	delete[] dataDims;
	
	// ----------------------------------------------------------
	//   check dataset type and type size and load numeric data 
	// ----------------------------------------------------------
	if (dataClassType == H5T_FLOAT) 
	{
		if (dataTypeSize==4 /*sizeof(float)*/) 
		{
			float *data = new float[dataSize];
			H5Dread(dataset, dataType, H5S_ALL, H5S_ALL, H5P_DEFAULT, data);
			buffer = static_cast<float>(data[0]);
			delete[] data;
		} 
		else if (dataTypeSize==8 /*sizeof(double)*/) 
		{
			double *data = new double[dataSize];
			H5Dread(dataset, dataType, H5S_ALL, H5S_ALL, H5P_DEFAULT, data);
			buffer = static_cast<float>(data[0]);
			delete[] data;
		} 
		else 
		{
			throw IOException("unsupported FLOAT type of dimension");
			return -1;
		}
	} 
	else if (dataClassType == H5T_INTEGER) 
	{
		if (dataTypeSize==1 ) 
		{
			char *data = new char[dataSize];
			H5Dread(dataset, dataType, H5S_ALL, H5S_ALL, H5P_DEFAULT, data);
			buffer = static_cast<float>(data[0]);
			delete[] data;
		} 
		else if (dataTypeSize==2 /*sizeof(short int)*/) 
		{
			short *data = new short[dataSize];
			H5Dread(dataset, dataType, H5S_ALL, H5S_ALL, H5P_DEFAULT, data);
			buffer = static_cast<float>(data[0]);
			delete[] data;
		} 
		else if (dataTypeSize==4 /*sizeof(int)*/) 
		{
			int *data = new int[dataSize];
			H5Dread(dataset, dataType, H5S_ALL, H5S_ALL, H5P_DEFAULT, data);
			buffer = static_cast<float>(data[0]);
			delete[] data;
		} 
		else if (dataTypeSize==8 /*sizeof(long int)*/) 
		{
			long *data = new long[dataSize];
			H5Dread(dataset, dataType, H5S_ALL, H5S_ALL, H5P_DEFAULT, data);
			buffer = static_cast<float>(data[0]);
			delete[] data;
		} 
		else
		{
			throw IOException("unsupported INTEGER type of dimension");
		}
		
	} 
	else 
	{
		throw IOException("unsupported type of dimension");
	}	
	
	H5Tclose(dataType);
	H5Sclose(datafileSpace);
	
	return buffer;
}


/****************************************************************************
	ObjCheck - static function passed for iterate function
 IMPORTANT: don't change the function attribute list -required by H5Literate()
*****************************************************************************/
herr_t HDF5Reader::ObjCheck(hid_t loc_id, const char *name, const H5L_info_t *info, void *object) 
{
	info = info;
	if (object) 
	{
		HDF5Reader *reader = (HDF5Reader *) object;
		return reader->ObjectCheck(loc_id, name);
	} 
	else 
		return -1;
}


/****************************************************************************
		ObjectCheck 
  used for iterating through groups in file and searching for image dataset 
 ****************************************************************************/
herr_t HDF5Reader::ObjectCheck(hid_t loc_id, const char *name) 
{	
	H5O_info_t infobuf;
	hid_t object;
		
	H5Oget_info_by_name(loc_id, name, &infobuf, H5P_DEFAULT);
	
	switch (infobuf.type) 
	{
		case H5O_TYPE_GROUP: 
			{				
#ifdef I3D_DEBUG
				//	cout << "    - Group: " << name << " (" << infobuf.addr << ")" << endl;
#endif			
				object = H5Gopen2(loc_id, name, H5P_DEFAULT);
				if ( (checkedGroups.find(infobuf.addr) == checkedGroups.end()) && (searchGroups.find(infobuf.addr) == searchGroups.end()) ) 
					searchGroups[infobuf.addr]=object;
				break;
			}
		case H5O_TYPE_DATASET: 
			{				
#ifdef I3D_DEBUG
				//	cout << "    - Dataset: " << name << " (" << infobuf.addr << ")" << endl;
#endif								
				object = H5Dopen2(loc_id, name, H5P_DEFAULT);

				if (dataSetID<0) 
				{					
					if (IsImage(object)) 
					{
						dataSetID = object;						
						return 0;
					}
				} 
				else 
				{
					if (name==string("ExcitationWavelength")	|| 
					    name==string("ExcitationPhotonCount")	|| 
					    name==string("EmissionWavelength") 		||
					    name==string("RefractiveIndexLensImmersionMedium")		||
					    name==string("RefractiveIndexSpecimenEmbeddingMedium")	||
					    name==string("NumericalAperture")		||
					    name==string("BackprojectedIlluminationPinholeRadius")	||
					    name==string("BackprojectedIlluminationPinholeSpacing") /*	||
						    name==string("BackprojectedPinholeRadius")	||
						    name==string("ExcitationBeamOverfillFactor")||
						    name==string("Magnification")		||
						    name==string("ObjectiveQuality") */
						   ) 
					{
						try 
						{
							//float value = ;
							std::ostringstream val;
							if (val << LoadObjectData(object))
								header.description[string(name)] = val.str();							
						} 
						catch (IOException &e) 
						{
							cerr << e << endl;
						}
					}										
				}
				H5Dclose(object);
				break;
			}
		case H5O_TYPE_NAMED_DATATYPE: 
			{				
#ifdef I3D_DEBUG
				//	cout << "    - Datatype: " << name << " (" << infobuf.addr << ")" << endl;
#endif				
				break;
			}
		default:				
#ifdef I3D_DEBUG
				//	cout << "    - Unknown type: " << name << " (" << infobuf.addr << ")" << endl;
#endif				
			break;
	}
	
	return 0;	
}

/************************************************************
                        HDF5Reader                        
 ************************************************************/
HDF5Reader::HDF5Reader(const char *fname, const VOI<PIXELS> *voi):ImageReader(fname, voi) 
{	
	// suppress error printing of libHDF5 to avoid 
	// confusion with error handling by exceptions
#ifndef I3D_DEBUG
	H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
#endif
	
	if ((fd = H5Fopen (fname, H5F_ACC_RDONLY, H5P_DEFAULT))<0)
		throw IOException("HDF5: Unable to open " + string(fname) + ".");

#ifdef I3D_DEBUG
	cout << "Iterating inside the file for Image dataset:" << endl;
#endif
				
	H5O_info_t infobuf;
		
	if (H5Oget_info (fd, &infobuf) < 0)
		throw IOException("HDF5: Unable to iterate through file");		
		
	// iteratively search in groups for image dataset until 
	// correct dataset ID is not set by ObjectCheck function
	dataSetID=-1;
	searchGroups[infobuf.addr]=fd;		
		
	while (!searchGroups.empty() && dataSetID<0) 
	{			
		hid_t group = searchGroups.begin()->second;	
			
		H5Literate (group, H5_INDEX_NAME, H5_ITER_NATIVE, NULL, ObjCheck, this);
						
		if (dataSetID<0) 
		{
			checkedGroups.insert(searchGroups.begin(), ++searchGroups.begin());
			searchGroups.erase(searchGroups.begin());
		} 
		else 
		{
			parentGroup = group;
			checkedGroups.clear();
			searchGroups.clear();
		}		
	}
		
	// fail when image dataset hasn't been found
	if (dataSetID < 0)
	{
		throw IOException("HDF5: File " + string(fname) + " does not contain image data.");
	}
			
	for (int i=0; i<4; i++) dimsPosition[i] = -1;
}


/***************************************************************************
                          LoadImageInfo                                  
 ***************************************************************************/
void HDF5Reader::LoadImageInfo() 
{		
	bool dimensionsSet = false;
	size_t      typeSize;  
	const char* DIMENSION_NAME[] = {"X", "Y", "Z", "C", "T"};
	const char* OFFSET[] = {"XOffset", "YOffset", "ZOffset"};
	const char* TRUECOLOR_IMAGE = "IMAGE_TRUECOLOR";
	const char* INTERLACE_PLANE = "INTERLACE_PLANE";
	const char* COLOR_INVERSION = "IMAGE_WHITE_IS_ZERO";
	const char* PLACE_OF_ORIGIN = "DISPLAY_ORIGIN"; 
		
	fileSpaceID = H5Dget_space(dataSetID);
	dimensions = H5Sget_simple_extent_ndims(fileSpaceID);
	if (!(dimensions>0)) throw IOException("HDF5: No dimensions found in the file.");
	
	dims = new hsize_t[dimensions];
	
	if (H5Sget_simple_extent_dims(fileSpaceID, dims, NULL) < 0)
		throw IOException("HDF5: error");
	
	
#ifdef I3D_DEBUG
		cout << " Dataset dimension sizes: ";
		for (int i=0; i<dimensions; i++) {
			if (i) cout << "x";
			cout << dims[i];
		}
		cout << endl;
#endif
	
	// begin block
	{
		// check image dimensions' labels and set position of dimensions according to them
		size_t sze = 255;
		char DIMENSION_LBLS[256];
#ifdef I3D_DEBUG
			cout << " Dimension labels:   ";
#endif
		
		for (int i=0; i<dimensions; i++) 
		{						
			if (H5DSget_label(dataSetID, i, DIMENSION_LBLS, sze)<=0)
				continue;
			
#			ifdef I3D_DEBUG
				cout << " Dim[" << i << "]: " << DIMENSION_LBLS << ", ";
#			endif
			
			if (!strcmp(DIMENSION_LBLS,DIMENSION_NAME[DIM_X])) 
			{
				dimsPosition[DIM_X] = i;
				dimensionsSet = true;
			} 
			else if ( !strcmp(DIMENSION_LBLS,DIMENSION_NAME[DIM_Y])) 
			{
				dimsPosition[DIM_Y] = i;
				dimensionsSet = true;
			} 
			else if ( !strcmp(DIMENSION_LBLS,DIMENSION_NAME[DIM_Z]) ) 
			{
				dimsPosition[DIM_Z] = i;
				dimensionsSet = true;
			} 
			else if ( !strcmp(DIMENSION_LBLS, DIMENSION_NAME[DIM_C]) ) 
			{
				if (channel>=0)
					if (channel < static_cast<int>(dims[i]) ) 
					{
						dimsPosition[DIM_C] = i;
#ifdef I3D_DEBUG
						cout << "(CHANNEL: " << channel << ") ;";
#endif
					}
			} 
			else if ( !strcmp(DIMENSION_LBLS, DIMENSION_NAME[DIM_T]) ) 
			{
#ifdef I3D_DEBUG
				cout << "HDF5: Time series are not supported, skipping to first frame; "; // leave Channel at the same place, we won't need to check it later
#endif
			} 
			else 
			{
#ifdef I3D_DEBUG
				cout << "HDF5: Unable to recognize dimension label \"" << DIMENSION_LBLS[i] << "\"; ";
#endif
			}
		}
		
#ifdef I3D_DEBUG
		cout << endl;
#endif				
	}
	// end block

	// ---------------------------------------------------------
	//    get data type and specify image voxel type using it
	// ---------------------------------------------------------
	dataType  = H5Dget_type(dataSetID);
	dataImageType = H5Tget_class(dataType);
	typeSize  = H5Tget_size(dataType);
		
	switch (dataImageType) 
	{
		case (H5T_INTEGER): 
			{				
#ifdef I3D_DEBUG
				cout << " HDF5 data image type: H5T_INTEGER" << endl;
#endif								
				char *attributeClass = NULL;
				if ( (attributeClass = GetAttribute(dataSetID, "IMAGE_SUBCLASS", H5T_STRING, attributeClass)) != NULL ) 
				{					
					// check if image class is correct RGB image
					if (strcmp(TRUECOLOR_IMAGE, attributeClass)==0 && dimensions>2) 
					{						
						// remove channel index for RGB images
						channel = -1;
						
						char *imageType = NULL;
						if ( (imageType = GetAttribute(dataSetID, "INTERLACE_MODE", H5T_STRING, imageType)) && (!strcmp(INTERLACE_PLANE ,imageType)) && (dims[0]==3)) 
						{								
							// override previous dimension labels and 
							// consecute according to image specification
							dimsPosition[DIM_C] = 0;
							dimsPosition[DIM_Y] = 1;
							dimsPosition[DIM_X] = 2;
								
							// load and set possible Z dimension
							if (dimensions>3) 
							{
								if (dimsPosition[DIM_Z]<=2) 
									dimsPosition[DIM_Z]=3;
							} 
							else 
								dimsPosition[DIM_Z]=-1;
																
							if (typeSize==1) 
							{
								header.itype = RGBVoxel;
								dataType = H5T_NATIVE_UCHAR;
							} 
							else if (typeSize==2) 
							{
								header.itype = RGB16Voxel;
								dataType = H5T_NATIVE_UINT16;
							} 
							else
							{
								throw IOException("HDF5: Wrong size of datatype for RGB images");
							}
																
							delete[] imageType;
							delete[] attributeClass;
							break;
						}
																		
						// if INTERLACE_PLANE is not set or dataset doesn't meet the 
						// image specification continue with INTERLACE_PIXEL specification
						
						// is channel dataset last one and with correct size
						if (dims[ dimensions-1 ]==3) 
						{
							cout << "INTERLACE_PIXEL" << endl;
							// override previous dimension labels and 
							// consecute according to image specification
							dimsPosition[DIM_C] = dimensions-1;
							dimsPosition[DIM_X] = dimensions-2;
							dimsPosition[DIM_Y] = dimensions-3;
																
							// load and set possible Z dimension
							if (dimensions>3) 
							{
								if (dimsPosition[DIM_Z] >= dimensions-3 || dimsPosition[DIM_Z] < 0)
									dimsPosition[DIM_Z] = dimensions-4;
							} 
							else 
								dimsPosition[DIM_Z]=-1;
																
							for (int i=0; i<4; i++)
								cout << "  Dim[" << i << "]: " << dimsPosition[i]; 
							cout << endl;
								
							if (typeSize==1) 
							{
								header.itype = RGBVoxel;
								dataType = H5T_NATIVE_UCHAR;
							} 
							else if (typeSize==2) 
							{
								header.itype = RGB16Voxel;
								dataType = H5T_NATIVE_UINT16;
							} 
							else
								throw IOException("HDF5: Wrong size of datatype for RGB images.");
															
							delete[] imageType;
							delete[] attributeClass;
							break;
						}																	
					} 
					else if (strcmp(/*classname,*/ "IMAGE_BITMAP", attributeClass)==0) 
					{
						if (typeSize==1) 
						{
							header.itype = BinaryVoxel;
							dataType = H5T_NATIVE_UCHAR;
						} 
						else
							header.itype = UnknownVoxel;
						
						break;
					} 
					else if (strcmp(/*classname,*/ "IMAGE_INDEXED", attributeClass)==0) 
					{						
						throw IOException("HDF5: Indexed images are not supported.");
						break;
					}					
				}
				
				// if attribute IMAGE_SUBCLASS was not recognized or set as 
				// grayscale image type or some error with specification occured 
				// load image as grayscale type
#ifdef I3D_DEBUG
				cout << " VOXEL Type: GRAYSCALE" << endl;
#endif
										
				if (typeSize==1) 
				{
					header.itype = Gray8Voxel;
					dataType = H5T_NATIVE_UCHAR;
				} 
				else if (typeSize==2) 
				{
					header.itype = Gray16Voxel;
					dataType = H5T_NATIVE_UINT16;
				} 
				else if (typeSize==4 /*sizeof(int)*/) 
				{
					header.itype = IntegerVoxel;
					dataType = H5T_NATIVE_INT;
				} 
				else
					throw IOException("HDF5: Wrong size of dataset type");
										
				delete[] attributeClass;
				break;
			}
		case (H5T_BITFIELD): 
			{				
#ifdef I3D_DEBUG
				cout << " HDF5 data image type: H5T_BITFIELD" << endl;
#endif
				header.itype = UnknownVoxel;
				break;
			}
		case (H5T_FLOAT): 
			{				
#ifdef I3D_DEBUG
				cout << " HDF5 data image type: H5T_FLOAT" << endl;
#endif
				
				if (typeSize==4 /*sizeof(float)*/)
					header.itype = FloatVoxel;
				else if (typeSize==8 /*sizeof(double)*/)
					header.itype = DoubleVoxel;
				else
					header.itype = UnknownVoxel;
				break;
			}
		case (H5T_COMPOUND): 
			{				
#ifdef I3D_DEBUG
				cout << " HDF5 data image type: H5T_COMPOUND" << endl;
#endif				
				int elements = H5Tget_nmembers(dataType);
				if (elements==2) 
				{					
					// ----------------------------------------------------------------------
					//    check for Complex number type compound:
					//   2 float number members with same type, size, without overlapping
					// ----------------------------------------------------------------------
					hid_t real = H5Tget_member_type(dataType, 0);
					hid_t imag = H5Tget_member_type(dataType, 1);
					
#ifdef I3D_DEBUG
					cout << "    Members: " << H5Tget_member_name(dataType, 0) << ", " << H5Tget_member_name(dataType, 1) << endl;
#endif
					
					// check if members are equal & if members don't overlap
					if ((H5Tget_class(real)==H5Tget_class(imag)) && (H5Tget_size(real)==H5Tget_size(imag)) && 							
						(!H5Tget_member_offset(dataType, 0)) && (H5Tget_member_offset(dataType, 1)==H5Tget_size(real))) 
					{							
						// check byte sizes of types  
						if (H5Tget_size(imag) == 4 /*sizeof(float)*/) 
						{
							header.itype = Complex32Voxel;
							break;
						} 
						else if (H5Tget_size(imag) == 8 /*sizeof(double)*/) 
						{
							header.itype = Complex64Voxel;
							break;
						}							
					}
					
				} 
				else if (elements==3) 
				{					
					// ----------------------------------------------------------------------
					//    check for Vector3D number type compound:
					//   3 float number members with same type, size, without overlapping
					// ----------------------------------------------------------------------
					hid_t x = H5Tget_member_type(dataType, 0);
					hid_t y = H5Tget_member_type(dataType, 1);
					hid_t z = H5Tget_member_type(dataType, 2);
					
#ifdef I3D_DEBUG
					cout << "    Members: " << H5Tget_member_name(dataType, 0) << ", " << H5Tget_member_name(dataType, 1) << ", " << H5Tget_member_name(dataType, 2) << endl;
#endif
									
					// check if members are equal & if members don't overlap
					if (((H5Tget_class(x)==H5Tget_class(y)) && (H5Tget_class(y)==H5Tget_class(z))) &&
						((H5Tget_size(x)==H5Tget_size(y)) && (H5Tget_size(y)==H5Tget_size(z))) && 
						((!H5Tget_member_offset(dataType, 0)) && (H5Tget_member_offset(dataType, 1)==H5Tget_size(x)) && 
						(H5Tget_member_offset(dataType, 2)==(H5Tget_size(x)+H5Tget_size(y))))) 
					{
						// check size of types
						if (H5Tget_size(x) == 4 /*sizeof(float)*/) 
						{
							header.itype = VectFloatVoxel;
							break;
						} 
						else if (H5Tget_size(x) == 8 /*sizeof(double)*/) 
						{
							header.itype = VectDoubleVoxel;
							break;
						}
					}
				}
								
				header.itype = UnknownVoxel;
				break;
			}
		default: 
			{
#ifdef I3D_DEBUG
				cout << " HDF5 data image type: UNKNOWN" << endl;
#endif
				header.itype = UnknownVoxel;
				break;
			}
	} // end of switch
	
	if (header.itype == UnknownVoxel)
		throw IOException("HDF5: Unknown dataset type.");
		
	// set position of dimension if they are not already set 
	if (!dimensionsSet && !(header.itype==RGBVoxel || header.itype==RGB16Voxel)) 
	{		
		int i = dimensions-1;
		dimsPosition[DIM_X] = i--;
		dimsPosition[DIM_Y] = i--;
		dimsPosition[DIM_Z] = i--;		
	}
		
	// set size of image according to dimensions sizes
	header.size.x = static_cast<size_t>((dimsPosition[DIM_X] >=0) ? dims[dimsPosition[DIM_X]] : 1); 
	header.size.y = static_cast<size_t>((dimsPosition[DIM_Y] >=0) ? dims[dimsPosition[DIM_Y]] : 1); 
	header.size.z = static_cast<size_t>((dimsPosition[DIM_Z] >=0) ? dims[dimsPosition[DIM_Z]] : 1); 
	
#ifdef I3D_DEBUG
	cout << "Image size: " << header.size.x << "x" << header.size.y << "x" << header.size.z << endl ;
#endif
				
	// -------------------------------------------------
	//  load resolution, offset and other header values
	// -------------------------------------------------
		
	// try to iterate through dataset scales to load image resolution
	header.resolution = new Vector3d<float>();
	for (int i =0; i<dimensions; i++)
		if ( H5DSiterate_scales(dataSetID, i, NULL, DimensionScaleIterate, this) < 0) 
		{
			delete header.resolution;
#ifdef I3D_DEBUG
			cout << "HDF5: Resolution of some dimension[" << i << "] has not been loaded correctly" << endl;
#endif
			break;
		}
		
	// Try to load color information about image inversion
	char *inversion = NULL;
	if ((inversion = GetAttribute(dataSetID, COLOR_INVERSION, H5T_INTEGER, inversion)) != NULL)
		if (*inversion == 1)
			header.color_inversion = true;
	delete[] inversion;
		
	// Try to load origin of image
	char *origin = NULL;
	hsize_t attributeSize;
	if ((origin = GetAttribute(dataSetID, PLACE_OF_ORIGIN, H5T_STRING, origin, NULL, &attributeSize)) != NULL)
		if (attributeSize>2) 
		{
			if (origin[0] == 'L')
				header.zero_at_the_top = false;
			if (origin[1] == 'R')
				header.zero_on_the_left = false;
		}
	delete[] origin;
		
	// Try to load offset information for X,Y(,Z) dimensions
	for (int i=0; i<std::min(dimensions,3); i++) 
	{
		hid_t offsetDS;
		float offset;
		try 
		{
			if (H5Lexists(parentGroup, OFFSET[i], H5P_DEFAULT) == 1)
			{
				if (!((offsetDS = H5Dopen2(parentGroup, OFFSET[i], H5P_DEFAULT))<0)) 
				{
					offset = LoadObjectData(offsetDS);
					header.offset[i] = METERS2MICRONS(offset);
					H5Dclose(offsetDS);
				}
			}
		} 
		catch (...) 
		{
#ifdef I3D_DEBUG
			cerr << "Unable to load offset for dimension " << DIMENSION_NAME[i] << endl;
#endif
		}
	}
	
	// ----------------------------------------
	//       Load texual image information
	// ----------------------------------------
	LoadDescription();			
	
	// ------------------------------------
	//       Load microscope info
	// ------------------------------------
	{
#ifdef I3D_DEBUG
		cout << "-----------------------------------" << endl
		     << "  Loading microscope information  " << endl;
#endif
				
		H5O_info_t infobuf;
		if (H5Oget_info (fd, &infobuf) < 0)
			throw IOException("HDF5: Can't iterate through the file and load microscope info.");
				
		searchGroups[infobuf.addr]=fd;
		while (!searchGroups.empty()) 
		{			
			hid_t group = searchGroups.begin()->second;
			
			H5Literate (group, H5_INDEX_NAME, H5_ITER_NATIVE, NULL, ObjCheck, this);
			
			checkedGroups.insert(searchGroups.begin(), ++searchGroups.begin());
			searchGroups.erase(searchGroups.begin());			
		}
		
		checkedGroups.clear();
		searchGroups.clear();		
		
#ifdef I3D_DEBUG
		cout << "-----------------------------------" << endl;
#endif
	}
}

/****************************************************************************
		LoadDescription
	Loads all information placed in Image info directory
 ****************************************************************************/
void HDF5Reader::LoadDescription() 
{
			
#ifdef I3D_DEBUG
	cout << "-----------------------------------" << endl
	     << "  Loading image description:  " << endl;
#endif
				
	hid_t group;
	herr_t status;
	const char* IMAGE_INFO = "AdditionalImageInformation";
		
	// Is the group "AdditionalImageInformation" present in the file?		
	if (H5Lexists(parentGroup, IMAGE_INFO, H5P_DEFAULT) == 0)
	{
#ifdef I3D_DEBUG
		cerr << "HDF5: Image description is not present" << endl;
#endif
		return;
	}

	// Open the group
	if ((group = H5Gopen2(parentGroup, IMAGE_INFO, H5P_DEFAULT)) < 0)
	{
#ifdef I3D_DEBUG
		cerr << "HDF5: Unable to load image description" << endl;
#endif
		return;
	}

	// Load the image description
	if ((status = H5Literate (group, H5_INDEX_NAME, H5_ITER_NATIVE, NULL, HDF5Reader::ImageDescriptionIterate, this))<0)
	{
#ifdef I3D_DEBUG
		cerr <<	"HDF5: Loading image description failed." << endl;
#endif
		return;
	}
				
#ifdef I3D_DEBUG
	map<string, string>::iterator iter;

	for (iter=header.description.begin(); iter!=header.description.end(); iter++)
	{
		cout << "  [" << iter->first << "]: " << iter->second << endl;
	}
#endif			
}

/****************************************************************************
		ImageDescriptionIterate
	static function passed for dimension iterating function
 IMPORTANT: don't change the function attribute list -required by H5DSiterate()
*****************************************************************************/
herr_t HDF5Reader::ImageDescriptionIterate(hid_t loc_id, const char *name, const H5L_info_t *info, void *object) 
{
	info = info;
	if (object) 
	{
		HDF5Reader *reader = (HDF5Reader *) object;
		return reader->ImageDescriptionCheck(loc_id, name);
	} 
	else 
		return -1;
}

/****************************************************************************
		ImageDescriptionCheck 
	used for iterating through datasets in image info group 
	and loading image description 
 ****************************************************************************/
herr_t HDF5Reader::ImageDescriptionCheck(hid_t loc_id, const char *name) 
{	
	H5O_info_t infobuf;
	hid_t object;
	
	H5Oget_info_by_name(loc_id, name, &infobuf, H5P_DEFAULT);
		
	if (infobuf.type == H5O_TYPE_DATASET) 
	{	
		if ((object = H5Dopen2(loc_id, name, H5P_DEFAULT))<0)
			return 0;
		
		hid_t datasetType = H5Dget_type(object);
		H5T_class_t typeClass = H5Tget_class(datasetType);
		
		
		if (typeClass == H5T_STRING) 
		{
			size_t size = static_cast<size_t>(H5Dget_storage_size(object));
			
			char *data = new char[size];

			if (!(H5Dread(object, datasetType, H5S_ALL, H5S_ALL, H5P_DEFAULT, data) < 0))
				header.description[string(name)] = string(data);

			delete[] data;
		}
		
		H5Tclose(datasetType);
		H5Dclose(object);
	}
	
	return 0;
}


/***************************************************************************
	~HDF5Reader 
 ***************************************************************************/
HDF5Reader::~HDF5Reader() 
{
	delete[] dims;
	
	// H5Tclose(dataImageType);
	H5Dclose(dataSetID);
	H5Sclose(fileSpaceID);
	H5Fclose(fd);
}

/***************************************************************************
 Test if the compression schemes (algorithm) used during
 the storage of the data are available in the library.
 ***************************************************************************/  
bool HDF5Reader::CheckCompressionAlgorithms()
{
	hid_t plist = H5Dget_create_plist(dataSetID);

	if (plist < 0)
		return false;
	
	bool compression_algorithms_avail = true;

	if (H5Pall_filters_avail(plist) != 1)
		compression_algorithms_avail = false;

	H5Pclose(plist);

	return compression_algorithms_avail;
}

/***************************************************************************
		LoadData 
     Load image data for all types of image excluding interlace RGB images 
 ***************************************************************************/
template <class T> void HDF5Reader::LoadData(T *data) 
{
	if (CheckCompressionAlgorithms() == false)
	{
		throw IOException("HDF5: Data stroed with unknown compression method");
	}
	
	int i = 0;
	hsize_t *chunkdims = new hsize_t[dimensions];
	hsize_t *offset = new hsize_t[dimensions];
		
	for (i=0; i<dimensions; i++)
		offset[i] = 0;
		
		
	if (crop) 
	{
		if ( crop->size > Vector3d<size_t>(0,0,0))
			
			if ( le( Vector3d<size_t>(crop->offset+crop->size) , header.size)) 
			{
				header.size = crop->size;
				for (int i = 0; i<dimensions; i++) 
				{
					if (i == dimsPosition[DIM_Y])
						offset[i] = crop->offset.y;
					else if (i == dimsPosition[DIM_X])
						offset[i] = crop->offset.x;
					else if (i == dimsPosition[DIM_Z])
						offset[i] = crop->offset.z;
				}				
			} 
			else
			{
				throw IOException("HDF5: Crop area exceeds image size.");			
			}
		else
		{
			throw IOException("HDF5: Zero crop size.");
		}
	}
		
	// specify image header size for each dimension
	//  according to its position previously set
	for (int i = 0; i<dimensions; i++) 
	{
		if (i == dimsPosition[DIM_Y])
			chunkdims[i] = header.size.y;
		else if (i == dimsPosition[DIM_X])
			chunkdims[i] = header.size.x;
		else if (i == dimsPosition[DIM_Z])
			chunkdims[i] = header.size.z;
		else if (i == dimsPosition[DIM_C]) 
		{
			if (channel<0)
				chunkdims[i] = 3;
			else 
			{
				if (channel < static_cast<int>(dims[i]))
					offset[i]=channel;
				chunkdims[i] = 1;
			}
		} 
		else
		{
			chunkdims[i] = 1;
		}		
	}
			
	for (int i=0; i<dimensions; i++)
	{
		if (offset[i]+chunkdims[i] > dims[i])
			throw IOException("HDF5: offset and crop size exceed size of dimension " + i);
	}		
		
	// create memory space for loading data
	hid_t memspace = H5Screate_simple(dimensions, chunkdims, NULL);

	// set size and position of data of loading image
	if ( H5Sselect_hyperslab(fileSpaceID, H5S_SELECT_SET, offset, NULL, chunkdims, NULL) < 0)
		throw IOException("HDF5: error setting hyperslab image data.");

	// read selected data from file
	if (H5Dread(dataSetID, dataType, memspace, fileSpaceID, H5P_DEFAULT, data) < 0)		
		throw IOException("HDF5: error reading image data.");
		
	delete[] offset;
	delete[] chunkdims;		
}

/***************************************************************************
		LoadData 
	Load data in RGB colors
 ***************************************************************************/
template <class R> void HDF5Reader::LoadRGBData(RGB_generic<R> *data) 
{
	if (CheckCompressionAlgorithms() == false)
	{
		throw IOException("HDF5: Data stroed with unknown compression method");
	}
	
	// if RGB Channel is not located as first dimension load data by traditional style
	if (dimsPosition[DIM_C] != 0)
	{
		LoadData(data);
	}
	else 
	{		
		// Loading of "interlaced" RGB images		
		int i=0;
		hsize_t *offset = new hsize_t[dimensions];
		hsize_t *chunkdims = new hsize_t[dimensions];
			
		for (i=0; i<dimensions; i++) 
		{
			offset[i]=0;
		}				
		
		// check correct cropping size and set cropped size of loaded image
		if (crop) 
		{
			if (crop->size > Vector3d<size_t>(0,0,0))						
				if ( le( Vector3d<size_t>(crop->offset+crop->size), header.size)) 
				{ // && (crop->offset.y+crop->size.x<=header.size.x) && (crop->offset.z+crop->size.z<=header.size.z))
					header.size = crop->size;
					for (int i = 0; i<dimensions; i++) 
					{
						if (i == dimsPosition[DIM_Y])
							offset[i] = crop->offset.y;
						else if (i == dimsPosition[DIM_X])
							offset[i] = crop->offset.x;
						else if (i == dimsPosition[DIM_Z])
							offset[i] = crop->offset.z;
					}						
				} 
				else
				{
					throw IOException("HDF5: Crop area exceeds image size.");
				}						
			else
			{
				throw IOException("HDF5: Zero crop size.");
			}
		}
			
		// ----------------------------------------------------------------
		// 		load the whole channel at once
		// ----------------------------------------------------------------		
		R* buffer = new R[header.size.y*header.size.x*header.size.z];
		for (int c=0; c<3; c++) 
		{			
			offset[ dimsPosition[DIM_C] ] = c;
			chunkdims[ dimsPosition[DIM_C] ] = 1;
			
			for (int i=0; i<dimensions; i++) 
			{
				if (i == dimsPosition[DIM_X])
					chunkdims[i] = header.size.x;
				else if (i == dimsPosition[DIM_Y])
					chunkdims[i] = header.size.y;
				else if (i == dimsPosition[DIM_Z])
					chunkdims[i] = header.size.z;
			}
			
			//  check the correct sizes of dimensions
			for (int i=0; i<dimensions; i++)
			{
				if (offset[i]+chunkdims[i] > dims[i])
				{
					throw IOException("HDF5: wrong size of dimension "+i);
				}
			}
				
			hid_t memspace = H5Screate_simple(dimensions, chunkdims, NULL);

			if ( H5Sselect_hyperslab(fileSpaceID, H5S_SELECT_SET, offset, NULL, chunkdims, NULL) < 0)
			{
				throw IOException("HDF5: error");
			}

			if (H5Dread(dataSetID, dataType, memspace, fileSpaceID, H5P_DEFAULT, buffer) < 0)
			{
				throw IOException("HDF5: error");
			}			
			
			// ---------------------------
			//    load data from buffer
			// ---------------------------
			size_t dataIndex, bufferIndex;
			for (size_t x=0; x<header.size.x; x++)
				for (size_t y=0; y<header.size.y; y++)
					for (size_t z=0; z<header.size.z; z++) 
					{
						
						dataIndex = z*header.size.y*header.size.x + y*header.size.x + x;
						bufferIndex = y*header.size.x*header.size.z + x*header.size.z + z;
						if (c==0)
							data[dataIndex].red = buffer[bufferIndex] ;
						else if (c==1)
							data[dataIndex].green = buffer[bufferIndex];
						else if (c==2)
							data[dataIndex].blue = buffer[bufferIndex];
					}			
		}
		delete[] offset;
		delete[] chunkdims;
		delete[] buffer;
	}
}

/*****************************************************************************
		createCompoundType
  creates compound HDF5 type according to set type, size and names of fields
  returns descriptor of created type
 *****************************************************************************/
template<class A> hid_t HDF5Writer::CreateCompoundType(const char* name[], size_t size, hid_t H5Tid) 
{	
	hid_t type;
	if ((type = H5Tcreate (H5T_COMPOUND, sizeof(A)*size))<0)
		throw IOException("HDF5: Unable to create compound data type");
	
	for (size_t i=0; i<size; i++) 
	{
		if (H5Tinsert(type, name[i], sizeof(A)*i, H5Tid)<0)
			throw IOException("HDF5: Unable to create compound data type");
	}

	return type;
}


/***************************************************************************
	SaveImageInfo
 ***************************************************************************/
void HDF5Writer::SaveImageInfo() 
{	
	if (!newFile) 
	{		
		// check if opened file has the same image type as being saved one
		if (header.itype != ReadImageType(header.name.c_str(), IMG_HDF5) )
			throw IOException("HDF5: Can't save incompatible dataTypes as one image.");
	}
	
		
	// use block to avoid allocating memory when just checking image type
	{ // begin block		
		/// names of complex number components
		const char *complex[2] = {"r", "i"};
		/// names of Vector3d<> components
		const char *vector_3d[3] = {"x", "y", "z"};		
		// names of datasets storing image offset information 
		const char* OFFSET[] = {"XOffset", "YOffset", "ZOffset"};
		// names of datasets storing image resolution
		const char* RESOLUTION[] = {"DimensionScaleX", "DimensionScaleY", "DimensionScaleZ"};
		// names of dimensions of image
		const char* DIMENSION_NAME[] = {"X", "Y", "Z", "C"};
		// labels of dimensions of image saved in file
		char** dimensionLabels;
		
		// image dataset class name
		const char* CLASS = "IMAGE";
		const char* IMAGE_VERSION = "1.2";
		
		
		// names of supported image subclasses
		const char *IMAGE_SUBCLASSES[] = {"IMAGE_GREYSCALE", "IMAGE_TRUECOLOR", "IMAGE_BITMAP"};
		// subclass of saving image
		const char *IMAGE_SUBCLASS;
		
		dimensions = 2;
		// number of real image dimensions (2 for x,y or 3 with z)
		int realDimensions = 2;
		hid_t params;
		checkMinMax = false;
		
		dimsPosition[DIM_X] = DIM_X;
		dimsPosition[DIM_Y] = DIM_Y;
		dimsPosition[DIM_Z] = DIM_Z;
		
		
		// choose datatype and image subclass (also with specifying wheater 
		// to check image minimum and maximum) according to image voxel type
		switch (header.itype) 
		{
			case (BinaryVoxel): 
				{
					IMAGE_SUBCLASS = IMAGE_SUBCLASSES[2];
					dataImageType = H5T_NATIVE_UCHAR;
					checkMinMax = true;
					break;
				}
			case (Gray8Voxel): 
				{
					IMAGE_SUBCLASS = IMAGE_SUBCLASSES[0];
					dataImageType = H5T_NATIVE_UCHAR;
					checkMinMax = true;
					break;
				}
			case (Gray16Voxel): 
				{
					IMAGE_SUBCLASS = IMAGE_SUBCLASSES[0];
					dataImageType = H5T_NATIVE_UINT16;
					checkMinMax = true;
					break;
				}
			case (IntegerVoxel): 
				{
					IMAGE_SUBCLASS = IMAGE_SUBCLASSES[0];
					dataImageType = H5T_NATIVE_INT32;
					checkMinMax = true;
					break;
				}
			case (RGBVoxel): 
				{
					IMAGE_SUBCLASS = IMAGE_SUBCLASSES[1];
					dimensions++;	// increase number of dimension for a channel dimension
					dataImageType = H5T_NATIVE_UCHAR;
					break;
				}
			case (RGB16Voxel): 
				{
					IMAGE_SUBCLASS = IMAGE_SUBCLASSES[1];
					dimensions++;	// increase number of dimension for a channel dimension
					dataImageType = H5T_NATIVE_UINT16;
					break;
				}		
			case (FloatVoxel): 
				{
					IMAGE_SUBCLASS = IMAGE_SUBCLASSES[0];
					dataImageType = H5T_NATIVE_FLOAT;
					checkMinMax = true;
					break;
				}
			case (DoubleVoxel): 
				{
					IMAGE_SUBCLASS = IMAGE_SUBCLASSES[0];
					dataImageType = H5T_NATIVE_DOUBLE;
					checkMinMax = true;
					break;
				}			
			case (Complex32Voxel): 
				{
					IMAGE_SUBCLASS = IMAGE_SUBCLASSES[0];
					dataImageType = CreateCompoundType<float>(complex, sizeof(complex)/sizeof(*complex), H5T_NATIVE_FLOAT);
					break;
				}
			case (Complex64Voxel): 
				{
					IMAGE_SUBCLASS = IMAGE_SUBCLASSES[0];
					dataImageType = CreateCompoundType<double>(complex, sizeof(complex)/sizeof(*complex), H5T_NATIVE_DOUBLE);
					break;
				}			
			case (VectFloatVoxel): 
				{
					IMAGE_SUBCLASS = IMAGE_SUBCLASSES[0];
					dataImageType = CreateCompoundType<double>(vector_3d, sizeof(vector_3d)/sizeof(*vector_3d), H5T_NATIVE_FLOAT);
					break;
				}
			case (VectDoubleVoxel): 
				{
					IMAGE_SUBCLASS = IMAGE_SUBCLASSES[0];
					dataImageType = CreateCompoundType<double>(vector_3d, sizeof(vector_3d)/sizeof(*vector_3d), H5T_NATIVE_DOUBLE);
					break;
				}			
			default: 
				{
					throw IOException("HDF5: Unknown Voxel type.");
					break;
				}
		}
		
		if (header.size.z>1) 
		{
			dimensions++;
			realDimensions++;
		}
				
		// specify sizes and labels of image dimensions
		hsize_t *dims = new hsize_t[dimensions];
		hsize_t *chunkdims = new hsize_t[dimensions];
		dimensionLabels = new char*[dimensions];
				
		// specify the positions of dimensions for 
		//  - RGB: CxYxXxZ and specify channel dimension
		//  - others: [Zx]YxX (Z dimension is optional)
		if (header.itype==RGBVoxel || header.itype==RGB16Voxel) 
		{
			dims[0] = 3;
			dimensionLabels[0] = const_cast<char*>(DIMENSION_NAME[DIM_C]);
			chunkdims[0] = 1;
			dimsPosition[DIM_Y]=1;
			dimsPosition[DIM_X]=2;
			dimsPosition[DIM_Z]=3;
		} 
		else 
		{
			int i =0;
			if (header.size.z > 1)
				dimsPosition[DIM_Z] = i++;
			dimsPosition[DIM_Y] = i++;
			dimsPosition[DIM_X] = i++;
		}
				
		if (newFile) 
		{
			// --------------------------------------------------
			//  if new file is being created specify and create 
			//     all image dataset required information
			// --------------------------------------------------				
		
#ifdef I3D_DEBUG
			cout << "  Writing image header" << endl;
#endif
						
			// --------------------------------------------------------------
			// specify sizes of dimensions, sizes of chunks and dimension 
			// labels for every available dimension 
			//
			//  - size of chunks are set to half the size of all dimensions 
			//    of the whole image to avoid big memory consumption during
			//    data saving transformations processed within libHDF5
			// --------------------------------------------------------------
			if (crop) 
			{			
				if ( !le( crop->size , header.size ))
					throw IOException("HDF5: Crop size exceeds image size.");		
			
				chunkdims[ dimsPosition[DIM_X] ] =
				dims[ dimsPosition[DIM_X] ] = crop->offset.x + crop->size.x;
				dims[ dimsPosition[DIM_Y] ] = crop->offset.y + crop->size.y;
				chunkdims[ dimsPosition[DIM_Y] ] = (header.itype == RGBVoxel || header.itype ==RGB16Voxel ? 1 : crop->offset.y + crop->size.y);
				dimensionLabels[ dimsPosition[DIM_X] ] = const_cast<char*>(DIMENSION_NAME[DIM_X]);
				dimensionLabels[ dimsPosition[DIM_Y] ] = const_cast<char*>(DIMENSION_NAME[DIM_Y]);
			
				if (crop->size.z > 1) 
				{
					dims[ dimsPosition[DIM_Z] ] = crop->offset.z + crop->size.z;
					chunkdims[ dimsPosition[DIM_Z] ] = (header.itype == RGBVoxel || header.itype ==RGB16Voxel ? crop->offset.z + crop->size.z : 1);
					dimensionLabels[ dimsPosition[DIM_Z] ] = const_cast<char*>(DIMENSION_NAME[DIM_Z]);
				}		
			} 
			else 
			{
				dims[ dimsPosition[DIM_X] ] = header.size.x;
				dims[ dimsPosition[DIM_Y] ] = header.size.y;
				chunkdims[ dimsPosition[DIM_X] ] =  header.size.x;
				chunkdims[ dimsPosition[DIM_Y] ] = (header.itype == RGBVoxel || header.itype ==RGB16Voxel ? 1 : header.size.y); 
			
				dimensionLabels[ dimsPosition[DIM_X] ] = const_cast<char*>(DIMENSION_NAME[DIM_X]);
				dimensionLabels[ dimsPosition[DIM_Y] ] = const_cast<char*>(DIMENSION_NAME[DIM_Y]);
			
				if (header.size.z > 1) 
				{
					chunkdims[ dimsPosition[DIM_Z] ] = (header.itype == RGBVoxel || header.itype ==RGB16Voxel ? header.size.z : 1);
					dims[ dimsPosition[DIM_Z] ] = header.size.z;
					dimensionLabels[ dimsPosition[DIM_Z] ] = const_cast<char*>(DIMENSION_NAME[DIM_Z]);
				}	
			}		
		
			// ---------------------------------------------------
			//                creating the main structures
			// ---------------------------------------------------		
			
			// create the data space for the dataset
			if((dataSpaceID = H5Screate_simple(dimensions, dims, NULL)) < 0)
				throw IOException("HDF5: error");
			
			// set parametres for image dataset and specify level of compression (probably no difference between levels)
			params = H5Pcreate(H5P_DATASET_CREATE);
			
			if (header.itype==BinaryVoxel)
				H5Pset_nbit(params);
			
			if (header.compression)			// levels of compression: 1-9 (other number = no compression)
				H5Pset_deflate(params, 1);	// there is just a slight difference between final file size but 1 is the fastest level
			
			// set main size of chunks
			if (H5Pset_chunk( params, dimensions, chunkdims) < 0)
				throw IOException("HDF5: error setting chunks");
			
			// create access list parametres to disable using extra cache mamory by libHDF5
			hid_t accessParams = H5Pcreate(H5P_DATASET_ACCESS);
			if (H5Pset_chunk_cache(accessParams, 0, 0, 0) <0 )
				throw IOException("HDF5: unable to set size of chunk cache size");
			
			// create the image dataset
			if((dataSetID = H5Dcreate2(fd, DATASET_NAME, dataImageType, dataSpaceID, H5P_DEFAULT, params, accessParams)) < 0)
				throw IOException("HDF5: Error with GOTO");
							
			// ------------------------------------------------------------------
			//                        Saving image header info
			// ------------------------------------------------------------------			
			
			for (int i=0; i<realDimensions; i++)
			{
				// Save offset Information
				float offset = MICRONS2METERS(header.offset[i]);
				if (SaveAdditionalInfo(parentGroup, OFFSET[i], H5T_NATIVE_FLOAT, &offset)<0)
						throw IOException("HDF5: unable to save offset info");
				
				// Save resolution Information
				hid_t dimScale;
				float res = (header.resolution != NULL) ? (*header.resolution)[i] : DefaultResolution[i];
				float spacing = MICRONS2METERS(1.0f/res);

				if ((dimScale = SaveAdditionalInfo(parentGroup, RESOLUTION[i], H5T_NATIVE_FLOAT, &spacing)) < 0)
						throw IOException("HDF5: unable to save resolution info");
				
				H5DSset_scale(dimScale, DIMENSION_NAME[i]);
				if ( H5DSattach_scale(dataSetID, dimScale, dimsPosition[i]) < 0)
						throw IOException("HDF5: unable to set dimension scale ");
			}
			
			
			
			// -------------------------------------------------------------------
			//           Save image class attributes
			// -------------------------------------------------------------------
			
			
			/* Attach the DIMENSION_LABELS attribute with dimension names */
			for (int i=0; i<dimensions; i++)
				H5DSset_label(dataSetID, i, dimensionLabels[i]);
							
			// Attach the CLASS attribute
			if ( H5LTset_attribute_string( fd, DATASET_NAME, "CLASS", CLASS ) < 0)
				throw IOException("HDF5: Failed saving image attribute - CLASS.");
			
			// Attach the IMAGE_VERSION attribute
			if ( H5LTset_attribute_string( fd, DATASET_NAME, "IMAGE_VERSION", IMAGE_VERSION ) < 0)
				throw IOException("HDF5: Failed saving image  attribute - IMAGE_VERSION.");
			
			// Attach the IMAGE_SUBCLASS attribute
			if ( H5LTset_attribute_string( fd, DATASET_NAME, "IMAGE_SUBCLASS",  IMAGE_SUBCLASS) < 0)
				throw IOException("HDF5: Failed saving image attribute - IMAGE_SUBCLASS.");
						
			// Attach the DISPLAY_ORIGIN attribute
			char ORIGIN[3] = {0,0,0};
			ORIGIN[0] = (header.zero_at_the_top  ? 'U' : 'L'); // "Upper" or "Lower" orientation
			ORIGIN[1] = (header.zero_on_the_left ? 'L' : 'R'); // "Left"  or "Right" orientation
			
			if ( H5LTset_attribute_string( fd, DATASET_NAME, "DISPLAY_ORIGIN",  ORIGIN) < 0)
				throw IOException("HDF5: Failed saving image attribute - IMAGE_SUBCLASS.");
			
			// -------------------------------------------------------------------
			//        inserts specific attributes for specific image subclass
			// -------------------------------------------------------------------
			if (header.itype == RGBVoxel || header.itype ==RGB16Voxel) 
			{					
				const char* INTERLACE_MODE = "INTERLACE_PLANE";
				const char* IMAGE_COLORMODEL = "RGB";
					
				// Attach the INTERLACE_MODE attribute. This attributes is only for true color images 
				if ( H5LTset_attribute_string( fd, DATASET_NAME, "INTERLACE_MODE", INTERLACE_MODE) < 0)
					throw IOException("HDF5: Failed saving image attribute - INTERLACE_MODE.");					
					
				if ( H5LTset_attribute_string( fd, DATASET_NAME, "IMAGE_COLORMODEL", IMAGE_COLORMODEL) < 0)
					throw IOException("HDF5: Failed saving image attribute - IMAGE_COLORMODEL.");										
			} 
			else 
			{					
				unsigned char IMAGE_WHITE_IS_ZERO = header.color_inversion;
				if ( H5LTset_attribute_uchar( fd, DATASET_NAME, "IMAGE_WHITE_IS_ZERO", &IMAGE_WHITE_IS_ZERO, 1) < 0)
					throw IOException("HDF5: Failed saving image attribute - IMAGE_WHITE_IS_ZERO.");
				
				// Maximal and minimal value attribute for greyscale
				//  images is written after saving image data										
			}
			// -------------------------------------------------------------------
			//                        end of specific attributes 
			// -------------------------------------------------------------------
		}
		
		SaveDescription();
		
		delete[] dims;
		delete[] chunkdims;
		delete[] dimensionLabels;
		
	} // end block
		
	H5Fflush(fd, H5F_SCOPE_GLOBAL);
}


/****************************************************************************
	Saves text information attached to the image
	returns: - number of saved information
		 - negative on fail
 ****************************************************************************/
int HDF5Writer::SaveDescription() 
{	
	// Group name for image information
	const char* IMAGE_INFO = "AdditionalImageInformation";
	hid_t infoGroupID;
	int returnValue = 0;
			
	// create group for storing string information about image
	// and continue saving if the group was correctly created
	if ((infoGroupID = H5Gcreate2(parentGroup, IMAGE_INFO, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT))<0) 
	{
		returnValue = -1;
#ifdef I3D_DEBUG
		cout << "Unable to create info group" << endl;
#endif
	} 
	else 
	{
		map<string, string>::iterator iter;
						
#ifdef I3D_DEBUG
		cout << "--------------------------------------" << endl
		     << "    HDF5: Writing image description:" << endl;
#endif	
		for (iter=header.description.begin(); iter!=header.description.end(); iter++) 
		{
				
#ifdef I3D_DEBUG
			cout << "  [" << iter->first << "]: " << iter->second << endl;
#endif								
			// create dataset storing additional string information
			if (SaveAdditionalInfo(infoGroupID, iter->first.c_str(), H5T_STRING, iter->second.c_str())<0)
				cerr << "Saving information \"" << iter->first << "\" failed" << endl;
			else
				returnValue++;				
		}
						
#ifdef I3D_DEBUG
		cout << "--------------------------------------" << endl;
#endif
	}
		
	return returnValue;
}

/****************************************************************************
			SaveAdditionalInfo
	help function for creating description datasets of image
	returns reference to created Dataset or negative error value
 ****************************************************************************/
hid_t HDF5Writer::SaveAdditionalInfo(const hid_t parentGroup, const char* datasetName, hid_t dataType, const void* data, int dimensions, hsize_t* dims) 
{
	
#ifdef I3D_DEBUG
	cout << "    - Creating dataset for info data: " << datasetName << endl;
#endif
		
	hid_t infoDataSpace, params, infoDataSet, filespace, memoryDataSpace;
	
	// create default dimension sizes of single dataset while none were set
	if (dims==NULL) 
	{
		dims = new hsize_t[dimensions];
		for (int i=0; i<dimensions; i++)
		{
			dims[i] = 1;
		}
	} 
		
	if (dataType==H5T_STRING) 
	{
		// if dataset type is string 				
		if ((dataType = H5Tcopy(H5T_C_S1)) < 0 )
			return -11;
				
		if (H5Tset_size(dataType, (strlen(static_cast<const char*>(data))+1) ) < 0)
			return -12;
				
		if (H5Tset_strpad(dataType, H5T_STR_NULLTERM) < 0)
			return -13;
				
		if ((infoDataSpace = H5Screate(H5S_SCALAR)) < 0)
			return -14;		
	} 
	else 
	{
		if ((infoDataSpace = H5Screate_simple(dimensions, dims, NULL)) < 0)
			return -2;
	}
		
	// create parametres for dataset creation, possible to add also compression
	params = H5Pcreate(H5P_DATASET_CREATE);
	// create dataset with specified parametres
	if ((infoDataSet = H5Dcreate2(parentGroup, datasetName, dataType, infoDataSpace, H5P_DEFAULT, params, H5P_DEFAULT)) < 0) 
	{
		H5Pclose(params);
		H5Sclose(infoDataSpace);
		return -3;
	}
	
	// create memory and file dataspaces and write data into dataset
	filespace = H5Dget_space(infoDataSet);
	memoryDataSpace = H5Screate_simple(dimensions, dims, NULL);
	if (H5Dwrite(infoDataSet, dataType, memoryDataSpace, filespace, H5P_DEFAULT, data) < 0) 
	{
		H5Sclose(memoryDataSpace);
		H5Sclose(filespace);
		H5Pclose(params);
		H5Sclose(infoDataSpace);
		H5Dclose(infoDataSet);
		return -4;
	}
	
	// close all opened objects
	H5Sclose(memoryDataSpace);
	H5Sclose(filespace);
	H5Pclose(params);
	H5Sclose(infoDataSpace);
		
	delete[] dims;
	
	// write all remaining data from memory into the file
	//H5Fflush(infoDataSet, H5F_SCOPE_LOCAL);
	return infoDataSet;
}


/*************************************************************************
	   SaveRGBData
	SaveData in RGB
 ************************************************************************/
template <class R> void HDF5Writer::SaveRGBData(const RGB_generic<R> *data) 
{ 
	// save only if there are data set
	if (data) 
	{			
#ifdef I3D_DEBUG
		cout << "  HDF5: Writing image data" << endl;
#endif
		
		// number of RGB channels
		const int channels = 3;
		// labels of RGB channels used for exporting info data
		const char rgb[3] = {'R','G','B'};
		int i;
		
		
		// disallow inserting zero sized image if cropping was set
		if (crop)
			if ((crop->size.x == 0) || (crop->size.y == 0) || (crop->size.z == 0))
				throw IOException("HDF5: Can't insert image with zero sized volume");
		
		
		// defines basic maximal sizes of image (also with eventual image shift)
		size_t maxSizes[3] = {header.size.x, header.size.y, header.size.z};
		if (crop) 
			for (i=0; i<3; i++)
				maxSizes[i] += crop->offset[i];
				
		if (!newFile) 
		{
			// ---------------------------------------------------
			//  If data being inserted to existing file, check 
			//  number of dimensions and their sizes if they fit
			// ---------------------------------------------------
					
#ifdef I3D_DEBUG
			cout << "  - Writing cropped area" << endl;
#endif
					
			int dimens = H5Sget_simple_extent_ndims(dataSpaceID);
			if (dimens>4)
				throw IOException("HDF5: Too many dimensions in original image file.");
						
			// check if we don't want to insert 3D image into 2D original image
			if (dimens<dimensions)
				throw IOException("HDF5: Can't insert image into smaller one.");
										
			// if we are inserting less dimensional image set dimensions 
			// to be equal and check correct size of dimension Z 
			if (dimens>dimensions) 
			{
				dimensions=dimens;
				if (crop->size.z != 1)
					throw IOException("HDF5: Incorrect size of VOI in Z-dimension.");
			}
												
			// find original image's dimension sizes
			hsize_t *dims = new hsize_t[dimens];
			if (H5Sget_simple_extent_dims(dataSpaceID, dims, NULL) < 0) 
			{
				delete[] dims;
				throw IOException("HDF5: Can't load dimension sizes of original image file.");
			}
						
			// check correct dimension sizes of original image and 
			// set maximal possible size of image shift  
			if ((dims[0]<3) || (dims[1] < crop->size.y+crop->offset.y) || 
				(dims[2] < crop->size.x+crop->offset.x) || 
				((dimens==4) && (dims[3] < crop->size.z+crop->offset.z) )) 
			{
				delete[] dims;
				throw IOException("HDF5: Inserting image position/size is out of range of original image file.");
				// ------------------------------------------------------------ //
				// TODO: //
				//  Je mozne tu nastavit namiesto vynimky zmenu velkosti 	//
				//  povodneho obrazka pomocou H5Dset_extent(dataSetID, dims);	//
				//  kde sa do dims nastavia velkosti jednotlivych dimenzii	//
				// ------------------------------------------------------------ //
			}
						
			// set up maximal possible sizes for shift of new image
			maxSizes[DIM_X] = static_cast<size_t>(dims[2]);
			maxSizes[DIM_Y] = static_cast<size_t>(dims[1]);
			if (dimens == 4) 
				maxSizes[DIM_Z] = static_cast<size_t>(dims[3]);
												
			delete[] dims;
		}
				
#ifdef I3D_DEBUG
		cout << " Maximal possible sizes of non-channel dimensions: [";
		for (i=0; i <(dimensions-1); i++) 
		{
			if (i) 
			{
				cout << "x";
			}
				
			cout << maxSizes[i];
		}
			
		cout << "]" << endl;
#endif
			
		hsize_t *hyperSlab = new hsize_t[dimensions];
		hsize_t *offset = new hsize_t[dimensions];
		hsize_t sizeZ = 1;
		hyperSlab[0] = 1; 
				
		if (crop) 
		{
			// ------------------------------------------------------------------
			//  check size of cropped area and compare it with maximal possible 
			//     size (header size or dimension sizes of original image)
			// ------------------------------------------------------------------
					
			// check if size of VOI is less or equal than image size, otherwise skip 
			// all settings and save image with offset [0,0,0] and original size
			if ((crop->size.x > header.size.x) || (crop->size.y > header.size.y) || (crop->size.z > header.size.z))
				throw IOException("HDF5: cropped image size exceeds size of saving image");
										
			if (dimensions==4) 
			{
				if (crop->size.z + crop->offset.z <= maxSizes[DIM_Z]) 
				{
					sizeZ = hyperSlab[3] = crop->size.z;
					offset[3] = crop->offset.z;
				} 
				else
					throw IOException("HDF5: Inserting image position/size is out of range of original image file.");
			}
					
			if ((crop->size.x + crop->offset.x <= maxSizes[DIM_X]) && (crop->size.y + crop->offset.y <= maxSizes[DIM_Y])) 
			{
				hyperSlab[1] = crop->size.y;
				offset[1] = crop->offset.y;
					
				hyperSlab[2] = crop->size.x;
				offset[2] = crop->offset.x;
			} 
			else
				throw IOException("HDF5: Inserting image position/size is out of range of original image file.");			
		} 
		else  
		{			
			// --------------------------------------------------------
			//   set hyperSlab for non-VOI images to whole image size 
			// --------------------------------------------------------
					
			hyperSlab[1] = header.size.y;
			hyperSlab[2] = header.size.x;
					
			offset[1] = 0;
			offset[2] = 0;
					
			if (dimensions==4) 
			{
				sizeZ = hyperSlab[3] = header.size.z;
				offset[3] = 0;
			}				
		}
				
		// ------------------------------------------------------
		//  set buffer for one channel data according to 
		//  already specified size of hyperslab for saving data
		// ------------------------------------------------------
		size_t imageSize = static_cast<size_t>(hyperSlab[1]*hyperSlab[2]*sizeZ);
		R* buffer = new R[imageSize];
						
#ifdef I3D_DEBUG
		cout << " Inserting image size: " << hyperSlab[1] << "x" << hyperSlab[2] << "x" << sizeZ << endl;
#endif
				
		// save every channel separatelly - best middle 
		// course of memory consumption and computing time
		for (int c=0; c<channels; c++) 
		{			
			// copy data of selected channel into buffer and select correct offset of it
			offset[0] = c;
			i=0;
			for (size_t y=0; y<hyperSlab[1]; y++) 
			{
				for (size_t x=0; x<hyperSlab[2]; x++) 
				{
					for (size_t z=0; z<sizeZ; z++) 
					{						
						size_t index = x + y*header.size.x + z*header.size.y*header.size.x;
						if (c==0)
							buffer[i] = data[index].red;
						else if(c==1)
							buffer[i] = data[index].green;
						else
							buffer[i] = data[index].blue;
						i++;
					}
				}
			}
						
			// get information about used filespace and set hyperslab in file for saving data in
			hid_t filespace = H5Dget_space(dataSetID);

			if ( H5Sselect_hyperslab(filespace, H5S_SELECT_SET, offset, NULL, hyperSlab, NULL)<0 ) 
			{
				delete[] buffer;
				throw IOException("HDF5: error setting hyperSlab for channel " + rgb[c]);
			}
			
			
#ifdef I3D_DEBUG
			cout << " - Writing channel " << rgb[c] << " data into file" << endl;
#endif
			
			// create memory space for saving data and write them all into file to avoid duplicing data in memory
			hid_t memoryDataSpace = H5Screate_simple(dimensions, hyperSlab, NULL);

			if (H5Dwrite(dataSetID, dataImageType, memoryDataSpace, filespace, H5P_DEFAULT, buffer) < 0) 
			{
				delete[] buffer;
				throw IOException("HDF5: error writing data into channel " + rgb[c]);
			}
		}
				
		delete[] hyperSlab;
		delete[] offset;
		delete[] buffer;
	}
}


/*************************************************************************
		SaveData 
	save greyscale images
************************************************************************/
template <class T> void HDF5Writer::SaveData(const T *data) 
{
	// Write the dataset only if there are data to write
	if (data) 
	{		
		T* buffer;
		T minMax[2];
		// defines basic maximal sizes of 
		hsize_t maxSizes[3];
		
#ifdef I3D_DEBUG
			cout << "  HDF5: Writing image data" << endl;
#endif
				
		if (crop) 
		{ 						
#ifdef I3D_DEBUG
			cout << "  - Writing only cropped area" << endl;
#endif
			if ((crop->size.x == 0) || (crop->size.y == 0) || (crop->size.z == 0))
				throw IOException("HDF5: Can't insert image with zero sized volume");
			
			for (int i=0; i<3; i++)
				maxSizes[i] = header.size[i]+crop->offset[i];
			
			if (!newFile) 
			{
				// ---------------------------------------------------
				//  If data being inserted to existing file, check 
				//  number of dimensions and their sizes if they fit
				// ---------------------------------------------------				
#ifdef I3D_DEBUG
				cout << "  - Checking dimensions in opened file" << endl;
#endif				
				// skip checking minimal and maximal image values										
				int dimens = H5Sget_simple_extent_ndims(dataSpaceID);
				if (dimens>3)
					throw IOException("HDF5: Too many dimensions in original image file.");
					
				// check if we don't want to insert 3D image into 2D original image
				if (dimens<dimensions) 
				{
					//if (crop->size.z>1)
					throw IOException("HDF5: Can't insert image into smaller one.");
				}
										
				// if we are inserting less dimensional image set dimensions 
				// to be equal and check correct size of dimension Z 
				if (dimens>dimensions) 
				{
#ifdef I3D_DEBUG
					cout << "  - Setting number of dimensions as in original file." << endl;
#endif	
					if (crop->size.z != 1)
						throw IOException("HDF5: Incorrect size of VOI in Z-dimension.");
						
					dimensions=dimens;
					dimsPosition[DIM_Z]=0;
					dimsPosition[DIM_Y]++;
					dimsPosition[DIM_X]++;
				}					
				
				// find original image's dimension sizes
				hsize_t *dims = new hsize_t[dimens];
				if (H5Sget_simple_extent_dims(dataSpaceID, dims, NULL) < 0)
					throw IOException("HDF5: Can't load dimension sizes of original image file.");
										
				// check correct dimension sizes of original image and 
				// set maximal possible size of image shift  
				if (((dimens==3) && (dims[ dimsPosition[DIM_Z] ] < crop->size.z+crop->offset.z)) || 
					(dims[ dimsPosition[DIM_Y] ] < crop->size.y+crop->offset.y) ||
					(dims[ dimsPosition[DIM_X] ] < crop->size.x+crop->offset.x))
				{
					throw IOException("HDF5: Inserting image position/size is out of range of original image file.");
					// TODO:
					//  Je mozne tu nastavit namiesto vynimky i zmenu velkosti 
					//  povodneho obrazka pomocou H5Dset_extent(dataSetID, dims),
					//  kde do pola dims sa nastavia velkosti jednotlivych dimenzii					
				}
								
				// set up maximal possible sizes for shift of new image
				if (dimens == 3) 
					maxSizes[DIM_Z] = dims[dimsPosition[DIM_Z]];

				maxSizes[DIM_Y] = dims[dimsPosition[DIM_Y]];
				maxSizes[DIM_X] = dims[dimsPosition[DIM_X]];
														
				delete[] dims;								
			}
			
			hsize_t *hyperSlab = new hsize_t[dimensions];
			hsize_t *offset = new hsize_t[dimensions];
									
			// ---------------------------------------------
			// check size of cropped area and compare it 
			//   with maximal possible dimensions sizes
			// ---------------------------------------------
			
			if ((crop->size.x > header.size.x) || (crop->size.y > header.size.y) || (crop->size.z > header.size.z))
				throw IOException("HDF5: size of insering image is bigger than image size");
			
			if (dimensions==3) 
			{
				if (crop->size.z + crop->offset.z <= maxSizes[DIM_Z]) 
				{
					hyperSlab[ dimsPosition[DIM_Z] ] = crop->size.z;
					offset[ dimsPosition[DIM_Z] ] = crop->offset.z;
				} 
				else 
				{
					throw IOException("HDF5: saving image data exceeds maximal possible Z-dimension size");
				}
			}
						
			if ((crop->size.y + crop->offset.y <= maxSizes[DIM_Y]) && (crop->size.x + crop->offset.x <= maxSizes[DIM_X])) 
			{
				hyperSlab[ dimsPosition[DIM_Y] ] = crop->size.y;
				offset[ dimsPosition[DIM_Y] ] = crop->offset.y;
				
				hyperSlab[ dimsPosition[DIM_X] ] = crop->size.x;
				offset[ dimsPosition[DIM_X] ] = crop->offset.x;
			} 
			else
			{
				throw IOException("HDF5: saving image data exceeds maximal possible dimensions sizes");
			}
			
			buffer = new T[static_cast<size_t>(hyperSlab[ dimsPosition[DIM_Z] ]*hyperSlab[ dimsPosition[DIM_Y] ]*hyperSlab[ dimsPosition[DIM_X] ])];
			
			size_t i = 0;

			for (size_t z = 0; z < hyperSlab[ dimsPosition[DIM_Z] ]; z++)
				for (size_t y = 0; y < hyperSlab[ dimsPosition[DIM_Y] ]; y++)
					for (size_t x = 0; x < hyperSlab[ dimsPosition[DIM_X] ]; x++) 
					{
						size_t index = x+y*header.size.x+z*header.size.y*header.size.x;
						buffer[i] = data[index];
								
						// find minimal and maximal values of image data for greyscale images
						if (checkMinMax)
						{
							if (i || z) 
							{
								minMax[0]= minValue(data[index],minMax[0]);
								minMax[1]= maxValue(data[index],minMax[1]);
							} 
							else
								minMax[0] = minMax[1] = data[index];
						}
						i++;
					}
									
			// get information about used filespace and set hyperslab in file for saving data in
			if ( H5Sselect_hyperslab(dataSpaceID, H5S_SELECT_SET, offset, NULL, hyperSlab, NULL)<0 ) 
			{
				delete[] buffer;
				throw IOException("HDF5: Error selecting hyperSlab, saving data could be corrupted");
			}						
			
#ifdef I3D_DEBUG
			cout << "  - Writing data" << endl;
#endif
			
			// create memory space for saving data and write them all into file to avoid duplicing data in memory
			hid_t memoryDataSpace = H5Screate_simple(dimensions, hyperSlab, NULL);

			if (H5Dwrite(dataSetID, dataImageType, memoryDataSpace, dataSpaceID, H5P_DEFAULT, buffer) < 0) 
			{
				delete[] buffer;
				throw IOException("HDF5: error writing image data");
			}
									
			// free all allocated arrays
			delete[] hyperSlab;
			delete[] offset;
			delete[] buffer;
			
		} 
		else  
		{
			// for non-cropped images use dafult saving to avoid high memory consuption
			if (checkMinMax) 
			{
				size_t imageSize = header.size.x*header.size.y*header.size.z;
				for (size_t i=0; i<imageSize; i++)
					if (i) 
					{
						minMax[0]= minValue(data[i],minMax[0]);
						minMax[1]= maxValue(data[i],minMax[1]);
					} 
					else
						minMax[0] = minMax[1] = data[i];
			}			
				
			if (H5Dwrite(dataSetID, dataImageType, H5S_ALL, H5S_ALL, H5P_DEFAULT, data) < 0)
				throw IOException("HDF5: error writing image data");			
		}
		
		// write minimal and maximal values of image data for greyscale images
		if (checkMinMax) 
		{
			const char* MIN_MAX_RANGE = "IMAGE_MINMAXRANGE";
				
			// if image is being inserted load previous values
			// if set and compare them with actual ones
			if (crop) 
			{					
				hid_t attribute;
					
				// check if attribute exists and were're able to open it
				if (!((attribute = H5Aopen(dataSetID, MIN_MAX_RANGE, H5P_DEFAULT))<0)) 
				{
					// check if attribute has equal type with image 
					hid_t attType = H5Aget_type(attribute);
					if (H5Tequal(attType, dataImageType)) 
					{														
						// load dataspace and check wheater its sizes fit 
						hid_t attributeSpace = H5Aget_space(attribute);
						int attributeDimensions = H5Sget_simple_extent_ndims(attributeSpace);
						hsize_t *attributeDims = new hsize_t[attributeDimensions];
							
						if (!(H5Sget_simple_extent_dims(attributeSpace, attributeDims, NULL)<0)) 
						{
							hsize_t dataSize = 1;
							for (int i=0; i<attributeDimensions; i++)
								dataSize*= attributeDims[i];
								
							T attData[2];
								
							// if attribute size is OK and we can read attribute data, load image 
							//   data and compare them with present minimum and maximum values
							if (dataSize == 2 && !(H5Aread(attribute, dataImageType, attData)<0)) 
							{									
								minMax[0] = minValue(minMax[0], attData[0]);
								minMax[1] = maxValue(minMax[1], attData[1]);
							}
						}
						delete[] attributeDims;
					}
					H5Aclose(attribute);
				}
					
				// remove previously set attribute so we can create a new one
				H5Adelete(dataSetID, MIN_MAX_RANGE);
			}
				
			hsize_t referenceDims[1] = {2};
			hid_t obj_id 	= H5Oopen(fd, DATASET_NAME, H5P_DEFAULT);
			hid_t sid 	= H5Screate_simple(1, referenceDims, NULL);
			hid_t attr_id 	= H5Acreate2(obj_id, MIN_MAX_RANGE, dataImageType, sid, H5P_DEFAULT, H5P_DEFAULT);
			H5Awrite(attr_id, dataImageType, minMax);
			H5Aclose(attr_id);
			H5Sclose(sid);  
			H5Oclose(obj_id);
		}				
	} 	
}

/****************************************************************************
	HDF5Writer
 ****************************************************************************/
HDF5Writer::HDF5Writer(const char* fname, const VOI<PIXELS> *voi) : ImageWriter(fname, voi) 
{
	// suppress error printing of libHDF5 to avoid 
	// confusion with error handling by exceptions
#ifndef I3D_DEBUG
	H5Eset_auto2(H5E_DEFAULT, NULL, NULL);	
#endif
	
	// set address and name for main image dataset used in files saved by i3d-core
	// (if dataset is not in a root group be sure to set correct descriptor of its parent group)
	DATASET_NAME = "Image";
	
	// set flag for creating new datasets
	newFile = true;
	
	// set error values for 
	dataSetID = -1;
	dataSpaceID = -1;
			
	// check inserting data in an already existing file
	if (voi)
	{
		if ( (fd = H5Fopen(fname, H5F_ACC_RDWR, H5P_DEFAULT)) < 0) 
		{
#ifdef I3D_DEBUG
			cout	<< "HDF5: Unable to open access to file \"" << fname << "\", creating new file" << endl;
#endif
		} 
		else 
		{
			if ((dataSetID = H5Dopen2(fd, DATASET_NAME, H5P_DEFAULT) ) < 0) 
			{	// file doesn't contain main image dataset used by i3d-core
				H5Fclose(fd);
				throw IOException("HDF5: Unable to open access to dataset \"" + string(DATASET_NAME) + "\"");
			} 
			else 
			{ 
				if ((dataSpaceID = H5Dget_space(dataSetID))<0) 
				{	// dataspace of image dataset is probably corrupted
					H5Dclose(dataSetID);
					H5Fclose(fd);
					throw IOException("HDF5: Error opening dataspace of dataset \"" + string(DATASET_NAME) + "\"");
				} 
				else 
				{
					newFile = false;	// set flag for skipping of creating new datasets
					parentGroup = fd;	// set parent group of image dataset as a root group
#ifdef I3D_DEBUG
					cout << "  HDF5: Image dataset correctly opened" << endl;
#endif					
					return; 
				}
			}
		}		
	}
	
#ifdef I3D_DEBUG
	cout << "  Writing new file" << endl;
#endif
	
	fd = H5Fcreate( fname, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
	parentGroup = fd;	// set parent group used for search as a root group
	if (fd < 0)
		throw IOException("HDF5: Unable to create " + string(fname) + ". ");
	
#ifdef I3D_DEBUG
		cout << "  HDF5: file successfully created" << endl;
#endif
}

/***************************************************************************
	~HDF5Writer
 ***************************************************************************/
HDF5Writer::~HDF5Writer() 
{	
	// Terminate access to the dataspace and dataset if some were open.
	if (!(dataSpaceID<0)) 
		H5Sclose(dataSpaceID);

	if (!(dataSetID<0)) 
		H5Dclose(dataSetID);
	
	// write all remainng memory data into file to avoid troubles 
	// in case of later program termination
	H5Fflush(fd, H5F_SCOPE_GLOBAL);
	H5Fclose(fd);
		
#ifdef I3D_DEBUG
	cout << "HDF5: file writing finnished" << endl;
#endif
	
}



} // end of namespace


#endif // end of WITH_HDF5
