/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** @file 
    @brief Ellipse fitting routines.

    @author Matus Kotry (Ko3@mail.muni.cz), Ondrej Danek (xdanek2@fi.muni.cz)
    @date 2011
*/ 

#ifdef __GNUG__
    #pragma implementation "EllipseFitting.h"
#endif

#include <math.h>
#include "regions.h"
#include "draw.h"
#include "EllipseFitting.h"

// Method parameters
const double MAX_CONC_ANGLE = 165.0;
const double MAX_HOLE_CONC_ANGLE = 150.0;
const double NEIGHB_PENALTY = 12;
const double MULTI_CONC_PENALTY = 1.7;
const double SEGMENTATION_STEP = 2;
const double MIN_SEGMENT_SIZE = 0.25;
const double ANGLE_SCORE = 1;
const double ANGLE_PENALIZATION = 2;
const double CONTAIN_HOLE_PENALIZATION = 100.0;
const double LIMIT_BASE = -1;
const double LIMIT_INCREMENT = 0.25;
const double MINIMAL_VOLUME = 0.02;
const double VOLUME_SCORE = 1.5;
const double MINIMAL_FINAL_SIZE = 0.1;

// Lapack external functions
#ifdef WITH_LAPACK
    typedef long int integer;
    typedef double doublereal;

    extern "C" void dgemm_(const char *transa, const char *transb, integer *m, integer *n,
	    integer *k, doublereal *alpha, doublereal *a, integer *lda,
	    doublereal *b, integer *ldb, doublereal *beta,
	    doublereal *c, integer *ldc);
    extern "C" int dgegv_(const char *jobvl, const char *jobvr, integer *n, doublereal *a, 
		   integer *lda, doublereal *b, integer *ldb, doublereal *alphar, 
		   doublereal *alphai, doublereal *beta, doublereal *vl, integer *ldvl, 
		   doublereal *vr, integer *ldvr, doublereal *work, integer *lwork, integer *info);
#endif

namespace i3d
{
    namespace EllipseFitting
    {
        ////////////////////////////////////////////////////////////////////////
        //
        // AUXILIARY STRUCTURES
        //
        ////////////////////////////////////////////////////////////////////////

        /** Concave point
          *
          * index = index of point in the original polygon
          * angle = angle of the adjacent lines of the point <0, 2PI>
          */
        struct concPoint 
        {
            unsigned int index;
            double angle;

            concPoint(unsigned int index, double angle) 
            { 
                this->index = index; 
                this->angle = angle; 
            }
        };

        /** Polygon segment. */
        class Segment 
        {
        private:
            VertexList points;
            int polygonID;
            int holeID;
            int segmentID;
            double prevConcavness;
            bool last;

        public:
            /** Constructor. */
            Segment(const VertexList &points, int polygonID, int segmentID, int holeID, double prevConcavness) 
            {
                this->points = points; 
                this->polygonID = polygonID; 
                this->holeID = holeID;
                this->segmentID = segmentID; 
                this->prevConcavness = prevConcavness; 
                last = false;
            }

            void setLast(bool islast)
            {
                last = islast;
            }

            const VertexList &getPoints() const 
            { 
                return points; 
            }

            int getPolygonID() const 
            { 
                return polygonID; 
            }

            int getHoleID() const
            { 
                return holeID; 
            }

            int getSegmentID() const 
            { 
                return segmentID; 
            }

            double getPrevConcavness() const
            { 
                return prevConcavness; 
            }

            void setPoints(const VertexList &points)
            { 
                this->points = points; 
            }

            bool isLast() const
            { 
                return last; 
            }

            bool isFirst() const
            { 
                return segmentID==0; 
            }
        };

        // Ellipse pair score
        struct PairScore 
        {
            PairScore(int i, int j) 
            { 
                this->i=i; 
                this->j=j; 
                score=0; 
            }; 

            int i,j;
            double score;
            Ellipse newEll;
            std::vector<Segment> origSegments;
        };

        ////////////////////////////////////////////////////////////////////////
        //
        // APPROXIMATION OF REGIONS BY POLYGONS - AUXILIARY METHODS
        //
        ////////////////////////////////////////////////////////////////////////

        const int NBS_X[8] = {  0,  1,  1,  1,  0, -1, -1, -1};
        const int NBS_Y[8] = { -1, -1,  0,  1,  1,  1,  0, -1};

        /** Distance of point from line.
        *	Calculates distance of a point from line given by two points.
        *
        * @param[in] pt The point.
        * @param[in] l1 First point defining the line.
        * @param[in] l2 Second point defining the line.
        *
        *	@retval double The distance.
        */
        static double pointLineDist2D(const Vertex &pt, const Vertex &l1, const Vertex &l2) 
        {
            double vx = l2.x - l1.x;
            double vy = l2.y - l1.y;
            double wx = pt.x - l1.x;
            double wy = pt.y - l1.y;

            double vSize = sqrt(pow(vx,2) + pow(vy,2));
            double wSize = sqrt(pow(wx,2) + pow(wy,2));

            if (vSize == 0 || wSize == 0)
            {
                return 0;
            }

            return wSize*sqrt(1-pow((vx*wx+vy*wy)/(vSize*wSize),2));
        }

        /** Approximation of a polygon.
        *	Approximates a polygon by replacing some unnecessary points (points that are almost 
        *	colinear) by single line.
        *
        * @param[in,out] poly The polygon that will be approximated.
        * @param[in] dTh Level of approximation (minimal distance of point from line given by 
        *				   two neighbour points)
        */
        static void approxPolygon(VertexList &poly, double dTh) 
        {
            if (poly.size() < 3)
            {
                return;
            }

            VertexList newPoly;

            size_t i = 0;
            size_t j = 2;
            size_t first = 0;
            size_t sz = poly.size();
            bool goOn = true;

            while (goOn && (j < sz || newPoly.size() > 2) && j < 2*sz) 
            {
                bool found = false;
                size_t k = i+1;

                while (k < j && !found) 
                {		
                    if (pointLineDist2D(poly[k%sz], poly[i%sz], poly[j%sz]) >= dTh) 
                    {
                        // New vertex found
                        i = k;

                        if (first == 0)
                        {
                            first = i + sz;
                        }
                        else if (i >= first) 
                        {	
                            // All points were processed => END
                            goOn = false;
                            break;
                        }

                        newPoly.push_back(poly[i%sz]);
                        found = true;
                    }

                    k++;
                }

                if (!found) 
                {
                    j++;
                }
            }

            poly = newPoly;	
        }

        void makePolygons(const Image3d<bool> &input, double dTh, PolygonSystem &polygons) 
        {
            LabeledImage3d<GRAY16, bool> labelIm;
            labelIm.CreateForegroundRegionsFF(input, nb2D_4);

            Image3d<bool> trackedPixels;
            trackedPixels.MakeRoom(input.GetSize());
            std::vector<bool> processedLabels(labelIm.get_max_label() + 1, false);

            for (size_t y = 0; y < input.GetSizeY(); y++) 
            {
                for (size_t x = 0; x < input.GetSizeX(); x++) 
                {
                    GRAY16 currLabel = labelIm.GetVoxel(x,y,0);

                    if (currLabel > 0) 
                    {
                        if (!processedLabels[currLabel-1]) 
                        {
                            // Track outer contour
                            Polygon poly;
                            int tx = int(x), ty = int(y), dir = 0;

                            processedLabels[currLabel-1] = true;

                            do 
                            {
                                poly.m_vertex_list.push_back(Vertex(tx,ty,0));
                                trackedPixels.SetVoxel(tx, ty, 0, true);

                                for (int k=dir; k<dir+8; k++) 
                                {
                                    int nx = tx+NBS_X[k%8];
                                    int ny = ty+NBS_Y[k%8];

                                    if (nx >= 0 && ny >= 0 && nx < (int)input.GetSizeX() && ny < (int)input.GetSizeY() && 
                                        labelIm.GetVoxel(nx,ny,0) == currLabel) 
                                    {
                                        tx = nx;
                                        ty = ny;
                                        dir = (k+6)%8;
                                        break;
                                    }
                                }
                            } while (tx != int(x) || ty != int(y));

                            approxPolygon(poly.m_vertex_list, dTh);

                            if (poly.m_vertex_list.size() > 2) 
                            {
                                poly.m_id = currLabel;
                                poly.m_is_hole = false;
                                polygons.push_back(poly);
                            }

                        } 
                        else if (y+1<input.GetSizeY() && labelIm.GetVoxel(x,y+1,0) == (GRAY16) 0
                            && trackedPixels.GetVoxel(x,y,0) == false) 
                        {
                            // Track inner contour
                            int tx = int(x), ty = int(y), dir = 4;
                            Polygon poly;

                            do 
                            {
                                poly.m_vertex_list.push_back(Vertex(tx,ty,0));
                                trackedPixels.SetVoxel(tx, ty, 0, true);

                                for (int k=dir; k<dir+8; k++) 
                                {
                                    int nx = tx+NBS_X[k%8];
                                    int ny = ty+NBS_Y[k%8];
                                    if (nx >= 0 && ny >= 0 && nx < (int)input.GetSizeX() && ny < (int)input.GetSizeY() && 
                                        labelIm.GetVoxel(nx,ny,0) == currLabel) 
                                    {
                                        tx = nx;
                                        ty = ny;
                                        dir = (k+6)%8;
                                        break;
                                    }
                                }
                            } while (tx != int(x) || ty != int(y));

                            approxPolygon(poly.m_vertex_list, dTh);

                            if (poly.m_vertex_list.size() > 2) 
                            {
                                poly.m_id = currLabel;
                                poly.m_is_hole = true;
                                polygons.push_back(poly);
                            }
                        }
                    }
                }
            }
        }

        /** Tests if a point lies in a polygon.
        *
        * @param[in] poly The polygon.
        * @param[in] pt The point.
        *
        *	@retval bool True if the point lies in the polygon.
        */
        static bool pointInPolygon(const VertexList &poly, const Vertex &pt) 
        {
            size_t j = poly.size() - 1;
            bool oddNodes = false;

            for (size_t i = 0; i < poly.size(); i++) 
            {
                if ((poly[i].y < pt.y && poly[j].y >= pt.y) || (poly[j].y < pt.y && poly[i].y >= pt.y))
                {
                    if (poly[i].y != poly[j].y) 
                    {		
                        double edge	= poly[i].x + ((pt.y - poly[i].y) / (poly[j].y - poly[i].y)) * (poly[j].x - poly[i].x);

                        if (edge < pt.x) 
                        {
                            oddNodes = !oddNodes; 
                        }
                    }
                }

                j = i;
            }

            return oddNodes;
        }

        /** Tests if a line given by two points lies in a polygon.
        *
        * @param[in] poly The polygon.
        * @param[in] pt1 First point defining the line.
        * @param[in] pt2 Second point defining the line.
        *
        *	@retval bool True if the line completely lies in the polygon.
        */
        static bool lineInPolygon(const VertexList &poly, const Vertex &pt1, const Vertex &pt2) 
        {
            Vertex vec = pt2 - pt1;

            int steps = int(std::max(fabs(vec.x), fabs(vec.y))) + 2; 
            vec /= double(steps);

            // Points pt1 and pt2 doesn't have to be inside - tolerance
            for (int i = 1; i < steps; i++) 
            {
                if (!pointInPolygon(poly, pt1 + double(i) * vec))
                {
                    return false;
                }
            }

            return true;
        }

        /** Calculates a reflection of a point over a line (in 2D)
        * @param[in] pt The point being reflected.
        * @param[in] ln1 First point defining a line.
        * @param[in] ln2 Second point defining a line.
        */
        static Vertex reflection2D(const Vertex &pt, const Vertex &ln1, const Vertex &ln2) 
        {
            Vertex vp = pt - ln1, vl = ln2 - ln1;
            double k = 2.0 * (Dot(vp, vl) / Dot(vl, vl));    
            return ln1 + ((k * vl) - vp);
        }

        /** Length of polygon
        * Can also calculate length of part of the polygon, when startIndex and endIndex are modified.
        * EndIndex can be also smaller than startIndex. Example:
        * polygon size = 10 points, startIndex = 8, endIndex = 2 => You will get length of segment 8-9-0-1-2.
        *
        * @param[in] poly The polygon.
        * @param[in] startIndex start index of the part of polygon.
        * @param[in] endIndex end index of the part of polygon.
        *
        *	@retval double Length of the polygon (or selected segment)
        */
        static double polygonLength(const VertexList &poly, size_t startIndex, size_t endIndex) 
        {
            size_t sz = poly.size();
            double size = 0;

            if (startIndex >= endIndex) 
            {
                endIndex += sz;
            }

            for (size_t i = startIndex; i < endIndex; i++) 
            {
                size += Norm(poly[(i+1)%sz] - poly[i%sz]);
            }

            return size;
        }

        /** Detection of concave points in polygon.
        * 
        * @param[in] poly The polygon.
        * @param[in] inverse If set to true, convex points will be detected instead.
        *
        *	@retval vector<concPoint> Vector of indexes and angles of the concave points
        */
        static void getConcavePoints(const VertexList &poly, bool inverse, std::vector<concPoint> &concavePoints) 
        {
            concavePoints.clear();

            size_t sz = poly.size();

            double avgLen = 0;
            for (size_t i = 0; i < sz; i++) 
            {
                avgLen += Norm(poly[(i+1)%sz] - poly[i]);
            }
            avgLen /= double(sz);

            for	(size_t i=1; i<=sz; i++) 
            {
                double a1 = atan2( (poly[(i-1)%sz].y - poly[i%sz].y) ,
                    (poly[(i-1)%sz].x - poly[i%sz].x) );
                double a2 = atan2( (poly[(i+1)%sz].y - poly[i%sz].y) ,
                    (poly[(i+1)%sz].x - poly[i%sz].x) );

                double conc = std::abs(a1-a2);

                if (conc > M_PI)
                {
                    conc = 2*M_PI - conc;
                }

                if (inverse == pointInPolygon(poly, (poly[(i+1)%sz] + poly[(i-1)%sz]) / 2.0))
                {	
                    if (i < sz) 
                    {
                        concavePoints.push_back(concPoint(i%sz,conc));
                    } 
                    else 
                    {
                        concavePoints.insert(concavePoints.begin(),concPoint(i%sz,conc));
                    }
                }
            }
        }

        ////////////////////////////////////////////////////////////////////////
        //
        // ELLIPSE FITTING
        //
        ////////////////////////////////////////////////////////////////////////

        bool Ellipse::Set(double a, double b, double c, double d, double f, double g, double shx, double shy) 
        {
            b = b/2.0; d = d/2.0; f = f/2.0;

            double dsc = b * b - a * c;	

            center.x = shx + ((c*d-b*f) / dsc);
            center.y = shy + ((a*f-b*d) / dsc);
            center.z = 0;

            double t1 = 2 * (a*f*f + c*d*d + g*b*b - 2*b*d*f - a*c*g);
            double t2 = sqrt((a-c) * (a-c) + 4*b*b);

            this->a = sqrt(t1 / (dsc * (t2 - (a+c)))); 
            this->b = sqrt(t1 / (dsc * (-t2 - (a+c))));

            if (b==0 && a<=c) this->angle = 0;
            if (b==0 && a>c)  this->angle = M_PI/2.0;
            if (b!=0 && a<=c) this->angle = atan(2*b/(a-c))/2.0;
            if (b!=0 && a>c)  this->angle = M_PI/2.0 + atan(2*b/(a-c))/2.0;
            if (this->b > this->a) this->angle += M_PI/2.0;

            return (this->a >= 2 && this->b >= 2);
        }

        bool Ellipse::containsPoint(const Vertex &pt, double tolerance) const
        {
            double major = getMajor() + tolerance;
            double minor = getMinor() + tolerance;

            double ptAngle = atan((pt.y - center.y) / (pt.x - center.x));
            ptAngle -= angle;

            if (cos(ptAngle) != 0)
            {
                ptAngle = atan((major/minor)*tan(ptAngle));
            }

            double maxDist = sqrt(pow(cos(ptAngle)*major,2) + pow(sin(ptAngle)*minor,2));

            return (Norm(center - pt) < maxDist);
        }

        /** Single ellipse fitting.
          * Fits an ellipse into given set of points.
          *
          * @param[in] poly The set of points.
          * @retval Ellipse The newly fitted ellipse.
          */
        static void fitEllipse(const VertexList &poly, Ellipse &el) 
        {
        #ifdef WITH_LAPACK
            size_t n = poly.size();

            if (n == 2) 
            {
                double a = Norm(poly[1] - poly[0]) / 2.0;		
                el.Set((poly[0] + poly[1]) / 2.0, a, a/2.0, atan2(poly[1].y - poly[0].y, poly[1].x - poly[0].x));
                return;
            }

            double *B = new double[6 * n];

            // build B - the coordinate system is shyfted by poly[0] to prevent numerical instability
            for (size_t i = 0; i < n; i++) 
            {
                B[i + 0 * n] = (poly[i].x - poly[0].x) * (poly[i].x - poly[0].x);
                B[i + 1 * n] = (poly[i].x - poly[0].x) * (poly[i].y - poly[0].y);
                B[i + 2 * n] = (poly[i].y - poly[0].y) * (poly[i].y - poly[0].y);
                B[i + 3 * n] = (poly[i].x - poly[0].x);
                B[i + 4 * n] = (poly[i].y - poly[0].y);
                B[i + 5 * n] = 1;
            }

            // M = B' * B
            double *M = new double[6*6];
            double *VR = new double[6*6];
            double *VL = new double[6*6];
            double *WORK = new double[2048];
            doublereal alpha=1, beta=0;
            integer m = 6, n_ = int(n), lwork=2048, info;
            dgemm_("t","n",&m,&m,&n_,&alpha,B,&n_,B,&n_,&beta, M, &m);

            // build C
            double C[36] = { 0, 0,-2, 0, 0, 0,
                0, 1, 0, 0, 0, 0,
                -2, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0};

            // solve generalized eigensystem
            double *vRe = new double[6];
            double *vIm = new double[6];
            double *vDen = new double[6];
            dgegv_("N","V",&m,M,&m,C,&m,vRe,vIm,vDen,VL,&m,VR,&m,WORK,&lwork,&info);

            delete [] B;
            delete [] M;

            int val = -1;
            double minim;

            // find smallest eigenvalue
            for (int i=0; i<6; i++) {
                if (vDen[i] != 0) {
                    if (val == -1 || minim > (vRe[i]/vDen[i])) {
                        minim = (vRe[i]/vDen[i]);
                        val = i;
                    } 
                }
            }

            double coefs[6];
            for (int i=0; i<6; i++) {
                coefs[i] = VR[6*val+i];
            }

            delete [] VR;
            delete [] VL;
            delete [] WORK;
            delete [] vRe;
            delete [] vIm;

            // make ellipse
            if (!el.Set(coefs[0], coefs[1], coefs[2], coefs[3], coefs[4], coefs[5], poly[0].x, poly[0].y))
            {
                double a = Norm(poly.front() - poly.back()) / 2.0;		
                el.Set((poly.front() + poly.back()) / 2.0, a, a/2.0, 
                    atan2(poly.back().y - poly.front().y, poly.back().x - poly.front().x));
            }
        #else
            throw i3d::InternalException("fitEllipse: LAPACK required.");
        #endif
        }

        /** Single ellipse fitting.
          * Unlike the fitEllipse() function, this function also resizes ellipses which exceed the set
          * of points too much, e.g. ellipses fitted to (almost) collinear set of points. This function
          * is internally used in the multiple ellipse fitting functions.
          *
          * @param[in] poly The set of points.
          * @param[in] tolerance How much must the ellipse exceed to be adjusted. 
          *
          * @retval Ellipse The newly fitted ellipse.
          */
        static void fitAndAdjustEllipse(const VertexList &poly, Ellipse &el, double tolerance = 1.2) 
        {
            fitEllipse(poly, el);

            double maxD = Norm(poly.back() - poly.front());
            for (size_t i = 0; i < poly.size() - 1; i++) 
            {
                for (size_t j = i + 1; j < poly.size(); j++) 
                {
                    double d = Norm(poly[j] - poly[i]);

                    if (d > maxD * 1.25)
                    {
                        return;
                    }
                }
            }

            if (el.getMajor() > tolerance*maxD) 
            {
                VertexList newPoly = poly;

                for (size_t i = 1; i < poly.size() - 1; i++) 
                {
                    newPoly.push_back(reflection2D(poly[i], poly.front(), poly.back()));
                }

                fitEllipse(newPoly, el);
            }
        }

        static void polyToPoints(const VertexList &poly, bool hole, size_t w, size_t h, VertexList &newPoly) 
        {
            newPoly.clear();

            for (size_t i = 0; i < poly.size() - 1; i++) 
            { 
                size_t j = i + 1;

                if ((poly[i].x < 4 || poly[i].x > w-5 || poly[i].y < 4 || poly[i].y > h-5) && 
                    (poly[j].x < 4 || poly[j].x > w-5 || poly[j].y < 4 || poly[j].y > h-5))
                {
                    newPoly.push_back(poly[i]);
                    continue;
                }

                Vertex dir = poly[j] - poly[i];
                double vectSize = Norm(dir);
                dir /= vectSize;

                for (double k = 0; k < vectSize; k += SEGMENTATION_STEP) 
                {
                    newPoly.push_back(poly[i] + k * dir);
                }
            }
        }

        static void combineSegments(const std::vector<Segment> &seg1, const std::vector<Segment> &seg2, 
            VertexList &result) 
        {
            result.clear();

            for (size_t i = 0; i < seg1.size(); i++) 
            {
                for (size_t j = 0; j < seg1[i].getPoints().size(); j++) 
                {
                    result.push_back(seg1[i].getPoints()[j]);
                }
            }

            for (size_t i = 0; i < seg2.size(); i++) 
            {
                for (size_t j = 0; j < seg2[i].getPoints().size(); j++) 
                {
                    result.push_back(seg2[i].getPoints()[j]);
                }
            }
        }

        static void combineSegments(const std::vector<Segment> &seg, VertexList &result) 
        {
            combineSegments(seg, std::vector<Segment>(), result);
        }

        // Test functions
        static double testNeighbourhood(const Ellipse &ell1, const Ellipse &ell2, 
            const std::vector<Segment> &seg1, const std::vector<Segment> &seg2)
        {	
            double angle;
            int none = true;
            for (size_t i = 0; i < seg1.size(); i++) 
            {
                for (size_t j = 0; j < seg2.size(); j++) 
                {
                    if (seg1[i].getHoleID() == seg2[j].getHoleID()) {
                        if ((seg1[i].getSegmentID() - seg2[j].getSegmentID() == 1)
                            ||	(seg2[j].isLast() && seg1[i].isFirst())) {
                                if (none) {
                                    angle = seg1[i].getPrevConcavness();
                                    none = false;
                                } else {
                                    return -NEIGHB_PENALTY;
                                }
                        }
                        if ((seg2[j].getSegmentID() - seg1[i].getSegmentID() == 1)
                            ||	(seg1[i].isLast() && seg2[j].isFirst())) {
                                if (none) {
                                    angle = seg1[i].getPrevConcavness();
                                    none = false;
                                } else {
                                    return -NEIGHB_PENALTY;
                                }
                        }
                    }
                }
            }

            if (none)
            {
                return 0.5;
            }

            return NEIGHB_PENALTY*(angle/M_PI-1);
        }

        static double testSize(const Ellipse &ell1, const Ellipse &ell2, 
            const std::vector<Segment> &seg1, const std::vector<Segment> &seg2, 
            double lowDist, double highDist)
        {
            double score = 0;

            // too small = bonus
            double optimalVol = lowDist*highDist*MINIMAL_VOLUME;
            double vol1 = ell1.getA()*ell1.getB();
            double vol2 = ell2.getA()*ell2.getB();

            if (vol1 < optimalVol*2 && vol1 < vol2) 
            {
                score += pow((optimalVol*2-vol1)*VOLUME_SCORE/(optimalVol),2);
            }

            if (vol2 < optimalVol*2 && vol2 < vol1) 
            {
                score += pow((optimalVol*2-vol2)*VOLUME_SCORE/(optimalVol),2);
            }

            // too big already = penalty
            if (ell1.getMinor() > lowDist) score -= (ell1.getMinor()-lowDist)/(lowDist/2.0);
            if (ell2.getMinor() > lowDist) score -= (ell2.getMinor()-lowDist)/(lowDist/2.0);

            // too thin ellipse = -1..inf each ellipse
            /*score += (ell1.getMajor()/ell1.getMinor()-highDist/lowDist)/ASPECT_DENOMINATOR;
            score += (ell2.getMajor()/ell2.getMinor()-highDist/lowDist)/ASPECT_DENOMINATOR;*/
            score += 1-2*(ell1.getMinor()/ell1.getMajor());
            score += 1-2*(ell2.getMinor()/ell2.getMajor());

            return score;
        }

        static double testEllipsesPosition(const Ellipse &ell1, const Ellipse &ell2, 
            const std::vector<Segment> &seg1, const std::vector<Segment> &seg2, 
            const Ellipse &newEll, double lowDist, double highDist) 
        {
            double score = 0;
            // low distance from newEll center = bonus
            double l, d;
            l = (lowDist+highDist)/2.0;
            d = Norm(newEll.getCenter() - ell1.getCenter());
            score += pow((l-d)/l, 2)*(l<d ? -1 : 1);
            d = Norm(newEll.getCenter() - ell2.getCenter());
            score += pow((l-d)/l, 2)*(l<d ? -1 : 1); 
            return score;
        }

        static bool testLinePosition(const Vertex &pt1, const Vertex &pt2, const std::vector<Segment> &seg1, 
            const std::vector<Segment> &seg2, size_t index, const PolygonSystem &ps) 
        {
            Vertex p1(pt1), p2(pt2);

            for (size_t i = 0; i < ps.size(); i++)
            {
                if (ps[i].m_id == index) 
                {
                    if (ps[i].m_is_hole == pointInPolygon(ps[i].m_vertex_list, p1)) 
                    {				
                        VertexList segmt; 
                        combineSegments(seg1, segmt);

                        double minDist = Norm(p1 - segmt[0]);
                        int minJ = 0;
                        for (size_t j = 1; j < segmt.size(); j++) 
                        {
                            double d = Norm(p1 - segmt[j]);

                            if (d < minDist) 
                            {
                                minDist = d;
                                minJ = j;
                            }
                        }

                        p1 = segmt[minJ];
                    }

                    if (ps[i].m_is_hole == pointInPolygon(ps[i].m_vertex_list, p2)) 
                    {
                        VertexList segmt;
                        combineSegments(seg2, segmt);

                        double minDist = Norm(p2 - segmt[0]);
                        int minJ = 0;
                        for (size_t j = 2; j < segmt.size(); j++) 
                        {
                            double d = Norm(p2 - segmt[j]);
                            if (d < minDist) {
                                minDist = d;
                                minJ = j;
                            }
                        }

                        p2 = segmt[minJ];
                    }

                    if (ps[i].m_is_hole == lineInPolygon(ps[i].m_vertex_list, p1, p2))
                    {
                        return false;
                    }
                }
            }

            return true;
        }

        static double testAngle(const Ellipse &ell1, const Ellipse &ell2) 
        {
            double score = 0;
            double lineAngle = atan((ell2.getCenter().y - ell1.getCenter().y) / (ell2.getCenter().x - ell1.getCenter().x));
            score += ANGLE_SCORE*(ANGLE_PENALIZATION*std::abs(sin(ell1.getAngle() - lineAngle))-ANGLE_PENALIZATION/2);
            score += ANGLE_SCORE*(ANGLE_PENALIZATION*std::abs(sin(ell2.getAngle() - lineAngle))-ANGLE_PENALIZATION/2);
            return score;
        }

        static int testLinePosition(const Ellipse &ell1, const Ellipse &ell2, 
            const std::vector<Segment> &seg1, const std::vector<Segment> &seg2, 
            size_t polyID, const PolygonSystem &ps, double thickness) 
        {
            if (!testLinePosition(ell1.getCenter(), ell2.getCenter(), seg1, seg2, polyID, ps))
            {
                return -100;
            }

            return 0;
        }

        static bool testCenterPosition(const Ellipse &ell, size_t index, double lowDist, const PolygonSystem &ps) 
        {
            for (size_t i = 0; i < ps.size(); i++) 
            {
                if (ps[i].m_id == index) 
                {
                    if (ps[i].m_is_hole == pointInPolygon(ps[i].m_vertex_list, ell.getCenter())) 
                    {
                        return false;
                    }
                }
            }
            return true;
        }

        static bool testHolePosition(const Ellipse &newEll, size_t index, const PolygonSystem &ps, double lowDist) 
        {
            for (size_t i=0; i<ps.size(); i++) 
            {
                if (ps[i].m_id == index && ps[i].m_is_hole) 
                {
                    double cx=0, cy=0;

                    for (size_t j = 0; j < ps[i].m_vertex_list.size(); j++) 
                    {
                        cx += ps[i].m_vertex_list[j].x;
                        cy += ps[i].m_vertex_list[j].y;
                    }

                    cx /= double(ps[i].m_vertex_list.size());
                    cy /= double(ps[i].m_vertex_list.size());

                    Vertex pt(cx, cy, 0);

                    if (pointInPolygon(ps[i].m_vertex_list, pt) && newEll.containsPoint(pt, lowDist/5.0))
                    {
                        return true;
                    }
                }
            }
            return false;
        }

        static bool bothHoles(const std::vector<Segment> &seg1, const std::vector<Segment> &seg2, 
            const PolygonSystem &ps) 
        {
            return (seg1.size() == 1 && seg2.size() == 1 
                && ps[seg1[0].getHoleID()].m_is_hole 
                && ps[seg2[0].getHoleID()].m_is_hole);
        }

        static bool isHoleOnly(const std::vector<Segment> &seg1, const PolygonSystem &ps) 
        {
            return (seg1.size() == 1 && ps[seg1[0].getHoleID()].m_is_hole);
        }

        // RECOMBINATION FUNCTION
        static void combineEllipses(std::vector<Ellipse> &ellipses, PolygonSystem &ps, 
            std::vector<std::vector<Segment> > &origSegments, 
            double lowDist, double highDist, size_t w, size_t h) 
        {
            std::vector<PairScore> scoreboard;
            std::vector<bool> checklist;
            std::vector<bool> keeplist;
            std::vector<Ellipse> newEllipses;
            std::vector<std::vector<Segment> > newSegments;
            bool added = true;
            double limit = LIMIT_BASE;

            // get score for each pair
            while (added) 
            {
                for (size_t i = 0; i < ellipses.size() - 1; i++) 
                {
                    for (size_t j = i + 1; j < ellipses.size(); j++) 
                    {
                        // Ellipses have to lie in the same polygon, and line connecting their centers
                        // have to lie inside too.
                        if (origSegments[i][0].getPolygonID() == origSegments[j][0].getPolygonID()
                            &&	!bothHoles(origSegments[i],origSegments[j],ps)) 
                        {
                            PairScore sc(i,j);

                            // distance: -inf..2 pt
                            sc.score += 2 * (1 - Norm(ellipses[j].getCenter() - ellipses[i].getCenter()) / highDist);

                            //testSize
                            sc.score += testSize(ellipses[i], ellipses[j], origSegments[i], origSegments[j], lowDist, highDist);

                            // test angles of major axes
                            sc.score += testAngle(ellipses[i], ellipses[j]);

                            //test if ellipse contains other
                            for (int q=0; q<4; q++) {
                                if (ellipses[i].containsPoint(ellipses[j].getCenter(), q*lowDist/8.0)) {
                                    sc.score += (10-3*q)*(isHoleOnly(origSegments[j], ps) ? 0.5 : 1);
                                    break;
                                }
                            }
                            for (int q=0; q<4; q++) {
                                if (ellipses[j].containsPoint(ellipses[i].getCenter(), q*lowDist/8.0)) {
                                    sc.score += (10-3*q)*(isHoleOnly(origSegments[i], ps) ? 0.5 : 1);
                                    break;
                                }
                            }

                            // if score is too low yet, there is no need to continue
                            if (sc.score < limit-5) {
                                continue;
                            }

                            // try to fit new ellipse
                            for (size_t l=0; l<origSegments[i].size(); l++) 
                            {
                                sc.origSegments.push_back(origSegments[i][l]);
                            }

                            for (size_t l=0; l<origSegments[j].size(); l++) 
                            {
                                sc.origSegments.push_back(origSegments[j][l]);
                            }

                            VertexList newSeg; 
                            combineSegments(sc.origSegments, newSeg);

                            Ellipse newEll; 
                            fitAndAdjustEllipse(newSeg, newEll);

                            // new ellipse center position: -4pt if not passed
                            if (!testCenterPosition(newEll, origSegments[i][0].getPolygonID(), lowDist, ps)) 
                            {
                                sc.score -= 4;
                                if (newEll.getCenter().x < 3 || newEll.getCenter().x > w-4
                                    ||  newEll.getCenter().y < 3 || newEll.getCenter().y > h-4) {
                                        sc.score += 2;
                                }
                            }

                            // new ellipse is not too big or small: 1pt or less
                            sc.score += 1.5-2*std::abs(newEll.getMajor()/newEll.getMinor()-highDist/lowDist);
                            if (newEll.getMajor() > highDist) {
                                sc.score -= pow(1.5 * ((newEll.getMajor()) / highDist), 3)-1;
                            } else if (newEll.getMajor() < lowDist) {
                                sc.score -= pow((lowDist-newEll.getMajor()) / (lowDist/2.0),2);
                            } else {
                                sc.score += 1;
                            }

                            // test if new ellipse contains a hole
                            if (testHolePosition(newEll, origSegments[i][0].getPolygonID(), ps, lowDist)) 
                            {
                                sc.score -= CONTAIN_HOLE_PENALIZATION;
                                if (testHolePosition(ellipses[i], origSegments[i][0].getPolygonID(), ps, lowDist)
                                    ||	testHolePosition(ellipses[j], origSegments[i][0].getPolygonID(), ps, lowDist)) 
                                {
                                    sc.score += CONTAIN_HOLE_PENALIZATION-3.0;
                                }
                            }

                            // various tests: -2pt..+2pt
                            sc.score += testEllipsesPosition(ellipses[i], ellipses[j], origSegments[i], origSegments[j], newEll, lowDist, highDist);

                            // test segment neighbourhood: -9pt..+0.5pt
                            sc.score += testNeighbourhood(ellipses[i], ellipses[j], origSegments[i], origSegments[j]);

                            // if score is too low yet, there is no need to continue
                            if (sc.score >= limit) 
                            {
                                // line between centers lies inside polygon: -4..-100pt if not passed
                                sc.score += testLinePosition(ellipses[i], ellipses[j], 
                                    origSegments[i], origSegments[j], 
                                    origSegments[i][0].getPolygonID(), ps, lowDist/8);

                                if (sc.score > limit) 
                                {
                                    size_t k = 0;
                                    sc.newEll = newEll;

                                    while (scoreboard.size() > k && scoreboard[k].score < sc.score) 
                                    {
                                        k++;
                                    }

                                    scoreboard.insert(scoreboard.begin()+k, sc);
                                } 
                            }
                        } 
                    }
                }

                // recombine ellipses
                checklist.resize(ellipses.size());
                keeplist.resize(ellipses.size());
                newEllipses.clear();
                newSegments.clear();

                for (size_t i = 0; i < ellipses.size(); i++) 
                {
                    checklist[i] = true;
                    keeplist[i] = true;
                }

                added = false;
                for (size_t i = scoreboard.size(); i-- > 0; )
                {
                    const PairScore &pair = scoreboard[i];

                    if (checklist[pair.i] && checklist[pair.j]) 
                    {
                        newEllipses.push_back(pair.newEll);
                        newSegments.push_back(pair.origSegments);
                        keeplist[pair.i] = false;
                        keeplist[pair.j] = false;
                        added = true;
                    }

                    checklist[pair.i] = false;
                    checklist[pair.j] = false;
                }

                // keep those, which are not recombined
                for (size_t i = 0; i < ellipses.size(); i++) 
                {
                    if (keeplist[i]) 
                    {
                        newEllipses.push_back(ellipses[i]);
                        newSegments.push_back(origSegments[i]);
                    }
                }

                ellipses = newEllipses;
                origSegments = newSegments;
                scoreboard.clear();
                limit += LIMIT_INCREMENT;
            }

            // delete too small ellipses
            newEllipses.clear();
            for (size_t i = 0; i < ellipses.size(); i++) 
            {
                if (ellipses[i].getA()*ellipses[i].getB() > MINIMAL_FINAL_SIZE*lowDist*highDist) 
                {
                    newEllipses.push_back(ellipses[i]);
                }
            }

            ellipses = newEllipses;
        }

        // POLYGON SEGMENTATION
        static void splitByConcavePoints(const VertexList &poly, unsigned int index, unsigned int hIndex, 
            bool inverse, double lowDist, size_t w, size_t h, std::vector<Segment> &result) 
        {
            result.clear();

            std::vector<concPoint> allConcaves; 
            getConcavePoints(poly, inverse, allConcaves);

            std::vector<concPoint> concaves;
            unsigned int sz = poly.size();
            double avgAngle = 0;
            int n = 0;

            for (size_t i=0; i<allConcaves.size(); i++) 
            {
                if (allConcaves[i].angle <= MAX_CONC_ANGLE*M_PI/180.0) 
                {
                    concaves.push_back(allConcaves[i]);
                }
            }

            if (concaves.size() <= 1) 
            {
                VertexList ptp;
                polyToPoints(poly, inverse, w, h, ptp);

                result.push_back(Segment(ptp, index, 0, hIndex, 0));
                result.back().setLast(true);
                return;
            }

            concaves.push_back(concPoint(concaves[0].index+sz,concaves[0].angle));
            int segId = 0;

            for (int i=concaves.size()-1; i>0; i--) {
                if (concaves[i].index - concaves[i-1].index == 1
                    &&	polygonLength(poly, concaves[i-1].index%sz, concaves[i].index%sz)<=lowDist*MIN_SEGMENT_SIZE) {
                        n++;
                        avgAngle += concaves[i-1].angle;
                } else {
                    break;
                }
            }

            for (size_t i = 0; i < concaves.size()-1; i++) 
            {
                VertexList newPoly;
                for (size_t j=concaves[i].index; j<=concaves[i+1].index; j++) 
                {
                    newPoly.push_back(poly[j%sz]);
                    if (poly[j%sz].x < 4 || poly[j%sz].x > w-5 || poly[j%sz].y < 4 || poly[j%sz].y > h-5) 
                    {
                        // -- novy segment --
                        n++;
                        avgAngle += concaves[i].angle;
                        if (newPoly.size() > 2 && polygonLength(poly, concaves[i].index, concaves[i+1].index) > lowDist*MIN_SEGMENT_SIZE) 
                        {
                            VertexList ptp;
                            polyToPoints(newPoly, inverse, w, h, ptp);

                            result.push_back(Segment(ptp, index, segId, hIndex, avgAngle/pow((double)n,MULTI_CONC_PENALTY)));
                            segId+=2;
                            n=0;
                            avgAngle=0;
                        } 
                        else if (newPoly.size() == 2 && polygonLength(poly, concaves[i].index, concaves[i+1].index) > lowDist*MIN_SEGMENT_SIZE) 
                        {
                            result.push_back(Segment(newPoly,index,segId,hIndex,avgAngle/pow((double)n,MULTI_CONC_PENALTY)));
                            segId+=2;
                            n=0;
                            avgAngle=0;
                        }
                        // ------------------
                        newPoly.clear();
                        while (j+2<concaves[i+1].index
                            && (poly[(j+2)%sz].x < 4 || poly[(j+2)%sz].x > w-5
                            || poly[(j+2)%sz].y < 4 || poly[(j+2)%sz].y > h-5)) {
                                j++;
                        }
                    }
                }
                // -- novy segment --
                n++;
                avgAngle += concaves[i].angle;
                if (newPoly.size() > 2 && polygonLength(poly, concaves[i].index, concaves[i+1].index) > lowDist*MIN_SEGMENT_SIZE) 
                {
                    VertexList ptp;
                    polyToPoints(newPoly, inverse, w, h, ptp);

                    result.push_back(Segment(ptp, index, segId, hIndex, avgAngle/pow((double)n,MULTI_CONC_PENALTY)));
                    segId++;
                    n=0;
                    avgAngle=0;
                } 
                else if (newPoly.size() == 2 && polygonLength(poly, concaves[i].index, concaves[i+1].index) > lowDist*MIN_SEGMENT_SIZE) 
                {
                    result.push_back(Segment(newPoly,index,segId,hIndex,avgAngle/pow((double)n,MULTI_CONC_PENALTY)));
                    segId++;
                    n=0;
                    avgAngle=0;
                }
                // ------------------
            }

            if (!result.empty()) 
            {
                result.back().setLast(true);
            }
        }

        static void splitByConcavePoints(PolygonSystem &polys, double lowDist, size_t w, size_t h, std::vector<Segment> &result) 
        {
            result.clear();

            std::vector<Segment> semiResult;

            for (size_t i=0; i < polys.size(); i++) 
            {
                splitByConcavePoints(polys[i].m_vertex_list, polys[i].m_id, i, polys[i].m_is_hole, lowDist, w, h, semiResult);

                for (size_t j=0; j < semiResult.size(); j++) 
                {
                    result.push_back(semiResult[j]);
                }
            }
        }

        static void reshapeHoles(PolygonSystem &polys) 
        {
            for (size_t i=0; i < polys.size(); i++) 
            {
                if (polys[i].m_is_hole) 
                {
                    VertexList newPoly;
                    std::vector<concPoint> concs; 
                    getConcavePoints(polys[i].m_vertex_list, true, concs);

                    size_t k=0;
                    for (size_t j=0; j < polys[i].m_vertex_list.size(); j++) 
                    {
                        if (k < concs.size() && concs[k].index == j) 
                        {
                            if (concs[k].angle > MAX_HOLE_CONC_ANGLE*M_PI/180.0) 
                            {
                                k++;
                                continue;
                            }
                            k++;
                        }

                        newPoly.push_back(polys[i].m_vertex_list[j]);
                    }

                    polys[i].m_vertex_list = newPoly;
                }
            }
        }

        // MULTIPLE ELLIPSES FITTING
        void fitEllipses(PolygonSystem &polys, double lowDist, double hiDist, size_t w, 
            size_t h, std::vector<Ellipse> &ellipses) 
        {
            ellipses.clear();

            reshapeHoles(polys);
            std::vector<Segment> segments;
            splitByConcavePoints(polys, lowDist, w, h, segments);

            std::vector<std::vector<Segment> > origSegments;

            for (size_t i=0; i<segments.size(); i++) 
            {
                Ellipse el;
                fitAndAdjustEllipse(segments[i].getPoints(), el);
                ellipses.push_back(el);

                if (segments[i].getPoints().size() == 2) 
                {
                    VertexList ptp;
                    polyToPoints(segments[i].getPoints(), polys[segments[i].getHoleID()].m_is_hole, w, h, ptp);
                    segments[i].setPoints(ptp);
                }

                std::vector<Segment> tmp;
                tmp.push_back(segments[i]);
                origSegments.push_back(tmp);		
            }

            combineEllipses(ellipses, polys, origSegments, lowDist, hiDist, w, h);
        }

        template<class VOXEL>
        void drawEllipse(const Ellipse &el, Image3d<VOXEL> &canvas, VOXEL color, bool showCenter, int steps)
        {
            double beta = -el.getAngle();
            double sinbeta = sin(beta);
            double cosbeta = cos(beta);

            int prevx = (int)round(el.getCenter().x + (el.getMajor() * cosbeta));
            int prevy = (int)round(el.getCenter().y - (el.getMajor() * sinbeta));

            for (double i=360.0/steps; i<=360; i+=360.0/steps) 
            {
                double alpha = i*(M_PI/180.0) ;
                double sinalpha = sin(alpha);
                double cosalpha = cos(alpha);

                Vector3d<size_t> pt;
                int ptx = (int)round(el.getCenter().x + (el.getMajor() * cosalpha * cosbeta - el.getMinor() * sinalpha * sinbeta));
                int pty = (int)round(el.getCenter().y - (el.getMajor() * cosalpha * sinbeta + el.getMinor() * sinalpha * cosbeta));

                if (ptx >= 0 && pty >= 0 && ptx < (int)canvas.GetSizeX() && pty < (int)canvas.GetSizeY() &&
                    prevx >= 0 && prevy >= 0 && prevx < (int)canvas.GetSizeX() && prevy < (int)canvas.GetSizeY())
                {
                    DrawLine(canvas, Vector3d<size_t>(prevx,prevy,0), Vector3d<size_t>(ptx,pty,0), color);
                }

                prevx = ptx;
                prevy = pty;
            }

            Vector3d<size_t> cs((size_t)el.getCenter().x, (size_t)el.getCenter().y, (size_t)el.getCenter().z);

            if (showCenter && cs.x >= 2 && cs.x < canvas.GetSizeX()-2 && cs.y >= 2 && cs.y < canvas.GetSizeY()-2) 
            {
                for (int i=-2; i<=2; i++) 
                {
                    canvas.SetVoxel(cs.x+i, cs.y+i, 0, color);
                    canvas.SetVoxel(cs.x-i, cs.y+i, 0, color);
                }
            }
        }

        // Explicit instantiations
        template I3D_ALGO_EXPORT void drawEllipse(const Ellipse &el, Image3d<bool> &canvas, bool color, bool showCenter, int steps);
        template I3D_ALGO_EXPORT void drawEllipse(const Ellipse &el, Image3d<GRAY8> &canvas, GRAY8 color, bool showCenter, int steps);
        template I3D_ALGO_EXPORT void drawEllipse(const Ellipse &el, Image3d<GRAY16> &canvas, GRAY16 color, bool showCenter, int steps);
        template I3D_ALGO_EXPORT void drawEllipse(const Ellipse &el, Image3d<RGB> &canvas, RGB color, bool showCenter, int steps);
    }
}
