/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * FastLevelSet.cc
 *
 * Fast level set-like algorithms intended for image segmentation as well as binary image deformations
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006 - 2011
 */
 
#include "FastLevelSet.h"
#include "DistanceTransform.h"
#include "fastlevelset/const_inflation_force.h"
#include "fastlevelset/variable_inflation_force.h"
#include "fastlevelset/p1_edge_stopping_function.h"
#include "fastlevelset/p2_edge_stopping_function.h"
#include "fastlevelset/const_edge_stopping_function.h"
#include "fastlevelset/geodesic_speed_function.h"
#include "fastlevelset/cv_speed_function.h"
#include "fastlevelset/rd_speed_function.h"
#include "fastlevelset/nha_mean_curvature_force.h"
#include "fastlevelset/nha_advection_force.h"
#include "fastlevelset/tp_nha_routine.h"
#include "fastlevelset/tp_nha_initializer.h"
#include "fastlevelset/tp_nha_domain_initializer.h"
#include "fastlevelset/tp_glob_reg_nha_routine.h"
#include "fastlevelset/tp_reg_nha_initializer.h"
#include "fastlevelset/tp_reg_nha_domain_initializer.h"
#include "fastlevelset/lab_object_manager.h"
#include "fastlevelset/lab_glob_reg_nha_routine.h"
#include "fastlevelset/lab_reg_nha_domain_initializer.h"
#include "fastlevelset/loc_reg_nha_routine.h"

namespace i3d
{
	namespace fls 
	{
		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                         Local Functions                     //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		/** Copy a given number of values from one buffer to the other one. */
		template <class VOXEL> static void Copy(const VOXEL *src, VOXEL *dst, Type::Size num, Type::Size step)
		{
			while (num > 0)
			{
				*dst = *src;
				dst += step;
				src += step;
				--num;
			}
		}

		/**********************************************************************/

		/** Copy a given number of values from one buffer to the other one. */
		template <class VOXEL> static void CopySlice(VOXEL *dst, Type::Size sx, Type::Size sy, Type::Size coef)
		{
			VOXEL *start = dst - 1;
			const VOXEL *src = dst + coef * sx;

			memcpy(dst, src, (sx - 2) * sizeof(VOXEL));

			dst += (sy - 1) * sx;
			src = dst - coef * sx;

			memcpy(dst, src, (sx - 2) * sizeof(VOXEL));

			dst = start;
			src = dst + coef;

			Copy(src, dst, sy, sx);

			dst = start + sx - 1;
			src = dst - coef;

			Copy(src, dst, sy, sx);
		}

		/**********************************************************************/

		/** Compute a gradient magnitude at one pixel. */
		template <class VOXEL> static void GradMag(const VOXEL *&src, Type::Prec *&dst, Type::Size right, Type::Size left, Type::Size upper, Type::Size lower, 
												   Type::Prec denom_x, Type::Prec denom_y)
		{
			Type::Prec dx = Type::Prec(*(src + right)) - Type::Prec(*(src - left));
			Type::Prec dy = Type::Prec(*(src + upper)) - Type::Prec(*(src - lower));

			*dst = Sqrt(dx * dx * denom_x + dy * dy * denom_y);
			++src, ++dst;
		}

		/**********************************************************************/

		/** Compute a gradient magnitude at each pixel of a row. */
		template <class VOXEL> static void RowGradMag(const VOXEL *&src, Type::Prec *&dst, Type::Size upper, Type::Size lower, Type::Prec denom_x, Type::Prec denom_y, Type::Size sx)
		{
			// process the first voxel in the row
			GradMag(src, dst, 1, 0, upper, lower, denom_x, denom_y);

			// process the remaining voxels in the row except the last one
			for (Type::Size x = 2; x < sx; ++x)
			{
				GradMag(src, dst, 1, 1, upper, lower, denom_x, denom_y);
			}

			// process the last voxel in the row
			GradMag(src, dst, 0, 1, upper, lower, denom_x, denom_y);

			dst += 2;
		}

		/**********************************************************************/

		/** Compute a gradient magnitude at one voxel. */
		template <class VOXEL> static void GradMag3D(const VOXEL *&src, Type::Prec *&dst, Type::Size right, Type::Size left, Type::Size upper, Type::Size lower, 
													 Type::Size top, Type::Size bottom, Type::Prec denom_x, Type::Prec denom_y, Type::Prec denom_z)
		{
			Type::Prec dx = Type::Prec(*(src + right)) - Type::Prec(*(src - left));
			Type::Prec dy = Type::Prec(*(src + upper)) - Type::Prec(*(src - lower));
			Type::Prec dz = Type::Prec(*(src + top)) - Type::Prec(*(src - bottom));

			*dst = Sqrt(dx * dx * denom_x + dy * dy * denom_y + dz * dz * denom_z);
			++src, ++dst;
		}

		/**********************************************************************/

		/** Compute a gradient magnitude at each voxel of a row. */
		template <class VOXEL> static void RowGradMag3D(const VOXEL *&src, Type::Prec *&dst, Type::Size upper, Type::Size lower, Type::Size top, 
														Type::Size bottom, Type::Prec denom_x, Type::Prec denom_y, Type::Prec denom_z, Type::Size sx)
		{
			// process the first voxel in the row
			GradMag3D(src, dst, 1, 0, upper, lower, top, bottom, denom_x, denom_y, denom_z);

			// process the remaining voxels in the row except the last one
			for (Type::Size x = 2; x < sx; ++x)
			{
				GradMag3D(src, dst, 1, 1, upper, lower, top, bottom, denom_x, denom_y, denom_z);
			}

			// process the last voxel in the row
			GradMag3D(src, dst, 0, 1, upper, lower, top, bottom, denom_x, denom_y, denom_z);

			dst += 2;
		}

		/**********************************************************************/

		/** Compute a gradient magnitude at each voxel of a slice. */
		template <class VOXEL> static void SliceGradMag3D(const VOXEL *&src, Type::Prec *&dst, Type::Size top, Type::Size bottom,  
														  Type::Prec denom_x, Type::Prec denom_y, Type::Prec denom_z, Type::Size sx, Type::Size sy)
		{
			// process the first row of the slice
			RowGradMag3D(src, dst, sx, 0, top, bottom, denom_x, denom_y, denom_z, sx);

			// process the remaining rows of the slice except the last one
			for (Type::Size y = 2; y < sy; ++y)
			{
				RowGradMag3D(src, dst, sx, sx, top, bottom, denom_x, denom_y, denom_z, sx);
			}

			// process the last row of the slice
			RowGradMag3D(src, dst, 0, sx, top, bottom, denom_x, denom_y, denom_z, sx);

			dst += 2 * sx + 4;
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                        NeumannBoundary                      //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		template <class VOXEL> void NeumannBoundary(Image3d<VOXEL> &img, bool central)
		{
			Type::Size coef = central ? 2 : 1;

			// the input image size
			Type::Size sx = img.GetSizeX();
			Type::Size sy = img.GetSizeY();
			Type::Size sz = img.GetSizeZ();

			if (sz == 1)
			{
				CopySlice(img.begin() + 1, sx, sy, coef);
			}
			else
			{
				Type::Size slice = sx * sy;
				VOXEL *dst = img.begin() + slice + 1;

				for (Type::Size z = 2; z < sz; ++z, dst += slice)
				{
					CopySlice(dst, sx, sy, coef);
				}

				dst = img.begin();
				memcpy(dst, dst + coef * slice, slice * sizeof(VOXEL));

				dst = img.end() - slice;
				memcpy(dst, dst - coef * slice, slice * sizeof(VOXEL));
			}
		}

		/**********************************************************************/

		/** Explicit instantiations. */
		template I3D_ALGO_EXPORT void NeumannBoundary(Image3d<GRAY8> &img, bool central);
		template I3D_ALGO_EXPORT void NeumannBoundary(Image3d<GRAY16> &img, bool central);
		template I3D_ALGO_EXPORT void NeumannBoundary(Image3d<float> &img, bool central);
		template I3D_ALGO_EXPORT void NeumannBoundary(Image3d<double> &img, bool central);

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                     GradientMagnitudeImage                  //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		template <class VOXEL> void GradientMagnitudeImage(const Image3d<VOXEL> &src, Image3d<Type::Prec> &dst)
		{
			// the input image size
			Type::Size sx = src.GetSizeX();
			Type::Size sy = src.GetSizeY();
			Type::Size sz = src.GetSizeZ();

			// check dimensionality
			bool is_2d = (sz == 1);

			// the input image resolution
			Vector3d<Type::Prec> res = src.GetResolution().GetRes();

			// prepare the output image
			dst.SetResolution(src.GetResolution());
			dst.SetOffset(src.GetOffset());
			dst.MakeRoom(sx + 2, sy + 2, is_2d ? 1 : sz + 2);

			// precomputed denominators intended for efficient computation of central differences in the x and y axes
			Type::Prec denom_x = Type::Prec(0.25) * res.x * res.x;
			Type::Prec denom_y = Type::Prec(0.25) * res.y * res.y;

			// go through the input image and compute the gradient magnitude using the central differences
			// the central differences are substituted with the right and left ones on the image border
			const VOXEL *src_ptr = src.begin();

			if (is_2d)
			{
				Type::Prec *dst_ptr = dst.begin() + sx + 3;

				// process the first row
				RowGradMag(src_ptr, dst_ptr, sx, 0, denom_x, denom_y, sx);

				// process the remaining rows except the last one
				for(Type::Size y = 2; y < sy; ++y)
				{
					RowGradMag(src_ptr, dst_ptr, sx, sx, denom_x, denom_y, sx);
				}

				// process the last row
				RowGradMag(src_ptr, dst_ptr, 0, sx, denom_x, denom_y, sx);
			}
			else
			{
				Type::Prec *dst_ptr = dst.begin() + (sx + 2) * (sy + 3) + 1;

				// precomputed denominator intended for efficient computation of central differences in the z axis
				Type::Prec denom_z = Type::Prec(0.25) * res.z * res.z;

				// the number of voxels in one slice
				Type::Size slice = sx * sy;

				// process the first slice
				SliceGradMag3D(src_ptr, dst_ptr, slice, 0, denom_x, denom_y, denom_z, sx, sy);

				// process the remaining slices except the last one
				for (Type::Size z = 2; z < sz; ++z)
				{
					SliceGradMag3D(src_ptr, dst_ptr, slice, slice, denom_x, denom_y, denom_z, sx, sy);
				}

				// process the last slice
				SliceGradMag3D(src_ptr, dst_ptr, 0, slice, denom_x, denom_y, denom_z, sx, sy);
			}

			// set the expanded image border according to Neumann (zero central differences)
			NeumannBoundary(dst);
		}

		/**********************************************************************/

		/** Explicit instantiations. */
		template I3D_ALGO_EXPORT void GradientMagnitudeImage(const Image3d<GRAY8> &src, Image3d<Type::Prec> &dst);
		template I3D_ALGO_EXPORT void GradientMagnitudeImage(const Image3d<GRAY16> &src, Image3d<Type::Prec> &dst);
		template I3D_ALGO_EXPORT void GradientMagnitudeImage(const Image3d<float> &src, Image3d<Type::Prec> &dst);
		template I3D_ALGO_EXPORT void GradientMagnitudeImage(const Image3d<double> &src, Image3d<Type::Prec> &dst);

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                     VarInflationForceField                  //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		template <class VOXEL> void VarInflationForceField(const Image3d<VOXEL> &src, VOXEL level, Image3d<Type::Prec> &dst)
		{
			// the input image size
			Type::Size sx = src.GetSizeX();
			Type::Size sy = src.GetSizeY();
			Type::Size sz = src.GetSizeZ();

			// check dimensionality
			bool is_2d = (sz == 1);

			// the input image resolution
			Vector3d<Type::Prec> res = src.GetResolution().GetRes();

			// prepare the output image
			dst.SetResolution(i3d::Resolution(1.0f, 1.0f, 1.0f));
			dst.SetOffset(src.GetOffset());
			dst.MakeRoom(sx + 2, sy + 2, is_2d ? 1 : sz + 2);

			// set border points to zero
			dst.SetBorderVoxels(Type::Prec(0));

			// find binary mask of given intensity level 
			const VOXEL *ptr_src = src.begin();

			if (is_2d)
			{
				Type::Prec *ptr_dst = dst.begin() + sx + 3;

				for (Type::Size y = 0; y < sy; ++y, ptr_dst += 2)
				{
					for (Type::Size x = 0; x < sx; ++x, ++ptr_src, ++ptr_dst)
					{
						*ptr_dst = (*ptr_src == level) ? Type::Prec(1) : Type::Prec(0);
					}
				}
			}
			else
			{
				Type::Prec *ptr_dst = dst.begin() + (sx + 2) * (sy + 3) + 1;
				Type::Size shift_z = 2 * (sx + 2);

				for (Type::Size z = 0; z < sz; ++z, ptr_dst += shift_z)
				{
					for (Type::Size y = 0; y < sy; ++y, ptr_dst += 2)
					{
						for (Type::Size x = 0; x < sx; ++x, ++ptr_src, ++ptr_dst)
						{
							*ptr_dst = (*ptr_src == level) ? Type::Prec(1) : Type::Prec(0);
						}
					}
				}
			}

			// compute a distance transform
			HybridSaito(dst, 0, Type::Prec(1), false);
		}

		/** Explicit instantiations. */
		template I3D_ALGO_EXPORT void VarInflationForceField(const Image3d<GRAY8> &src, GRAY8 level, Image3d<Type::Prec> &dst);
		template I3D_ALGO_EXPORT void VarInflationForceField(const Image3d<GRAY16> &src, GRAY16 level, Image3d<Type::Prec> &dst);
		template I3D_ALGO_EXPORT void VarInflationForceField(const Image3d<float> &src, float level, Image3d<Type::Prec> &dst);
		template I3D_ALGO_EXPORT void VarInflationForceField(const Image3d<double> &src, double level, Image3d<Type::Prec> &dst);

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                        RegionSpeedImage                     //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		template <class VOXEL> void RegionSpeedImage(const Image3d<VOXEL> &src, Image3d<Type::Prec> &dst)
		{
			// the output image size
			Type::Size sx = src.GetSizeX() + 2;
			Type::Size sy = src.GetSizeY() + 2;
			Type::Size sz = src.GetSizeZ();

			if (sz > 1)
			{
				sz += 2;
			}

			// prepare the output image
			dst.SetResolution(src.GetResolution());
			dst.SetOffset(src.GetOffset());
			dst.MakeRoom(sx, sy, sz);

			// find minimal and maximal values in the input image
			VOXEL min, max;
			src.GetRange(min, max);

			// if the input image is constant, the speed is equal to 1 everywhere
			if (max == min)
			{
				dst.SetAllVoxels(Type::Prec(1));
			}
			else
			{
				// expand the input image by 1 in each direction and stretch image data to be in the interval [0,1]
				Type::Prec coef = Reciproc(Type::Prec(max - min));
				const VOXEL *src_ptr = src.begin();

				if (sz == 1)
				{
					Type::Prec *dst_ptr = dst.begin() + sx + 1;

					for (Type::Size y = 2; y < sy; ++y, dst_ptr += 2)
					{
						for (Type::Size x = 2; x < sx; ++x, ++src_ptr, ++dst_ptr)
						{
							*dst_ptr = Type::Prec(*src_ptr - min) * coef;
						}
					}
				}
				else
				{
					Type::Prec *dst_ptr = dst.begin() + sx * (sy + 1) + 1;

					for (Type::Size z = 2; z < sz; ++z, dst_ptr += 2 * sx)
					{
						for (Type::Size y = 2; y < sy; ++y, dst_ptr += 2)
						{
							for (Type::Size x = 2; x < sx; ++x, ++src_ptr, ++dst_ptr)
							{
								*dst_ptr = Type::Prec(*src_ptr - min) * coef;
							}
						}
					}
				}
			}
		}

		/**********************************************************************/

		/** Explicit instantiations. */
		template I3D_ALGO_EXPORT void RegionSpeedImage(const Image3d<GRAY8> &src, Image3d<Type::Prec> &dst);
		template I3D_ALGO_EXPORT void RegionSpeedImage(const Image3d<GRAY16> &src, Image3d<Type::Prec> &dst);
		template I3D_ALGO_EXPORT void RegionSpeedImage(const Image3d<float> &src, Image3d<Type::Prec> &dst);
		template I3D_ALGO_EXPORT void RegionSpeedImage(const Image3d<double> &src, Image3d<Type::Prec> &dst);

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                           SegRoutine                        //
		//                                                             //
		/////////////////////////////////////////////////////////////////	

		/**********************************************************************/

		Routine<int> *SegRoutine(const Image3d<Type::Prec> &grad_mag, const Image3d<Type::BinVoxel> &mask, const Image3d<Type::BinVoxel> *domain, bool preserve_topology, 
				                 Type::Size radius, Type::Prec inflation_impact, Type::Prec curvature_impact, Type::Prec advection_impact, Type::Size exponent)
		{
			// grid size is given as the size of the gradient magnitude image
			Vector3d<Type::Size> grid_size = grad_mag.GetSize();

			// the gradient magnitude image has been constructed in such a way to have zero central differences on its border
			// therefore, its size has grown by two pixels in each dimension
			Vector3d<Type::Size> size = grid_size - Vector3d<Type::Size>(2, 2, grid_size.z > 1 ? 2 : 0);

			// check the size of the mask image as well as the domain one
			if (mask.GetSize() != size || (domain && domain -> GetSize() != size))
			{
				throw InternalException("Incompatible image size!");
			}

			// check whether the exponent is equal to 1 or 2, nothing else is supported
			if (exponent != 1 && exponent != 2)
			{
				throw InternalException("Unsupported exponent of the edge-stopping function!");
			}

			// grid resolution is given as the resolution of the gradient magnitude image
			Vector3d<Type::Prec> grid_res = grad_mag.GetResolution().GetRes();

			// base of the Nilsson-Heyden algorithm		
			NHA_Base *base = new NHA_Base(grid_size.x, grid_size.y, grid_size.z, grid_res.x, grid_res.y, grid_res.z);

			// circular neighbourhood at which the curvature is approximated
			NHA_CircularNeighbourhood *circular_neigh = new NHA_CircularNeighbourhood(*base, radius);

			// edge-stopping function
			EdgeStoppingFunction *esf = (exponent == 1) ? new P1EdgeStoppingFunction(grad_mag.begin()) : new P2EdgeStoppingFunction(grad_mag.begin());

			// inflation force
			InflationForce *inf_force = new ConstInflationForce(inflation_impact);

			// mean curvature force used in the Nilsson-Heyden algorithm
			MeanCurvatureForce *curv_force = new NHA_MeanCurvatureForce(base -> m_curvature.begin(), circular_neigh -> ReciprocSize());

			// advection force used in the Nilsson-Heyden algorithm
			AdvectionForce *adv_force = (size.z > 1) ? new NHA_AdvectionForce3D(*base, grad_mag.begin()) : new NHA_AdvectionForce(*base, grad_mag.begin());

			// speed function used in the Nilsson-Heyden algorithm
			SpeedFunction *speed_function = new GeodesicSpeedFunction(esf, inf_force, curv_force, curvature_impact, adv_force, advection_impact);

			int num = 0;
			if (size.z > 1)
			{
				++num;
			}

			if (domain)
			{
				num += 2;
			}

			if (preserve_topology)
			{
				num += 4;
			}

			switch (num)
			{
				case 0:
					
					return new NHA_Routine(base,
										   new NHA_Initializer(*base, mask.begin()),
										   new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
										   speed_function, 
										   new NHA_Curvature(*base, circular_neigh, speed_function, base -> m_heap), 
										   new NHA_ResultCreator(*base));

				case 1:

					return new NHA_Routine(base, 
										   new NHA_Initializer3D(*base, mask.begin()),
										   new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
										   speed_function, 
										   new NHA_Curvature3D(*base, circular_neigh, speed_function, base -> m_heap), 
										   new NHA_ResultCreator3D(*base));

				case 2:

					return new NHA_Routine(base, 
                						   new NHA_DomainInitializer(*base, mask.begin(), domain -> begin()),
										   new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
										   speed_function, 
										   new NHA_Curvature(*base, circular_neigh, speed_function, base -> m_heap), 
										   new NHA_ResultCreator(*base));

				case 3:

					return new NHA_Routine(base,
										   new NHA_DomainInitializer3D(*base, mask.begin(), domain -> begin()),
										   new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
										   speed_function, 
										   new NHA_Curvature3D(*base, circular_neigh, speed_function, base -> m_heap), 
										   new NHA_ResultCreator3D(*base));

				case 4:

					return new TP_NHA_Routine(base,
											  new TP_NHA_Initializer(*base, mask.begin()),
											  new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
											  speed_function, 
											  new NHA_Curvature(*base, circular_neigh, speed_function, base -> m_heap), 
											  new NHA_ResultCreator(*base),
											  new NHA_SimplePointInspector(*base));

				case 5:

					return new TP_NHA_Routine(base, 
											  new TP_NHA_Initializer3D(*base, mask.begin()),
											  new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
											  speed_function, 
											  new NHA_Curvature3D(*base, circular_neigh, speed_function, base -> m_heap), 
											  new NHA_ResultCreator3D(*base),
											  new NHA_SimplePointInspector3D(*base));

				case 6:

					return new TP_NHA_Routine(base,
  											  new TP_NHA_DomainInitializer(*base, mask.begin(), domain -> begin()),
											  new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
											  speed_function, 
											  new NHA_Curvature(*base, circular_neigh, speed_function, base -> m_heap), 
											  new NHA_ResultCreator(*base),
											  new NHA_SimplePointInspector(*base));

				case 7:

					return new TP_NHA_Routine(base,
											  new TP_NHA_DomainInitializer3D(*base, mask.begin(), domain -> begin()),
											  new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
											  speed_function, 
											  new NHA_Curvature3D(*base, circular_neigh, speed_function, base -> m_heap), 
											  new NHA_ResultCreator3D(*base),
											  new NHA_SimplePointInspector3D(*base));

				default: throw InternalException("Unreachable code executed!");
			}
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                          VarSegRoutine                      //
		//                                                             //
		/////////////////////////////////////////////////////////////////	

		/**********************************************************************/

		Routine<int> *VarSegRoutine(const Image3d<Type::Prec> &grad_mag, const Image3d<Type::BinVoxel> &mask, const Image3d<Type::BinVoxel> *domain, bool preserve_topology, 
								    Type::Size radius, const Image3d<Type::Prec> &inflation_impact, Type::Prec curvature_impact, Type::Prec advection_impact, Type::Size exponent)
		{
			// grid size is given as the size of the gradient magnitude image
			Vector3d<Type::Size> grid_size = grad_mag.GetSize();

			// the gradient magnitude image has been constructed in such a way to have zero central differences on its border
			// therefore, its size has grown by two pixels in each dimension
			Vector3d<Type::Size> size = grid_size - Vector3d<Type::Size>(2, 2, grid_size.z > 1 ? 2 : 0);

			// check the size of the mask image as well as the domain one
			if (mask.GetSize() != size || (domain && domain -> GetSize() != size))
			{
				throw InternalException("Incompatible image size!");
			}

			// check the size of the inflation force field
			if (inflation_impact.GetSize() != grid_size)
			{
				throw InternalException("Incompatible size of the inflation force field!");
			}

			// check whether the exponent is equal to 1 or 2, nothing else is supported
			if (exponent != 1 && exponent != 2)
			{
				throw InternalException("Unsupported exponent of the edge-stopping function!");
			}

			// grid resolution is given as the resolution of the gradient magnitude image
			Vector3d<Type::Prec> grid_res = grad_mag.GetResolution().GetRes();

			// base of the Nilsson-Heyden algorithm		
			NHA_Base *base = new NHA_Base(grid_size.x, grid_size.y, grid_size.z, grid_res.x, grid_res.y, grid_res.z);

			// circular neighbourhood at which the curvature is approximated
			NHA_CircularNeighbourhood *circular_neigh = new NHA_CircularNeighbourhood(*base, radius);

			// edge-stopping function
			EdgeStoppingFunction *esf = (exponent == 1) ? new P1EdgeStoppingFunction(grad_mag.begin()) : new P2EdgeStoppingFunction(grad_mag.begin());

			// inflation force
			InflationForce *inf_force = new VariableInflationForce(inflation_impact.begin());

			// mean curvature force used in the Nilsson-Heyden algorithm
			MeanCurvatureForce *curv_force = new NHA_MeanCurvatureForce(base -> m_curvature.begin(), circular_neigh -> ReciprocSize());

			// advection force used in the Nilsson-Heyden algorithm
			AdvectionForce *adv_force = (size.z > 1) ? new NHA_AdvectionForce3D(*base, grad_mag.begin()) : new NHA_AdvectionForce(*base, grad_mag.begin());

			// speed function used in the Nilsson-Heyden algorithm
			SpeedFunction *speed_function = new GeodesicSpeedFunction(esf, inf_force, curv_force, curvature_impact, adv_force, advection_impact);

			int num = 0;
			if (size.z > 1)
			{
				++num;
			}

			if (domain)
			{
				num += 2;
			}

			if (preserve_topology)
			{
				num += 4;
			}

			switch (num)
			{
				case 0:

					return new NHA_Routine(base, 
										   new NHA_Initializer(*base, mask.begin()),
										   new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
										   speed_function, 
										   new NHA_Curvature(*base, circular_neigh, speed_function, base -> m_heap), 
										   new NHA_ResultCreator(*base));

				case 1:

					return new NHA_Routine(base,
										   new NHA_Initializer3D(*base, mask.begin()),
										   new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
										   speed_function, 
										   new NHA_Curvature3D(*base, circular_neigh, speed_function, base -> m_heap),  
										   new NHA_ResultCreator3D(*base));

				case 2:

					return new NHA_Routine(base,
      						               new NHA_DomainInitializer(*base, mask.begin(), domain -> begin()),
										   new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
										   speed_function, 
										   new NHA_Curvature(*base, circular_neigh, speed_function, base -> m_heap), 
										   new NHA_ResultCreator(*base));

				case 3:

					return new NHA_Routine(base, 
										   new NHA_DomainInitializer3D(*base, mask.begin(), domain -> begin()),
										   new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
										   speed_function, 
										   new NHA_Curvature3D(*base, circular_neigh, speed_function, base -> m_heap), 
										   new NHA_ResultCreator3D(*base));

				case 4:

					return new TP_NHA_Routine(base, 
 											  new TP_NHA_Initializer(*base, mask.begin()),
											  new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
											  speed_function, 
											  new NHA_Curvature(*base, circular_neigh, speed_function, base -> m_heap), 
											  new NHA_ResultCreator(*base),
											  new NHA_SimplePointInspector(*base));

				case 5:

					return new TP_NHA_Routine(base, 
											  new TP_NHA_Initializer3D(*base, mask.begin()),
											  new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
											  speed_function, 
											  new NHA_Curvature3D(*base, circular_neigh, speed_function, base -> m_heap), 
											  new NHA_ResultCreator3D(*base),
											  new NHA_SimplePointInspector3D(*base));

				case 6:

					return new TP_NHA_Routine(base, 
											  new TP_NHA_DomainInitializer(*base, mask.begin(), domain -> begin()),
											  new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
											  speed_function, 
											  new NHA_Curvature(*base, circular_neigh, speed_function, base -> m_heap), 
											  new NHA_ResultCreator(*base),
											  new NHA_SimplePointInspector(*base));

				case 7:

					return new TP_NHA_Routine(base, 
											  new TP_NHA_DomainInitializer3D(*base, mask.begin(), domain -> begin()),
											  new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
											  speed_function, 
											  new NHA_Curvature3D(*base, circular_neigh, speed_function, base -> m_heap), 
											  new NHA_ResultCreator3D(*base),
											  new NHA_SimplePointInspector3D(*base));

				default: throw InternalException("Unreachable code executed!");
			}
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                           DefRoutine                        //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		Routine<int> *DefRoutine(const Image3d<Type::Prec> &speed, const Image3d<Type::BinVoxel> &mask, const Image3d<Type::BinVoxel> *domain, 
								 Type::Size radius, Type::Prec inflation_impact, Type::Prec curvature_impact)
		{
			// grid size is given as the size of the speed image
			Vector3d<Type::Size> grid_size = speed.GetSize();

			// the speed image has been constructed in such a way to have zero central differences on its border
			// therefore, its size has grown by two pixels in each dimension
			Vector3d<Type::Size> size = grid_size - Vector3d<Type::Size>(2, 2, grid_size.z > 1 ? 2 : 0);

			// check the size of the mask image as well as the domain one
			if (mask.GetSize() != size || (domain && domain -> GetSize() != size))
			{
				throw InternalException("Incompatible image size!");
			}

			// grid resolution is given as the resolution of the speed image
			Vector3d<Type::Prec> grid_res = speed.GetResolution().GetRes();

			// base of the Nilsson-Heyden algorithm		
			NHA_Base *base = new NHA_Base(grid_size.x, grid_size.y, grid_size.z, grid_res.x, grid_res.y, grid_res.z);

			// circular neighbourhood at which the curvature is approximated
			NHA_CircularNeighbourhood *circular_neigh = new NHA_CircularNeighbourhood(*base, radius);

			// speed function used in the Nilsson-Heyden algorithm
			// the speed at the i-th grid point is given as speed[i] * (inflation_impact + curvature_impact * curv[i])
			SpeedFunction *speed_function = new GeometricSpeedFunction(new ConstEdgeStoppingFunction(speed.begin()), 
                                                        			   new ConstInflationForce(inflation_impact), 
																	   new NHA_MeanCurvatureForce(base -> m_curvature.begin(), circular_neigh -> ReciprocSize()),
																	   curvature_impact);

			int num = 0;
			if (size.z > 1)
			{
				++num;
			}

			if (domain)
			{
				num += 2;
			}

			switch (num)
			{
				case 0:

					return new TP_NHA_Routine(base, 
											  new TP_NHA_Initializer(*base, mask.begin()),
											  new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
											  speed_function, 
											  new NHA_Curvature(*base, circular_neigh, speed_function, base -> m_heap), 
											  new NHA_ResultCreator(*base),
											  new NHA_SimplePointInspector(*base));

				case 1:

					return new TP_NHA_Routine(base, 
											  new TP_NHA_Initializer3D(*base, mask.begin()),
											  new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
											  speed_function, 
											  new NHA_Curvature3D(*base, circular_neigh, speed_function, base -> m_heap), 
											  new NHA_ResultCreator3D(*base),
											  new NHA_SimplePointInspector3D(*base));

				case 2:

					return new TP_NHA_Routine(base, 
											  new TP_NHA_DomainInitializer(*base, mask.begin(), domain -> begin()),
											  new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
											  speed_function, 
											  new NHA_Curvature(*base, circular_neigh, speed_function, base -> m_heap), 
											  new NHA_ResultCreator(*base),
											  new NHA_SimplePointInspector(*base));

				case 3:

					return new TP_NHA_Routine(base, 
											  new TP_NHA_DomainInitializer3D(*base, mask.begin(), domain -> begin()),
											  new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
											  speed_function, 
											  new NHA_Curvature3D(*base, circular_neigh, speed_function, base -> m_heap), 
											  new NHA_ResultCreator3D(*base),
											  new NHA_SimplePointInspector3D(*base));

				default: throw InternalException("Unreachable code executed!");
			}
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                       GlobChanVeseRoutine                   //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		Routine<int> *GlobChanVeseRoutine(const Image3d<Type::Prec> &input, const Image3d<Type::BinVoxel> &mask, const Image3d<Type::BinVoxel> *domain, bool preserve_topology, 
									      Type::Size radius, Type::Prec curvature_impact, Type::Prec foreground_impact, Type::Prec background_impact)
		{
			// grid size is given as the size of the input image
			Vector3d<Type::Size> grid_size = input.GetSize();

			// the input image has been constructed in such a way to have zero central differences on its border
			// therefore, its size has grown by two pixels in each dimension
			Vector3d<Type::Size> size = grid_size - Vector3d<Type::Size>(2, 2, grid_size.z > 1 ? 2 : 0);

			// check the size of the mask image as well as the domain one
			if (mask.GetSize() != size || (domain && domain -> GetSize() != size))
			{
				throw InternalException("Incompatible image size!");
			}

			// grid resolution is given as the resolution of the input image
			Vector3d<Type::Prec> grid_res = input.GetResolution().GetRes();

			// base of the global region-based version of the Nilsson-Heyden algorithm		
			Glob_Reg_NHA_Base *base = new Glob_Reg_NHA_Base(grid_size.x, grid_size.y, grid_size.z, grid_res.x, grid_res.y, grid_res.z);

			// circular neighbourhood at which the curvature is approximated
			NHA_CircularNeighbourhood *circular_neigh = new NHA_CircularNeighbourhood(*base, radius);

			// manager for efficient computation of foreground and background object statistics
			ObjectManager *obj_manager = new ObjectManager(input.begin());

			// speed function used in the Chan-Vese model
			ChanVeseSpeedFunction *speed_function = new ChanVeseSpeedFunction(new NHA_MeanCurvatureForce(base -> m_curvature.begin(), circular_neigh -> ReciprocSize()),
																	          curvature_impact, obj_manager, foreground_impact, background_impact);

			int num = 0;
			if (size.z > 1)
			{
				++num;
			}

			if (domain)
			{
				num += 2;
			}

			if (preserve_topology)
			{
				num += 4;
			}

			switch (num)
			{
				case 0:

					return new Glob_Reg_NHA_Routine(base,
													obj_manager, 
													new Reg_NHA_Initializer(*base, mask.begin(), *obj_manager),
													new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
													speed_function,
													new NHA_CurvatureBase(*base, circular_neigh), 
													new NHA_ResultCreator(*base));

				case 1:

					return new Glob_Reg_NHA_Routine(base, 
													obj_manager, 
													new Reg_NHA_Initializer3D(*base, mask.begin(), *obj_manager),
													new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
													speed_function,
													new NHA_CurvatureBase3D(*base, circular_neigh), 
													new NHA_ResultCreator3D(*base));

				case 2:

					return new Glob_Reg_NHA_Routine(base, 
													obj_manager,	
													new Reg_NHA_DomainInitializer(*base, mask.begin(), *obj_manager, domain -> begin()),
													new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
													speed_function,
													new NHA_CurvatureBase(*base, circular_neigh), 
													new NHA_ResultCreator(*base));

				case 3:

					return new Glob_Reg_NHA_Routine(base, 
													obj_manager,
													new Reg_NHA_DomainInitializer3D(*base, mask.begin(), *obj_manager, domain -> begin()),
													new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
													speed_function,
													new NHA_CurvatureBase3D(*base, circular_neigh), 
													new NHA_ResultCreator3D(*base));

				case 4:

					return new TP_Glob_Reg_NHA_Routine(base, 
													   obj_manager,
													   new TP_Reg_NHA_Initializer(*base, mask.begin(), *obj_manager),
													   new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
													   speed_function,
													   new NHA_CurvatureBase(*base, circular_neigh), 
													   new NHA_ResultCreator(*base),
													   new NHA_SimplePointInspector(*base));

				case 5:

					return new TP_Glob_Reg_NHA_Routine(base, 
													   obj_manager,
													   new TP_Reg_NHA_Initializer3D(*base, mask.begin(), *obj_manager),
													   new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
													   speed_function,
													   new NHA_CurvatureBase3D(*base, circular_neigh), 
													   new NHA_ResultCreator3D(*base),
													   new NHA_SimplePointInspector3D(*base));

				case 6:

					return new TP_Glob_Reg_NHA_Routine(base, 
													   obj_manager,
													   new TP_Reg_NHA_DomainInitializer(*base, mask.begin(), *obj_manager, domain -> begin()),
													   new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
													   speed_function,
													   new NHA_CurvatureBase(*base, circular_neigh), 
													   new NHA_ResultCreator(*base),
													   new NHA_SimplePointInspector(*base));

				case 7:

					return new TP_Glob_Reg_NHA_Routine(base,
													   obj_manager,
													   new TP_Reg_NHA_DomainInitializer3D(*base, mask.begin(), *obj_manager, domain -> begin()),
													   new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
													   speed_function,
													   new NHA_CurvatureBase3D(*base, circular_neigh), 
													   new NHA_ResultCreator3D(*base),
													   new NHA_SimplePointInspector3D(*base));

				default: throw InternalException("Unreachable code executed!");
			}
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                        LocChanVeseRoutine                   //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		Routine<int> *LocChanVeseRoutine(const Image3d<Type::Prec> &input, const Image3d<Type::BinVoxel> &mask, const Image3d<Type::BinVoxel> *domain,
									     Type::Size radius, Type::Prec curvature_impact, Type::Prec foreground_impact, Type::Prec background_impact)
		{
			// grid size is given as the size of the input image
			Vector3d<Type::Size> grid_size = input.GetSize();

			// the input image has been constructed in such a way to have zero central differences on its border
			// therefore, its size has grown by two pixels in each dimension
			Vector3d<Type::Size> size = grid_size - Vector3d<Type::Size>(2, 2, grid_size.z > 1 ? 2 : 0);

			// check the size of the mask image as well as the domain one
			if (mask.GetSize() != size || (domain && domain -> GetSize() != size))
			{
				throw InternalException("Incompatible image size!");
			}

			// grid resolution is given as the resolution of the input image
			Vector3d<Type::Prec> grid_res = input.GetResolution().GetRes();

			// base of the local region-based version of the Nilsson-Heyden algorithm		
			Loc_Reg_NHA_Base *base = new Loc_Reg_NHA_Base(grid_size.x, grid_size.y, grid_size.z, grid_res.x, grid_res.y, grid_res.z);

			// circular neighbourhood at which the curvature is approximated
			NHA_CircularNeighbourhood *circular_neigh = new NHA_CircularNeighbourhood(*base, radius);

			// manager for efficient computation of foreground and background object statistics
			ObjectManager *obj_manager = new ObjectManager(input.begin());

			// speed function used in the Chan-Vese model
			ChanVeseSpeedFunction *speed_function = new ChanVeseSpeedFunction(new NHA_MeanCurvatureForce(base -> m_curvature.begin(), circular_neigh -> ReciprocSize()),
																	          curvature_impact, obj_manager, foreground_impact, background_impact);

			int num = 0;
			if (size.z > 1)
			{
				++num;
			}

			if (domain)
			{
				num += 2;
			}


			switch (num)
			{
				case 0:

					return new Loc_Reg_NHA_Routine(base,
												   obj_manager, 
												   new Reg_NHA_Initializer(*base, mask.begin(), *obj_manager),
												   new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
												   speed_function,
												   new NHA_CurvatureBase(*base, circular_neigh), 
												   new NHA_ResultCreator(*base));

				case 1:

					return new Loc_Reg_NHA_Routine(base, 
												   obj_manager, 
												   new Reg_NHA_Initializer3D(*base, mask.begin(), *obj_manager),
												   new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
												   speed_function,
												   new NHA_CurvatureBase3D(*base, circular_neigh), 
												   new NHA_ResultCreator3D(*base));

				case 2:

					return new Loc_Reg_NHA_Routine(base, 
												   obj_manager,	
												   new Reg_NHA_DomainInitializer(*base, mask.begin(), *obj_manager, domain -> begin()),
												   new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
												   speed_function,
												   new NHA_CurvatureBase(*base, circular_neigh), 
												   new NHA_ResultCreator(*base));

				case 3:

					return new Loc_Reg_NHA_Routine(base, 
												   obj_manager,
												   new Reg_NHA_DomainInitializer3D(*base, mask.begin(), *obj_manager, domain -> begin()),
												   new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
												   speed_function,
												   new NHA_CurvatureBase3D(*base, circular_neigh), 
												   new NHA_ResultCreator3D(*base));

				default: throw InternalException("Unreachable code executed!");
			}
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                      LabGlobChanVeseRoutine                 //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		Routine<int> *LabGlobChanVeseRoutine(const Image3d<Type::Prec> &input, const Image3d<Type::LabVoxel> &mask, const Image3d<Type::BinVoxel> *domain, 
                              			     Type::Size radius, Type::Prec curvature_impact, Type::Prec foreground_impact, Type::Prec background_impact, bool multiple)
		{
			// grid size is given as the size of the input image
			Vector3d<Type::Size> grid_size = input.GetSize();

			// the input image has been constructed in such a way to have zero central differences on its border
			// therefore, its size has grown by two pixels in each dimension
			Vector3d<Type::Size> size = grid_size - Vector3d<Type::Size>(2, 2, grid_size.z > 1 ? 2 : 0);

			// check the size of the mask image as well as the domain one
			if (mask.GetSize() != size || (domain && domain -> GetSize() != size))
			{
				throw InternalException("Incompatible image size!");
			}

			// grid resolution is given as the resolution of the input image
			Vector3d<Type::Prec> grid_res = input.GetResolution().GetRes();

			// base of the labelled global region-based version of the Nilsson-Heyden algorithm		
			Lab_Glob_Reg_NHA_Base *base = new Lab_Glob_Reg_NHA_Base(grid_size.x, grid_size.y, grid_size.z, grid_res.x, grid_res.y, grid_res.z);

			// circular neighbourhood at which the curvature is approximated
			NHA_CircularNeighbourhood *circular_neigh = new NHA_CircularNeighbourhood(*base, radius);

			// manager for efficient computation of average foreground and background object statistics
			ObjectManager *obj_manager = multiple ? new LabObjectManager(input.begin(), base -> m_labels.begin()) : new ObjectManager(input.begin());

			// speed function used in the Chan-Vese model
			ChanVeseSpeedFunction *speed_function = new ChanVeseSpeedFunction(new NHA_MeanCurvatureForce(base -> m_curvature.begin(), circular_neigh -> ReciprocSize()),
                                                							  curvature_impact, obj_manager, foreground_impact, background_impact);

			int num = 0;
			if (size.z > 1)
			{
				++num;
			}

			if (domain)
			{
				num += 2;
			}

			switch (num)
			{
				case 0:

					return new Lab_Glob_Reg_NHA_Routine(base, 
														obj_manager,	
														new Lab_Reg_NHA_Initializer(*base, base -> m_labels, obj_manager, mask.begin()),
														new Lab_NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>, base -> m_labels),
														speed_function, 
														new Lab_NHA_Curvature(*base, circular_neigh, base -> m_labels),  
														new Lab_NHA_ResultCreator(*base, base -> m_labels));

				case 1:

					return new Lab_Glob_Reg_NHA_Routine(base,
													    obj_manager,	
														new Lab_Reg_NHA_Initializer3D(*base, base -> m_labels, obj_manager, mask.begin()),
														new Lab_NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>, base -> m_labels),
														speed_function, 
														new Lab_NHA_Curvature3D(*base, circular_neigh, base -> m_labels),  
														new Lab_NHA_ResultCreator3D(*base, base -> m_labels));

				case 2:

					return new Lab_Glob_Reg_NHA_Routine(base, 
														obj_manager,	 
														new Lab_Reg_NHA_DomainInitializer(*base, base -> m_labels, obj_manager, mask.begin(), domain -> begin()),
														new Lab_NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>, base -> m_labels),
														speed_function, 
														new Lab_NHA_Curvature(*base, circular_neigh, base -> m_labels),  
														new Lab_NHA_ResultCreator(*base, base -> m_labels));

				case 3:

					return new Lab_Glob_Reg_NHA_Routine(base, 
														obj_manager,	
														new Lab_Reg_NHA_DomainInitializer3D(*base, base -> m_labels, obj_manager, mask.begin(), domain -> begin()),
														new Lab_NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>, base -> m_labels),
														speed_function, 
														new Lab_NHA_Curvature3D(*base, circular_neigh, base -> m_labels),  
														new Lab_NHA_ResultCreator3D(*base, base -> m_labels));

				default: throw InternalException("Unreachable code executed!");
			}
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                    GlobRoussonDericheRoutine                //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		Routine<int> *GlobRoussonDericheRoutine(const Image3d<Type::Prec> &input, const Image3d<Type::BinVoxel> &mask, const Image3d<Type::BinVoxel> *domain, 
			                                    bool preserve_topology, Type::Size radius, Type::Prec curvature_impact)
		{
			// grid size is given as the size of the input image
			Vector3d<Type::Size> grid_size = input.GetSize();

			// the input image has been constructed in such a way to have zero central differences on its border
			// therefore, its size has grown by two pixels in each dimension
			Vector3d<Type::Size> size = grid_size - Vector3d<Type::Size>(2, 2, grid_size.z > 1 ? 2 : 0);

			// check the size of the mask image as well as the domain one
			if (mask.GetSize() != size || (domain && domain -> GetSize() != size))
			{
				throw InternalException("Incompatible image size!");
			}

			// grid resolution is given as the resolution of the input image
			Vector3d<Type::Prec> grid_res = input.GetResolution().GetRes();

			// base of the global region-based version of the Nilsson-Heyden algorithm		
			Glob_Reg_NHA_Base *base = new Glob_Reg_NHA_Base(grid_size.x, grid_size.y, grid_size.z, grid_res.x, grid_res.y, grid_res.z);

			// circular neighbourhood at which the curvature is approximated
			NHA_CircularNeighbourhood *circular_neigh = new NHA_CircularNeighbourhood(*base, radius);

			// manager for efficient computation of foreground and background object statistics
			ObjectManager *obj_manager = new ObjectManager(input.begin());

			// speed function used in the Rousson-Deriche model
			RoussonDericheSpeedFunction *speed_function = new RoussonDericheSpeedFunction(new NHA_MeanCurvatureForce(base -> m_curvature.begin(), circular_neigh -> ReciprocSize()), 
																						  curvature_impact, obj_manager);

			int num = 0;
			if (size.z > 1)
			{
				++num;
			}

			if (domain)
			{
				num += 2;
			}

			if (preserve_topology)
			{
				num += 4;
			}

			switch (num)
			{
				case 0:

					return new Glob_Reg_NHA_Routine(base,
													obj_manager, 
													new Reg_NHA_Initializer(*base, mask.begin(), *obj_manager),
													new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
													speed_function,
													new NHA_CurvatureBase(*base, circular_neigh), 
													new NHA_ResultCreator(*base));

				case 1:

					return new Glob_Reg_NHA_Routine(base, 
													obj_manager, 
													new Reg_NHA_Initializer3D(*base, mask.begin(), *obj_manager),
													new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
													speed_function,
													new NHA_CurvatureBase3D(*base, circular_neigh), 
													new NHA_ResultCreator3D(*base));

				case 2:

					return new Glob_Reg_NHA_Routine(base, 
													obj_manager,	
													new Reg_NHA_DomainInitializer(*base, mask.begin(), *obj_manager, domain -> begin()),
													new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
													speed_function,
													new NHA_CurvatureBase(*base, circular_neigh), 
													new NHA_ResultCreator(*base));

				case 3:

					return new Glob_Reg_NHA_Routine(base, 
													obj_manager,
													new Reg_NHA_DomainInitializer3D(*base, mask.begin(), *obj_manager, domain -> begin()),
													new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
													speed_function,
													new NHA_CurvatureBase3D(*base, circular_neigh), 
													new NHA_ResultCreator3D(*base));

				case 4:

					return new TP_Glob_Reg_NHA_Routine(base, 
													   obj_manager,
													   new TP_Reg_NHA_Initializer(*base, mask.begin(), *obj_manager),
													   new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
													   speed_function,
													   new NHA_CurvatureBase(*base, circular_neigh), 
													   new NHA_ResultCreator(*base),
													   new NHA_SimplePointInspector(*base));

				case 5:

					return new TP_Glob_Reg_NHA_Routine(base, 
													   obj_manager,
													   new TP_Reg_NHA_Initializer3D(*base, mask.begin(), *obj_manager),
													   new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
													   speed_function,
													   new NHA_CurvatureBase3D(*base, circular_neigh), 
													   new NHA_ResultCreator3D(*base),
													   new NHA_SimplePointInspector3D(*base));

				case 6:

					return new TP_Glob_Reg_NHA_Routine(base, 
													   obj_manager,
													   new TP_Reg_NHA_DomainInitializer(*base, mask.begin(), *obj_manager, domain -> begin()),
													   new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
													   speed_function,
													   new NHA_CurvatureBase(*base, circular_neigh),  
													   new NHA_ResultCreator(*base),
													   new NHA_SimplePointInspector(*base));

				case 7:

					return new TP_Glob_Reg_NHA_Routine(base,
													   obj_manager,
													   new TP_Reg_NHA_DomainInitializer3D(*base, mask.begin(), *obj_manager, domain -> begin()),
													   new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateGreater<0>),
													   speed_function,
													   new NHA_CurvatureBase3D(*base, circular_neigh), 
													   new NHA_ResultCreator3D(*base),
													   new NHA_SimplePointInspector3D(*base));

				default: throw InternalException("Unreachable code executed!");
			}
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                     LocRoussonDericheRoutine                //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		Routine<int> *LocRoussonDericheRoutine(const Image3d<Type::Prec> &input, const Image3d<Type::BinVoxel> &mask, const Image3d<Type::BinVoxel> *domain, 
											   Type::Size radius, Type::Prec curvature_impact)
		{
			// grid size is given as the size of the input image
			Vector3d<Type::Size> grid_size = input.GetSize();

			// the input image has been constructed in such a way to have zero central differences on its border
			// therefore, its size has grown by two pixels in each dimension
			Vector3d<Type::Size> size = grid_size - Vector3d<Type::Size>(2, 2, grid_size.z > 1 ? 2 : 0);

			// check the size of the mask image as well as the domain one
			if (mask.GetSize() != size || (domain && domain -> GetSize() != size))
			{
				throw InternalException("Incompatible image size!");
			}

			// grid resolution is given as the resolution of the input image
			Vector3d<Type::Prec> grid_res = input.GetResolution().GetRes();

			// base of the local region-based version of the Nilsson-Heyden algorithm		
			Loc_Reg_NHA_Base *base = new Loc_Reg_NHA_Base(grid_size.x, grid_size.y, grid_size.z, grid_res.x, grid_res.y, grid_res.z);

			// circular neighbourhood at which the curvature is approximated
			NHA_CircularNeighbourhood *circular_neigh = new NHA_CircularNeighbourhood(*base, radius);

			// manager for efficient computation of foreground and background object statistics
			ObjectManager *obj_manager = new ObjectManager(input.begin());

			// speed function used in the Rousson-Deriche model
			RoussonDericheSpeedFunction *speed_function = new RoussonDericheSpeedFunction(new NHA_MeanCurvatureForce(base -> m_curvature.begin(), circular_neigh -> ReciprocSize()), 
																						  curvature_impact, obj_manager);

			int num = 0;
			if (size.z > 1)
			{
				++num;
			}

			if (domain)
			{
				num += 2;
			}

			switch (num)
			{
				case 0:

					return new Loc_Reg_NHA_Routine(base,
												   obj_manager, 
												   new Reg_NHA_Initializer(*base, mask.begin(), *obj_manager),
												   new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
												   speed_function,
												   new NHA_CurvatureBase(*base, circular_neigh),  
												   new NHA_ResultCreator(*base));

				case 1:

					return new Loc_Reg_NHA_Routine(base, 
												   obj_manager, 
												   new Reg_NHA_Initializer3D(*base, mask.begin(), *obj_manager),
												   new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
												   speed_function,
												   new NHA_CurvatureBase3D(*base, circular_neigh), 
												   new NHA_ResultCreator3D(*base));

				case 2:

					return new Loc_Reg_NHA_Routine(base, 
												   obj_manager,	
												   new Reg_NHA_DomainInitializer(*base, mask.begin(), *obj_manager, domain -> begin()),
												   new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
												   speed_function,
												   new NHA_CurvatureBase(*base, circular_neigh), 
												   new NHA_ResultCreator(*base));

				case 3:

					return new Loc_Reg_NHA_Routine(base, 
												   obj_manager,
												   new Reg_NHA_DomainInitializer3D(*base, mask.begin(), *obj_manager, domain -> begin()),
												   new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
												   speed_function,
												   new NHA_CurvatureBase3D(*base, circular_neigh), 
												   new NHA_ResultCreator3D(*base));

				default: throw InternalException("Unreachable code executed!");
			}
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                   LabGlobRoussonDericheRoutine              //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		Routine<int> *LabGlobRoussonDericheRoutine(const Image3d<Type::Prec> &input, const Image3d<Type::LabVoxel> &mask, const Image3d<Type::BinVoxel> *domain, 
                              				       Type::Size radius, Type::Prec curvature_impact, bool multiple)
		{
			// grid size is given as the size of the input image
			Vector3d<Type::Size> grid_size = input.GetSize();

			// the input image has been constructed in such a way to have zero central differences on its border
			// therefore, its size has grown by two pixels in each dimension
			Vector3d<Type::Size> size = grid_size - Vector3d<Type::Size>(2, 2, grid_size.z > 1 ? 2 : 0);

			// check the size of the mask image as well as the domain one
			if (mask.GetSize() != size || (domain && domain -> GetSize() != size))
			{
				throw InternalException("Incompatible image size!");
			}

			// grid resolution is given as the resolution of the input image
			Vector3d<Type::Prec> grid_res = input.GetResolution().GetRes();

			// base of the labelled global region-based version of the Nilsson-Heyden algorithm		
			Lab_Glob_Reg_NHA_Base *base = new Lab_Glob_Reg_NHA_Base(grid_size.x, grid_size.y, grid_size.z, grid_res.x, grid_res.y, grid_res.z);

			// circular neighbourhood at which the curvature is approximated
			NHA_CircularNeighbourhood *circular_neigh = new NHA_CircularNeighbourhood(*base, radius);

			// manager for efficient computation of average foreground and background object statistics
			ObjectManager *obj_manager = multiple ? new LabObjectManager(input.begin(), base -> m_labels.begin()) : new ObjectManager(input.begin());

			// speed function used in the Chan-Vese model
			RoussonDericheSpeedFunction *speed_function = new RoussonDericheSpeedFunction(new NHA_MeanCurvatureForce(base -> m_curvature.begin(), circular_neigh -> ReciprocSize()),
                                                										  curvature_impact, obj_manager);

			int num = 0;
			if (size.z > 1)
			{
				++num;
			}

			if (domain)
			{
				num += 2;
			}

			switch (num)
			{
				case 0:

					return new Lab_Glob_Reg_NHA_Routine(base, 
														obj_manager,	
														new Lab_Reg_NHA_Initializer(*base, base -> m_labels, obj_manager, mask.begin()),
														new Lab_NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>, base -> m_labels),
														speed_function, 
														new Lab_NHA_Curvature(*base, circular_neigh, base -> m_labels),  
														new Lab_NHA_ResultCreator(*base, base -> m_labels));

				case 1:

					return new Lab_Glob_Reg_NHA_Routine(base,
										 				obj_manager,	
														new Lab_Reg_NHA_Initializer3D(*base, base -> m_labels, obj_manager, mask.begin()),
														new Lab_NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>, base -> m_labels),
														speed_function, 
														new Lab_NHA_Curvature3D(*base, circular_neigh, base -> m_labels),  
														new Lab_NHA_ResultCreator3D(*base, base -> m_labels));

				case 2:

					return new Lab_Glob_Reg_NHA_Routine(base, 
														obj_manager,	 
														new Lab_Reg_NHA_DomainInitializer(*base, base -> m_labels, obj_manager, mask.begin(), domain -> begin()),
														new Lab_NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>, base -> m_labels),
														speed_function, 
														new Lab_NHA_Curvature(*base, circular_neigh, base -> m_labels),  
														new Lab_NHA_ResultCreator(*base, base -> m_labels));

				case 3:

					return new Lab_Glob_Reg_NHA_Routine(base, 
														obj_manager,	
														new Lab_Reg_NHA_DomainInitializer3D(*base, base -> m_labels, obj_manager, mask.begin(), domain -> begin()),
														new Lab_NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>, base -> m_labels),
														speed_function, 
														new Lab_NHA_Curvature3D(*base, circular_neigh, base -> m_labels),  
														new Lab_NHA_ResultCreator3D(*base, base -> m_labels));

				default: throw InternalException("Unreachable code executed!");
			}
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                           MCMRoutine                        //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		Routine<int> *MCMRoutine(const Image3d<Type::BinVoxel> &input, const Image3d<Type::BinVoxel> *domain, Type::Size radius)
		{
			// check dimensionality
			bool is_3d = (input.GetSizeZ() > 1);

			// check the size of the domain image
			if (domain && domain -> GetSize() != input.GetSize())
			{
				throw InternalException("Incompatible image size!");
			}

			// grid size is given as the size of the input image expanded by 1 at each direction
			Vector3d<Type::Size> grid_size = input.GetSize() + Vector3d<Type::Size>(2, 2, is_3d ? 2 : 0);

			// grid resolution is given as the resolution of the input image
			Vector3d<Type::Prec> grid_res = input.GetResolution().GetRes();

			// base of the Nilsson-Heyden algorithm
			NHA_Base *base = new NHA_Base(grid_size.x, grid_size.y, grid_size.z, grid_res.x, grid_res.y, grid_res.z);

			// circular neighbourhood at which the curvature is approximated
			NHA_CircularNeighbourhood *circular_neigh = new NHA_CircularNeighbourhood(*base, radius);

			// speed function used in the Nilsson-Heyden algorithm
			MeanCurvatureSpeedFunction *speed_function = new MeanCurvatureSpeedFunction(new NHA_MeanCurvatureForce(base -> m_curvature.begin(), circular_neigh -> ReciprocSize()), Type::Prec(1));

			int num = 0;
			if (is_3d)
			{
				++num;
			}

			if (domain)
			{
				num += 2;
			}

			switch (num)
			{
				case 0:
					
					return new NHA_Routine(base, 
										   new NHA_Initializer(*base, input.begin()),
										   new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
										   speed_function, 
										   new NHA_Curvature(*base, circular_neigh, speed_function, base -> m_heap),  
										   new NHA_ResultCreator(*base));
			
				case 1:
					
					return new NHA_Routine(base, 
										   new NHA_Initializer3D(*base, input.begin()),
										   new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
										   speed_function,
										   new NHA_Curvature3D(*base, circular_neigh, speed_function, base -> m_heap), 
										   new NHA_ResultCreator3D(*base));

				case 2:
					
					return new NHA_Routine(base, 
										   new NHA_DomainInitializer(*base, input.begin(), domain -> begin()),
										   new NHA_Neighbourhood(*base, new NeighbourIndices2D_4(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
										   speed_function, 
										   new NHA_Curvature(*base, circular_neigh, speed_function, base -> m_heap),  
										   new NHA_ResultCreator(*base));
			
				case 3:
					
					return new NHA_Routine(base, 
										   new NHA_DomainInitializer3D(*base, input.begin(), domain -> begin()),
										   new NHA_Neighbourhood(*base, new NeighbourIndices3D_6(*base), new NHA_PredicateEqual<-1>, new NHA_PredicateEqual<1>),
										   speed_function,
										   new NHA_Curvature3D(*base, circular_neigh, speed_function, base -> m_heap), 
										   new NHA_ResultCreator3D(*base));

				default: throw InternalException("Unreachable code executed!");
			}
		}

		/**********************************************************************/

	} // namespace fls
	
} // namespace i3d

