/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2015   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Freere
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/** 
 * \file lbp.cc
   \brief Local Binary Patterns
	
	This file contains functions for calculating Local Binary Patterns.

	T. Ojala, M. Pietikinen, and T. Maenpaa. Multiresolution gray-scale and
	rotation invariant texture classification with local binary patterns. IEEE
	Transactions on Pattern Analysis and Machine Intelligence, 24(7):971987,
	July 2002.
	
	\author Tomas Majtner <172786@mail.muni.cz>
	\date 2015
*/

#include "image3d.h"

using namespace std;
using namespace i3d;


template <class T> void ComputeLBP(Image3d<T> &img , int P, int R, int RADIUS_STEP, int NUMBER_OF_R)
{	
	vector<double> feature_vector;
	int hist_len = (int)pow(2.0, P);  // length of the histogram
	int *result = new int[hist_len];  // for storing LBP pattern
	int sum;	// sum for normalization

	for (int rad = R; rad < R+(RADIUS_STEP*NUMBER_OF_R); rad+=RADIUS_STEP)
	{
		// clear the histogram
		for (int i = 0; i < hist_len; i++) result[i] = 0;

		for (unsigned y = rad; y < img.GetSizeY() - rad; y++)
		{
			for (unsigned x = rad; x < img.GetSizeX() - rad; x++)
			{
				int pattern = 0;
				for (int p = 0; p < P; p++)
				{
					// finding x and y coordinates of point in the neighbourhood
					double xp = x + (((rad*cos(2*M_PI*p/P) > -0.05) && (rad*cos(2*M_PI*p/P) < 0.05)) ? 0 : rad*cos(2*M_PI*p/P));
					double yp = y - (((rad*sin(2*M_PI*p/P) > -0.05) && (rad*sin(2*M_PI*p/P) < 0.05)) ? 0 : rad*sin(2*M_PI*p/P));
		
					// if we are right in the centre of the pixel, we can take its value
					if ((xp == (int)xp) || (yp == (int)yp))
					{
						pattern += ((img.GetVoxel((int)(xp+0.1), (int)(yp+0.1), 0) < img.GetVoxel(x, y, 0)) ? 0 : 1) << p;
					}										
					else // if we are not, we need to use bilinear interpolation to get the value
					{
						int x1 = (int)floor(xp), x2 = (int)ceil(xp), y1 = (int)floor(yp), y2 = (int)ceil(yp);
						pattern += (((img.GetVoxel(x1,y1,0)*(x2-xp)*(y2-yp) +
									img.GetVoxel(x2,y1,0)*(xp-x1)*(y2-yp) +
									img.GetVoxel(x1,y2,0)*(x2-xp)*(yp-y1) +
									img.GetVoxel(x2,y2,0)*(xp-x1)*(yp-y1)) < img.GetVoxel(x, y, 0)) ? 0 : 1) << p;
					}				
				}
				result[pattern]++;
			}
		}
		sum = 0;
		for (int i = 0; i<hist_len; i++) sum += result[i];  // sum for normalization
		for (int i = 0; i<hist_len; i++) feature_vector.push_back((double)(result[i])/sum);
	}
	delete [] result;

	// printing feature vector
	for (unsigned i = 0; i < feature_vector.size()-1; i++) {
		cout << feature_vector[i] << ","; 
	}
	cout << feature_vector[feature_vector.size()-1];
}