/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2015   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Freere
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/** 
 * \file moments.cc
   \brief Central moments
	
	This file contains functions for calculating Central moments.

	J. Flusser, B. Zitov, and T. Suk. Moments and Moment Invariants in Pattern
	Recognition. Wiley Publishing, 2009.
	
	\author Tomas Majtner <172786@mail.muni.cz>
	\date 2015
*/

#include "image3d.h"
#include "transform.h"
#include <math.h>

using namespace std;
using namespace i3d;


template <class T> void ComputeMoments(Image3d<T> &img)
{
	// *** Resample To Desired Resolution ***
	Vector3d<float> res = img.GetResolution().GetRes();
	 if ((res.x != res.y) || (res.x != res.z))
	 {
		  float max_res = max(max(res.x, res.y), res.z);
		  Vector3d<float> new_res = max_res;
		  ResampleToDesiredResolution(img, new_res, LANCZOS);
	 }

	int Nx = img.GetSizeX();
	int Ny = img.GetSizeY();
	int Nz = img.GetSizeZ();
	int min_val = img.GetMaxValue();

	// finding minimal intensity value in the image
	for (int k = 0; k < Nz; k++) {
		for (int j = 0; j < Ny; j++) {
			for (int i = 0; i < Nx; i++) {
				if ((img.GetVoxel(i,j,k) > 0) && (img.GetVoxel(i,j,k) < min_val)) min_val = img.GetVoxel(i,j,k);				
			}
		}
	}	

	double average = 0;
	int sum = 0;
	double u2 = 0;  // 2nd central moment
	double u3 = 0;  // 3rd central moment
	double u4 = 0;  // 4th central moment
	double range = img.GetMaxValue() - min_val + 1;  // range of intensity values
	vector<int> values((int)floor(range));

	// finding average intensity value of the image
	for (int k = 0; k < Nz; k++) {
		for (int j = 0; j < Ny; j++) {
			for (int i = 0; i < Nx; i++) {
				if (img.GetVoxel(i,j,k) > 0)
				{
					values[(int) (img.GetVoxel(i,j,k) - min_val)]++;
					average += (double) (img.GetVoxel(i,j,k) - min_val) / range;
					sum++;
				}
			}
		}
	}
	average = average / sum;

	// deriving and print 2nd, 3rd, and 4th central moment
	for (unsigned i = 0; i < values.size(); i++)
	{
		u2 += pow(((double)i / range - average), 2.0) * values[i] / sum;
		u3 += pow(((double)i / range - average), 3.0) * values[i] / sum;
		u4 += pow(((double)i / range - average), 4.0) * values[i] / sum;
	}
	cout << u2 << ',' << u3 << ',' << u4 << endl;
}