/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2015   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Freere
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/** 
 * \file tamura.cc
   \brief 2D & 3D Tamura features.
	
	This file contains functions for calculating 2D/3D Tamura features.

	H. Tamura, S. Mori, and T. Yamawaki. Textural features corresponding to
	visual perception. IEEE Transactions on Systems, Man and Cybernetics,
	8(6):460473, 1978.
	
	T. Majtner, D. Svoboda. Extension of Tamura Texture Features for 3D Fluorescence
	Microscopy. In 2nd Joint 3DIM/3DPVT Conference: 3D Imaging,
	Modeling, Processing, Visualization and Transmission (3DimPVT). Los Alamitos,
	California, USA: IEEE Computer Society, 2012. Pages 301307, ISBN
	978-0-7695-4873-9.
	
	\author Tomas Majtner <172786@mail.muni.cz>
	\date 2015
*/

#include <stdlib.h>
#include <vector>
#include <string>
#include <fstream>

#include "convolution.h"
#include "transform.h"
#include "math.h"

using namespace std;
using namespace i3d;

// sum the neighbours in 2D image
template <class T> double Neighbours2D (const Image3d<T> &img, int from_y, int to_y, int from_x, int to_x, int min_val, int normal)
{
	double result = 0;
	for (int y = from_y; y < to_y; y++) {
		for (int x = from_x; x < to_x; x++) 
		{
			result += (double) (img.GetVoxel(x,y,0) - min_val) / normal;
		}
	}
	return result;
}

// sum the neighbours in 3D image
template <class T> double Neighbours3D (const Image3d<T> &img, int from_z, int to_z, int from_y, int to_y, int from_x, int to_x, int min_val, int normal)
{
	double result = 0;
	for (int z = from_z; z < to_z; z++) {
		for (int y = from_y; y < to_y; y++) {
			for (int x = from_x; x < to_x; x++) 
			{
				result += (double) (img.GetVoxel(x,y,z) - min_val) / normal;
			}
		}
	}
	return result;
}

// *****************************************************************
//
//                      Tamura features in 2D
//
// *****************************************************************

template <class T> void ComputeTamura2D (const Image3d<T> &img)
{
	#ifdef I3D_DEBUG
	cout << "Computing Tamura features in 2D." << '\n';
	#endif	

	// getting size of an image
	int Nx = img.GetSizeX();
	int Ny = img.GetSizeY();


// *****************************************************************
// 2D Coarseness
// *****************************************************************	

	// 2D vectors for computation
	// Ex - horizontal direction, Ey - vertical direction
	vector<vector<double> > Ex, Ey;
	vector<vector<int>> sBest;

	// resizing vectors to appropriate size
	sBest.resize(Nx); Ex.resize(Nx); Ey.resize(Nx);
	for (int i = 0; i < Nx; i++) {
		sBest[i].resize(Ny); Ex[i].resize(Ny); Ey[i].resize(Ny);
	}

	int min_val = img.GetMinValue();
	int normal = img.GetMaxValue() - min_val;

	// *** Neighbourhood 1x1 ***
	// sBest is 2 for every point, Eh and Ev calculated according formula
	sBest[0][Ny-1] = 2;

	for (int j = 0; j < Ny; j++) {
		for (int i = 1; i < Nx; i++) {
			Ex[i][j] = (double) (((img.GetVoxel(i,j,0)-min_val)/normal) - ((img.GetVoxel(i-1,j,0)-min_val)/normal))/2;
			sBest[i][j] = 2;
		}
	}

	for (int j = 0; j < Ny-1; j++) {
		for (int i = 0; i < Nx; i++) {
			Ey[i][j] = (double) (((img.GetVoxel(i,j,0)-min_val)/normal) - ((img.GetVoxel(i,j+1,0)-min_val)/normal))/2;
			sBest[i][j] = 2;
		}
	}

	for (int qq = 0; qq < 5; qq++)
	{
		if (Nx >= pow(2.0,(qq+2)) && Ny >= pow(2.0,(qq+2)))
		{
			for (int j = (int)pow(2.0, qq) - 1; j < Ny - pow(2.0, qq); j++) {
				for (int i = (int)pow(2.0, (qq+1)); i < Nx - pow(2.0, (qq+1))+1; i++) {
										
					// calculating left operand in x case
					double A_left = Neighbours2D(img, j-((int)pow(2.0,qq)-1), j+((int)pow(2.0,qq)+1), 
						i, i+(int)pow(2.0,(qq+1)), min_val, normal);
									
					// calculating right operand in x case
					double A_right = Neighbours2D(img, j-((int)pow(2.0,qq)-1), j+((int)pow(2.0,qq)+1), 
						i-(int)pow(2.0,(qq+1)), ((qq==4) ? i-30 : i), min_val, normal);
					
					double A_x_y = abs((A_left - A_right) / pow(2.0,((qq+1)*2)));

					// if grater than previous value, update variables
					if (A_x_y > max(Ex[i][j], Ey[i][j]))
					{					
						Ex[i][j] = A_x_y;
						sBest[i][j] = (int)pow(2.0,(qq+2));
					}
				}
			}

			for (int j = (int)pow(2.0,(qq+1))-1; j < Ny-pow(2.0,(qq+1)); j++) {
				for (int i = (int)pow(2.0,qq); i < Nx-pow(2.0,qq)+1; i++) {
					
					// calculating left operand in y case
					double A_left = Neighbours2D(img, j-((int)pow(2.0,qq+1)-1), j+1, 
						i-(int)pow(2.0,qq), i+(int)pow(2.0,qq), min_val, normal);
					
					// calculating right operand in y case
					double A_right = Neighbours2D(img, j+1, j+(int)pow(2.0,(qq+1))+1, 
						i-(int)pow(2.0,qq), i+(int)pow(2.0,qq), min_val, normal);

					double A_x_y = (double) abs((A_left - A_right)/pow(2.0,((qq+1)*2)));

					// if grater than previous value, update variables
					if (A_x_y > max(Ex[i][j], Ey[i][j]))
					{					
						Ey[i][j] = A_x_y;
						sBest[i][j] = (int)pow(2.0,(qq+2));
					}
				}
			}
		}
	}
	
	// result is sum of all sBest-s divided by number of pixels
	double result = 0;
	for (int j = 0; j < Ny; j++) {
		for (int i = 0; i < Nx; i++) {
			result += sBest[i][j];
		}
	}
	cout << result / (Nx * Ny) << ",";

// *****************************************************************
// 2D Contrast
// *****************************************************************

	// important variables
	double average = 0;
	double u2 = 0;  // 2nd central moment
	double u4 = 0;  // 4th central moment
	int range = img.GetMaxValue() - img.GetMinValue() + 1;  // range of values
	double min = img.GetMinValue();  // minimal value
	vector<int> values(range);

	// counting frequencies of intensities and average value
	for (int j = 0; j < Ny; j++) {
		for (int i = 0; i < Nx; i++) {
			values[(int) (img.GetVoxel(i,j,0) - min)]++;
			average += (double) (img.GetVoxel(i,j,0) - min) / range;
		}
	}
	average = average / (Nx*Ny);

	// calculation 2nd and 4th central moment
	for (unsigned i = 0; i < values.size(); i++)
	{
		u2 += pow((i / (double)range - average), 2.0) * values[i] / (Nx*Ny);
		u4 += pow((i / (double)range - average), 4.0) * values[i] / (Nx*Ny);
	}

	// contrast is calculated from the formula
	cout << sqrt(u2) / pow((u4 / pow(u2, 2.0)), 0.25) << ",";


// *****************************************************************
// 2D Direction
// *****************************************************************

	// important 2D vectors
	vector<vector<double> > deltaH (Nx, vector<double> (Ny));  // for convolution with horizontal kernel
	vector<vector<double> > deltaV (Nx, vector<double> (Ny));  // for convolution with vertical kernel
	vector<double> theta (Nx*Ny);
	vector<vector<int> > kernelH (3, vector<int> (3));   // horizontal kernel
	vector<vector<int> > kernelV (3, vector<int> (3));   // vertical kernel

																 // horizontal kernel
	kernelH[0][0] = -1; kernelH[0][1] = -1; kernelH[0][2] = -1;  // -1 -1 -1
	kernelH[1][0] = 0; kernelH[1][1] = 0; kernelH[1][2] = 0;     //  0  0  0
	kernelH[2][0] = 1; kernelH[2][1] = 1; kernelH[2][2] = 1;     //  1  1  1

																 // vertical kernel
	kernelV[0][0] = -1; kernelV[0][1] = 0; kernelV[0][2] = 1;	 // -1  0  1
	kernelV[1][0] = -1; kernelV[1][1] = 0; kernelV[1][2] = 1;    // -1  0  1
	kernelV[2][0] = -1; kernelV[2][1] = 0; kernelV[2][2] = 1;    // -1  0  1

	// calculating convolution to deltaH and deltaV
	int mm, nn, ii, jj;
	for (int i = 0; i < Nx; ++i)
	{
		for (int j = 0; j < Ny; ++j)
		{
			for (int m = 0; m < 3; ++m)
			{
				mm = 3 - 1 - m;
				for (int n = 0; n < 3; ++n)
				{
					nn = 3 - 1 - n;
					ii = i + (m - 1);
					jj = j + (n - 1);

					if (ii >= 0 && ii < Nx && jj >= 0 && jj < Ny)
					{
						deltaH[i][j] += (double)img.GetVoxel(ii,jj,0) * kernelH[mm][nn];
						deltaV[i][j] += (double)img.GetVoxel(ii,jj,0) * kernelV[mm][nn];
					}
				}
			}
		}
	}

	// calculating theta from deltaH and deltaV according to formula
	int k = 0;
	for (int i = 0; i < Nx; ++i)
	{
		for (int j = 0; j < Ny; ++j)
		{
			if (deltaH[i][j] == 0) theta[k] = M_PI;  // avoiding division by zero
			else theta[k] = (M_PI/2 + atan(deltaV[i][j] / deltaH[i][j]));
			k++;
		}
	}

	// vector of direction values, 314196 = M_PI * 100000, to be able to use for cycle with integer 
	vector<int> direction_values(314196);

	// counting frequencies of theta values
	for (int i = 0; i < Nx*Ny; i++) direction_values[(int) (theta[i]*100000)]++;

	// finding max value in direction values
	int max_index = 0;
	int max_value = 0;
	for (int i = 0; i < 314196; i++) {
		if (direction_values[i] > max_value) {
			max_value = direction_values[i];
			max_index = i;
		}
	}

	// calculating direction according to formula
	double direction = 0;
	for (int i = 0; i < 314196; i++)
	{
		if ((double) (direction_values[i]) / (Nx*Ny) > 0.01) {
		direction += pow(((double)(i)/100000 - (double) (max_index)/100000), 2.0) * 
			(double) (direction_values[i]) / (Nx*Ny);
		}
	}
	cout << direction << endl;
}

// *****************************************************************
//
//                      Tamura features in 3D
//
// *****************************************************************
 
template <class T> void ComputeTamura3D (const Image3d<T> &img_orig)
{
	#ifdef I3D_DEBUG
	cout << "Computing Tamura features in 3D." << '\n';
	#endif	

	Image3d<T> img = img_orig;

	// *** Resample To Desired Resolution ***
	Vector3d<float> res = img.GetResolution().GetRes();

	 if ((res.x != res.y) || (res.x != res.z))
	 {
		  float max_res = max(max(res.x, res.y), res.z);
		  Vector3d<float> new_res = max_res;
		  ResampleToDesiredResolution(img, new_res, LANCZOS);
	 }

	// getting size of an image
	int Nx = img.GetSizeX();
	int Ny = img.GetSizeY();
	int Nz = img.GetSizeZ();


// *****************************************************************
// 3D Coarseness
// *****************************************************************

	typedef std::vector<double> Double1D; typedef std::vector<short int> Int1D;
	typedef std::vector<Double1D> Double2D; typedef std::vector<Int1D> Int2D;
	typedef std::vector<Double2D> Double3D; typedef std::vector<Int2D> Int3D;

	// important 3D vectors
	Int3D sBest(Nx,Int2D(Ny,Int1D(Nz)));
	Double3D Ex(Nx,Double2D(Ny,Double1D(Nz)));
	Double3D Ey(Nx,Double2D(Ny,Double1D(Nz)));
	Double3D Ez(Nx,Double2D(Ny,Double1D(Nz)));

	// resizing vectors to appropriate size
	sBest.resize(Nx); Ex.resize(Nx); Ey.resize(Nx); Ez.resize(Nx);
	for (int i = 0; i < Nx; i++) {
		sBest[i].resize(Ny); Ex[i].resize(Ny); Ey[i].resize(Ny); Ez[i].resize(Ny);
	}

	for (int j = 0; j < Ny; j++)
	{
		for (int i = 0; i < Nx; i++) {
			sBest[i][j].resize(Nz); Ex[i][j].resize(Nz); Ey[i][j].resize(Nz); Ez[i][j].resize(Nz);
		}
	}

	int min_val = img.GetMaxValue();
	for (int k = 0; k < Nz; k++) {
		for (int j = 0; j < Ny; j++) {
			for (int i = 0; i < Nx; i++) {
				if ((img.GetVoxel(i,j,k) > 0) && (img.GetVoxel(i,j,k) < min_val)) min_val = img.GetVoxel(i,j,k);
			}
		}
	}	
	int normal = img.GetMaxValue() - min_val;

	//*** Neighbourhood 1x1x1 ***
	sBest[0][Ny-1][Nz-1] = 2;
	for (int k = 0; k < Nz; k++) {
		for (int j = 0; j < Ny; j++) {
			for (int i = 1; i < Nx; i++) {
				Ex[i][j][k] = (double) (((img.GetVoxel(i,j,k)-min_val)/normal) - ((img.GetVoxel(i-1,j,k)-min_val)/normal))/2;
				sBest[i][j][k] = 2;
			}
		}
	}

	for (int k = 0; k < Nz; k++) {
		for (int j = 0; j < Ny-1; j++) {
			for (int i = 0; i < Nx; i++) {
				Ey[i][j][k] = (double) (((img.GetVoxel(i,j,k)-min_val)/normal) - ((img.GetVoxel(i,j+1,k)-min_val)/normal))/2;
				sBest[i][j][k] = 2;
			}
		}
	}

	for (int k = 0; k < Nz-1; k++) {
		for (int j = 0; j < Ny; j++) {
			for (int i = 0; i < Nx; i++) {
				Ez[i][j][k] = (double) (((img.GetVoxel(i,j,k)-min_val)/normal) - ((img.GetVoxel(i,j,k+1)-min_val)/normal))/2;
				sBest[i][j][k] = 2;
			}
		}
	}
	
	for (int qq = 0; qq < 1; qq++)
	{
		if (Nx >= pow(2.0,(qq+2)) && Ny >= pow(2.0,(qq+2)) && Nz >= pow(2.0,(qq+2)))
		{
			for (int k = (int)pow(2.0,qq)-1; k < Nz-pow(2.0,qq); k++) {
				for (int j = (int)pow(2.0,qq)-1; j < Ny-pow(2.0,qq); j++) {
					for (int i = (int)pow(2.0,(qq+1)); i < Nx-pow(2.0,(qq+1))+1; i++) {
						// calculating left operand in x case
						double A_left = Neighbours3D(img, k-((int)pow(2.0,qq)-1), k+((int)pow(2.0,qq)+1), 
							j-((int)pow(2.0,qq)-1), j+((int)pow(2.0,qq)+1), i, i+(int)pow(2.0,(qq+1)), min_val, normal);
										
						// calculating right operand in x case
						double A_right = Neighbours3D(img, k-((int)pow(2.0,qq)-1), k+((int)pow(2.0,qq)+1), 
							j-((int)pow(2.0,qq)-1), j+((int)pow(2.0,qq)+1), i-(int)pow(2.0,(qq+1)), 
							((qq==4) ? i-30 : i), min_val, normal);
						
						double A_x_y_z = abs((A_left - A_right)/pow(2.0,((qq+1)*2)));

						// if grater than previous value, update variables
						if (A_x_y_z > max(max(Ex[i][j][k], Ey[i][j][k]), Ez[i][j][k]))
						{					
							Ex[i][j][k] = A_x_y_z;
							sBest[i][j][k] = (int)pow(2.0,(qq+2));
						}
					}
				}
			}

			for (int k = (int)pow(2.0,qq); k < Nz-pow(2.0,qq)+1; k++) {
				for (int j = (int)pow(2.0,(qq+1))-1; j < Ny-pow(2.0,(qq+1)); j++) {
					for (int i = (int)pow(2.0,qq); i < Nx-pow(2.0,qq)+1; i++) {					
						// calculating left operand in y case
						double A_left = Neighbours3D(img, k-(int)pow(2.0,qq), k+(int)pow(2.0,qq), 
							j-((int)pow(2.0,qq+1)-1), j+1, i-(int)pow(2.0,qq), i+(int)pow(2.0,qq), min_val, normal);
						
						// calculating right operand in y case
						double A_right = Neighbours3D(img, k-(int)pow(2.0,qq), k+(int)pow(2.0,qq), 
							j+1, j+(int)pow(2.0,(qq+1))+1, i-(int)pow(2.0,qq), i+(int)pow(2.0,qq), min_val, normal);

						double A_x_y_z = (double) abs((A_left - A_right)/pow(2.0,((qq+1)*2)));

						// if grater than previous value, update variables
						if (A_x_y_z > max(max(Ex[i][j][k], Ey[i][j][k]), Ez[i][j][k]))
						{					
							Ey[i][j][k] = A_x_y_z;
							sBest[i][j][k] = (int)pow(2.0,(qq+2));
						}
					}
				}
			}	

			for (int k = (int)pow(2.0,(qq+1))-1; k < Nz-pow(2.0,(qq+1)); k++) {
				for (int j = (int)pow(2.0,qq); j < Ny-pow(2.0,qq)+1; j++) {
					for (int i = (int)pow(2.0,qq); i < Nx-pow(2.0,qq)+1; i++) {
						// calculating left operand in z case
						double A_left = Neighbours3D(img, k-((int)pow(2.0,qq+1)-1), k+1, 
							j-(int)pow(2.0,qq), j+(int)pow(2.0,qq), i-(int)pow(2.0,qq), i+(int)pow(2.0,qq), min_val, normal);
						
						// calculating right operand in z case
						double A_right = Neighbours3D(img, k+1, k+(int)pow(2.0,(qq+1))+1, 
							j-(int)pow(2.0,qq), j+(int)pow(2.0,qq), i-(int)pow(2.0,qq), i+(int)pow(2.0,qq), min_val, normal);

						double A_x_y_z = (double) abs((A_left - A_right)/pow(2.0,((qq+1)*2)));

						// if grater than previous value, update variables
						if (A_x_y_z > max(max(Ex[i][j][k], Ey[i][j][k]), Ez[i][j][k]))
						{					
							Ez[i][j][k] = A_x_y_z;
							sBest[i][j][k] = (int)pow(2.0,(qq+2));
						}
					}
				}
			}
		}
	}

	// result is sum of all sBest-s divided by number of pixels
	double result = 0; 
	int sum = 0;
	for (int k = 0; k < Nz; k++) {
		for (int j = 0; j < Ny; j++) {
			for (int i = 0; i < Nx; i++) {
				result += sBest[i][j][k];
				sum++;
			}
		}
	}	
	cout << result / sum << ',';


// *****************************************************************
// 3D Contrast
// *****************************************************************

	// important variables
	double average = 0;
	double u2 = 0;  // 2nd central moment
	double u4 = 0;  // 4th central moment
	int range = img.GetMaxValue() - min_val + 1;  // range of values
	vector<int> values(range);

	// counting frequencies of intensities and average value
	for (int k = 0; k < Nz; k++) {
		for (int j = 0; j < Ny; j++) {
			for (int i = 0; i < Nx; i++) {
				if (img.GetVoxel(i,j,k) > 0)
				{
					values[(int) (img.GetVoxel(i,j,k) - min_val)]++;
					average += (double) (img.GetVoxel(i,j,k) - min_val) / range;
				}
			}
		}
	}
	average = average / sum;

	// calculation 2nd and 4th central moment
	for (unsigned i = 0; i < values.size(); i++)
	{
		u2 += pow((i / (double)range - average), 2.0) * values[i] / sum;
		u4 += pow((i / (double)range - average), 4.0) * values[i] / sum;
	}

	// contrast is calculated from the formula
	cout << sqrt(u2) / pow((u4 / pow(u2, 2.0)), 0.25) << ',';


// *****************************************************************
// 3D Direction
// *****************************************************************

	// important 3D vectors
	Double3D deltaX(Nx,Double2D(Ny,Double1D(Nz)));  // for convolution with X kernel
	Double3D deltaZ(Nx,Double2D(Ny,Double1D(Nz)));  // for convolution with Z kernel
	vector<double> theta (Nx*Ny*Nz);
	Double3D kernelX(3,Double2D(3,Double1D(3)));  // X kernel
	Double3D kernelZ(3,Double2D(3,Double1D(3)));  // Z kernel

	// creating 3D kernels
	for (int j = 0; j < 3; j++)
	{
		for (int i = 0; i < 3; i++)
		{
			kernelX[0][i][j] = -1; kernelX[1][i][j] = 0; kernelX[2][i][j] = 1;
			kernelZ[i][j][0] = -1; kernelZ[i][j][1] = 0; kernelZ[i][j][2] = 1;
		}
	}

	// calculating convolution to deltaX and deltaY and deltaZ
	int mm, nn, oo, ii, jj, kk;
	for (int i = 0; i < Nx; ++i)
	{
		for (int j = 0; j < Ny; ++j)
		{
			for (int k = 0; k < Nz; ++k)
			{
				for (int m = 0; m < 3; ++m)
				{
					mm = 3 - 1 - m;
					for (int n = 0; n < 3; ++n)
					{
						nn = 3 - 1 - n;
						for (int o = 0; o < 3; ++o)
						{
							oo = 3 - 1 - o;
							ii = i + (m - 1);
							jj = j + (n - 1);
							kk = k + (o - 1);

							if (ii >= 0 && ii < Nx && jj >= 0 && jj < Ny && kk >= 0 && kk < Nz)
							{
								deltaX[i][j][k] += (double)img.GetVoxel(ii,jj,kk) * kernelX[mm][nn][oo];
								deltaZ[i][j][k] += (double)img.GetVoxel(ii,jj,kk) * kernelZ[mm][nn][oo];
							}
						}
					}
				}
			}
		}
	}

	#ifdef I3D_DEBUG
	cout << "Computing Tamura features in 3D." << '\n';
		
	Image3d<float> obr;
    obr.MakeRoom(Nx, Ny, Nz);
	for (int i = 0; i < Nx; ++i)
	{
		for (int j = 0; j < Ny; ++j)
		{
			for (int k = 0; k < Nz; ++k)
			{
				obr.SetVoxel(i, j, k, deltaZ[i][j][k]);
			}
		}
	}
	obr.SaveImage("_deltaZ.ics");
	#endif

	// calculating theta and phi from deltaX and deltaY and deltaZ according to formula
	int h = 0;
	for (int i = 0; i < Nx; ++i) {
		for (int j = 0; j < Ny; ++j) {
			for (int k = 0; k < Nz; ++k) {
				if (deltaX[i][j][k] == 0) theta[h] = M_PI;    // avoiding division by zero				 
				else theta[h] = (M_PI/2 + atan(deltaZ[i][j][k] / deltaX[i][j][k]));

				h++;
			}
		}
	}

	//vector of direction values, 314196 = M_PI * 100000, to be able to use for cycle with integer 
	vector<int> direction_values(314196);

	// counting frequencies of theta values
	for (int i = 0; i < Nx*Ny*Nz; i++) direction_values[(int) (theta[i]*100000)]++;
	
	// calculating direction according to formula
	double direction = 0;
	for (int i = 0; i < 314196; i++)
	{
		if (((double) (direction_values[i]) / sum > 0.005))
		{
			direction += pow((double)(i)/100000 - M_PI, 2.0) * 
				(((double) (direction_values[i]) / sum));
		}
	}
	cout << direction << endl;
}