/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file draw.cc

  \brief Basic drawing, filling and rasterization algorithms - implementation

  \author Pavel Matula (pam@fi.muni.cz) 2002
  \author David Svoboda (xsvobod2@fi.muni.cz) 2002 - FloodFill functions
  \author Petr Matula (pem@fi.muni.cz) 2004 - DrawCube
  
*/

#ifdef __GNUG__
#	pragma implementation
#endif

#include <iostream>
#include <algorithm>
#include <math.h>
#include <stack>
#include "neighbours.h"
#include "draw.h"

#ifdef _WIN32
#	define rint(x) (floor((x)+0.5))
#endif

using namespace std;
using std::abs;

namespace i3d {
	/*****************************************************************************\
	*
	*                            Local functions
	*
	\******************************************************************************/
	//--------------------------------------------------------------------------
	// PrepareLine
	//--------------------------------------------------------------------------
	size_t PrepareLine(Vector3d<size_t> p1, Vector3d <size_t> p2, Vector3d<double> &d)
	{
		Vector3d<int> v1(p1), v2(p2);
		v2 -= v1;
		size_t lng = std::max(std::max(::abs(v2.x), ::abs(v2.y)), ::abs(v2.z));
		d = Vector3d<double>(v2.x / (double) lng, v2.y / (double) lng, v2.z / (double) lng);
		return lng;
	} // PrepareLine

	//--------------------------------------------------------------------------
	// ShiftPenAlongLine
	//--------------------------------------------------------------------------
	inline void ShiftPenAlongLine(Vector3d<double> &D, int &dx, int &dy, int &dz)
	{
		if (D.x < 0 && (D.x < (dx - 1)))
			--dx;
		if (D.x > 0 && (D.x > (dx + 1)))
			++dx;
		if (D.y < 0 && (D.y < (dy - 1)))
			--dy;
		if (D.y > 0 && (D.y > (dy + 1)))
			++dy;
		if (D.z < 0 && (D.z < (dz - 1)))
			--dz;
		if (D.z > 0 && (D.z > (dz + 1)))
			++dz;
	} // ShiftPenAlongLine

	//--------------------------------------------------------------------------
	// RasterizeLineBres
	//--------------------------------------------------------------------------
	/** Drawing line using Bressenham algorithm. `Dest` is an appropriate functor 
	storing generated points to required place in memory/display. */
	template <class T> void RasterizeLineBres(const T &Dest, Vector3d<size_t> p1, Vector3d<size_t> p2)
	{
		int i, l, m, n, x_inc, y_inc, z_inc, err_1, err_2, dx2, dy2, dz2;
		Vector3d<size_t> point = p1;

		if (p2.x < p1.x)
		{ 
			x_inc = -1;
			l = static_cast<int>(p1.x - p2.x);
		}
		else
		{
			x_inc = 1;
			l = static_cast<int>(p2.x - p1.x);
		}

		if (p2.y < p1.y)
		{ 
			y_inc = -1;
			m = static_cast<int>(p1.y - p2.y);
		}
		else
		{
			y_inc = 1;
			m = static_cast<int>(p2.y - p1.y);
		}

		if (p2.z < p1.z)
		{ 
			z_inc = -1;
			n = static_cast<int>(p1.z - p2.z);
		}
		else
		{
			z_inc = 1;
			n = static_cast<int>(p2.z - p1.z);
		}

		dx2 = l << 1;
		dy2 = m << 1;
		dz2 = n << 1;

		if ((l >= m) && (l >= n))
		{
			err_1 = dy2 - l;
			err_2 = dz2 - l;
			for (i = 0; i < l; i++)
			{
				Dest.StorePoint (point);
				if (err_1 > 0)
				{
					point.y += y_inc;
					err_1 -= dx2;
				}
				if (err_2 > 0)
				{
					point.z += z_inc;
					err_2 -= dx2;
				}
				err_1 += dy2;
				err_2 += dz2;
				point.x += x_inc;
			}
		}
		else if ((m >= l) && (m >= n))
		{
			err_1 = dx2 - m;
			err_2 = dz2 - m;

			for (i = 0; i < m; i++)
			{
				Dest.StorePoint (point);
				if (err_1 > 0)
				{
					point.x += x_inc;                
					err_1 -= dy2;             
				}
				if (err_2 > 0)
				{
					point.z += z_inc;
					err_2 -= dy2;
				}
				err_1 += dx2;
				err_2 += dz2;
				point.y += y_inc;
			}
		}
		else
		{
			err_1 = dy2 - n;
			err_2 = dx2 - n;
			for (i = 0; i < n; i++)
			{
				Dest.StorePoint (point);
				if (err_1 > 0)
				{
					point.y += y_inc;
					err_1 -= dz2;
				}
				if (err_2 > 0)
				{
					point.x += x_inc;
					err_2 -= dz2;
				}
				err_1 += dy2;
				err_2 += dx2;
				point.z += z_inc;
			}
		}
		Dest.StorePoint (point);
	} // RasterizeLineBres

	//--------------------------------------------------------------------------
	// FillRow
	//--------------------------------------------------------------------------
	template <class VOXEL1, class VOXEL2> size_t FillRow(
		std::stack<size_t> &buf,
		const Image3d<VOXEL1> &input,
		Image3d<VOXEL2> &output, 
		VOXEL2 NewColor,
		const Neighbourhood &neib)
	{
		std::vector<const VOXEL1 *> adjacent;
		size_t max_x = input.GetSizeX() - 1;
		size_t Current, Addr, RowSeed;
		size_t i, j, Overlap, ScanLineLength = 0;
		RowSeed = buf.top();
		buf.pop();
		Addr = RowSeed;
		// Test whether the voxel has been already filled with the NewColor
		if (output.GetVoxel(Addr) == NewColor)
			return 0;

		// In case the flood has not already reached the current voxel we pick up
		// its color = color of FloodFilled region
		VOXEL1 OldColor = input.GetVoxel(Addr);
		// Mark the scanline voxels (bounded by object voxels) with NewColor
		do
		{
			output.SetVoxel(Addr, NewColor);
			ScanLineLength++;
			if ((input.GetX (Addr) == 0) || (input.GetVoxel (Addr - 1) != OldColor))
				break;
			Addr--;
		}
		while (true);
		// Find the apropriate neighbouring voxels
		GetWindow(input, input.GetX(RowSeed), input.GetY(RowSeed), input.GetZ(RowSeed), neib, adjacent);
		for (i = 0; i < adjacent.size(); i++)
		{
			Current = adjacent[i] - input.GetFirstVoxelAddr();

			// Eliminate all the voxels in the same line as the seed is.
			if ((input.GetY(Current) == input.GetY(RowSeed)) && (input.GetZ(Current) == input.GetZ(RowSeed)))
				continue;

			// Before implating new seed we have to find an apropriate place
			// for new seed. It should be the rightmost voxel in the currently
			// passed scanline. The variable "Overlap" measures the distance
			// from voxel "Current" to the rightmost voxel.
			Overlap = 0;
			// The first voxel in the scanline may be added to the
			// queue as a new seed if its color is not equal to new color .
			if ((input.GetVoxel(Current) == OldColor) && (output.GetVoxel(Current) != NewColor))
			{
				// Go to the rightmost component voxel in the line
				// parallel to the currently filled scanline.
				while ((input.GetX(Current) < max_x) && (input.GetVoxel(Current + 1) == OldColor))
				{
					Current++;
					Overlap++;
				}

				buf.push(Current);
			}

			// Verify the scanline - the same length as currently filled
			// scanline.
			for (j = 0; j < (ScanLineLength + Overlap - 1); j++)
			{
				if (input.GetX(Current) == 0)
					break;

				if ((input.GetVoxel (Current) != OldColor) && 
					(input.GetVoxel (Current - 1) == OldColor) && 
					(output.GetVoxel (Current - 1) != NewColor))
					buf.push (Current - 1);

				Current--;
			}
		}

		return ScanLineLength;
	} // FillRow

	/**************************************************************************\

	*

	*					Drawing and rasterization algorithms

	*

	\**************************************************************************/
	//--------------------------------------------------------------------------
	// DrawLine
	//--------------------------------------------------------------------------
	template <class T> void DrawLine(Image3d<T> &img, const Offset &p1, const Offset &p2, T color)
	{
		//cout << "DrawLine microns" << endl; 
		//cout << "p1 = " << p1 << endl; 
		//cout << "p2 = " << p2 << endl; 
		Vector3d<float> pp1 = MicronsToPixels(p1 - img.GetOffset(), img.GetResolution());
		Vector3d<float> pp2 = MicronsToPixels(p2 - img.GetOffset(), img.GetResolution());
		//cout << "pp1 = " << pp1 << endl; 
		//cout << "pp2 = " << pp2 << endl; 
		//FIXME: clipping!!!

		Vector3d<size_t> pix1 = Vector3d<size_t>(
			(size_t) rint (pp1.x),
			(size_t) rint (pp1.y),
			(size_t) rint (pp1.z));

		Vector3d<size_t> pix2 = Vector3d<size_t>(
			(size_t) rint (pp2.x),
			(size_t) rint (pp2.y),
			(size_t) rint (pp2.z));

		DrawLine(img, pix1, pix2, color);
		//cout << "DrawLine microns OK" << endl; 
	} // DrawLine

	template <class T> void DrawLine(Image3d<T> &img, Vector3d<size_t> p1, Vector3d<size_t> p2, T color)
	{
		int dx = 0, dy = 0, dz = 0;
		Vector3d<double> D(0), d;
		size_t lng = PrepareLine(p1, p2, d);

		for (size_t i = 0; i <= lng; ++i)
		{
			img.SetVoxel(p1.x + dx, p1.y + dy, p1.z + dz, color);
			D += d;
			ShiftPenAlongLine(D, dx, dy, dz);
		}
		img.SetVoxel(p2.x, p2.y, p2.z, color);
	} // DrawLine

	//--------------------------------------------------------------------------
	// Explicit instantiations
	//--------------------------------------------------------------------------
	template I3D_ALGO_EXPORT void DrawLine(
		Image3d<bool> &img,
		Vector3d<size_t> p1,
		Vector3d<size_t> p2, 
		bool color);

	template I3D_ALGO_EXPORT void DrawLine(
		Image3d<GRAY8> &img,
		Vector3d<size_t> p1,
		Vector3d<size_t> p2, 
		GRAY8 color);

	template I3D_ALGO_EXPORT void DrawLine(
		Image3d<GRAY16> &img,
		Vector3d<size_t> p1,
		Vector3d<size_t> p2, 
		GRAY16 color);

	template I3D_ALGO_EXPORT void DrawLine(
		Image3d<RGB> &img,
		Vector3d<size_t> p1,
		Vector3d<size_t> p2, 
		RGB color);

	template I3D_ALGO_EXPORT void DrawLine(
		Image3d<RGB16> &img,
		Vector3d<size_t> p1,
		Vector3d<size_t> p2, 
		RGB16 color);

	//--------------------------------------------------------------------------
	// DrawLines
	//--------------------------------------------------------------------------
	template <class T> void DrawLines(Image3d<T> &img, const std::vector<Offset> &pnts, T color)
	{
		//cout << "DrawLines" << endl; 
		for (size_t i = 0; i < pnts.size() / 2; ++i)
		{
			DrawLine (img, pnts[2 * i], pnts[2 * i + 1], color);
		}
		//cout << "DrawLines OK" << endl; 
	} // DrawLines

	//--------------------------------------------------------------------------
	// Explicit instantiations
	//--------------------------------------------------------------------------
	template I3D_ALGO_EXPORT void DrawLines(
		Image3d<GRAY8> &img, 
		const std::vector<Offset> &pnts, 
		GRAY8 color);

	template I3D_ALGO_EXPORT void DrawLines(
		Image3d<GRAY16> &img,
		const std::vector<Offset> &pnts,
		GRAY16 color);

	template I3D_ALGO_EXPORT void DrawLines(
		Image3d<RGB> &img,
		const std::vector<Offset> &pnts, 
		RGB color);

	template I3D_ALGO_EXPORT void DrawLines(
		Image3d<RGB16> &img,
		const std::vector<Offset> &pnts,
		RGB16 color);

	//--------------------------------------------------------------------------
	// DrawLineBres
	//--------------------------------------------------------------------------
	template <class T> void DrawLineBres(Image3d <T> &img, Vector3d <size_t> p1, Vector3d <size_t> p2, T color)
	{
		RasterizeLineBres(StoreToImage<T>(&img, color), p1, p2);
	} // DrawLineBres

	void DrawLineBres(std::list< Vector3d<int> > &ll, Vector3d<size_t> p1, Vector3d<size_t> p2)
	{
		RasterizeLineBres(StoreToMem(&ll), p1, p2);
	}  // DrawLineBres

	//--------------------------------------------------------------------------
	// Explicit instantiations
	//--------------------------------------------------------------------------
	template I3D_ALGO_EXPORT void DrawLineBres(
		Image3d<bool> &img,
		Vector3d<size_t> p1,
		Vector3d<size_t> p2,
		bool color);

	template I3D_ALGO_EXPORT void DrawLineBres(
		Image3d<GRAY8> &img,
		Vector3d<size_t> p1,
		Vector3d<size_t> p2, 
		GRAY8 color);

	template I3D_ALGO_EXPORT void DrawLineBres(
		Image3d<GRAY16> &img,
		Vector3d<size_t> p1,
		Vector3d<size_t> p2, 
		GRAY16 color);

	//--------------------------------------------------------------------------
	// DrawCube
	//--------------------------------------------------------------------------
	template <class T> void DrawCube(Image3d<T> &img, const VOI<PIXELS> &voi, T color)
	{
		// make sure that voi do not address elements out of image
		VOI<PIXELS> img_voi(0,img.GetSize());
		VOI<PIXELS> v = voi * img_voi; // make VOI intersection

		// addr of first pixel to draw
		size_t i = 
				  v.offset.x +   
				  v.offset.y * img.GetSizeX() + 
				  v.offset.z * img.GetSliceSize();

		size_t dy = img.GetSizeX() - v.size.x;
		size_t dz = img.GetSliceSize() - img.GetSizeX() * v.size.y;

		T* data = img.GetVoxelAddr(i);

		for (size_t z = 0; z < v.size.z; z++)
		{
			for (size_t y = 0; y < v.size.y; y++)
			{
				for (size_t x = 0; x < v.size.x; x++)
				{
					*data = color;
					++data;
				}
				data += dy;
			}
			data += dz;
		}
	} // DrawCube

	//--------------------------------------------------------------------------
	// Explicit instantiations
	//--------------------------------------------------------------------------
	template I3D_ALGO_EXPORT void DrawCube (Image3d<bool> &img, const VOI<PIXELS> &voi, bool color);
	template I3D_ALGO_EXPORT void DrawCube (Image3d<GRAY8> &img, const VOI<PIXELS> &voi, GRAY8 color);

	/**************************************************************************\

	*

	*						Filling algorithms

	*

	\**************************************************************************/
	//--------------------------------------------------------------------------
	// FloodFill
	//--------------------------------------------------------------------------
	template <class VOXEL1, class VOXEL2> ComponentInfo<VOXEL2, VOXEL1> FloodFill(
		const Image3d<VOXEL1> &input,
		Image3d<VOXEL2> &output, 
		VOXEL2 color,
		size_t x,
		size_t y, 
		size_t z, 
		const Neighbourhood &neib)
	{
		return FloodFill(input, output, color, input.GetIndex(x, y, z), neib);
	} // FloodFill

	template <class VOXEL1, class VOXEL2> ComponentInfo<VOXEL2, VOXEL1> FloodFill(
		const Image3d<VOXEL1> &input,
		Image3d<VOXEL2> &output, 
		VOXEL2 color,
		size_t seed_index,
		const Neighbourhood &neib)
	{
		size_t LengthOfRow, RightEnd, LeftEnd;
		std::stack<size_t> buf;
		ComponentInfo<VOXEL2, VOXEL1> component;
		const VOXEL1 *ptr = input.GetVoxelAddr(seed_index);
		size_t max_x = input.GetSizeX() - 1;
		size_t seed_x = input.GetX(seed_index);
		
		while ((seed_x < max_x) && (*ptr == *(ptr + 1)))
		{
			++ptr;
			++seed_index;
			++seed_x;
		}

		buf.push(seed_index);
		while (!buf.empty())
		{
			RightEnd = buf.top();
			LengthOfRow = FillRow(buf, input, output, color, neib);
			if (LengthOfRow > 0)
			{
				LeftEnd = RightEnd - LengthOfRow + 1;

				// Enlarge the VOI of component
				if (component.volume == 0)
				{
					component.voi = VOI<UPIXELS>(input.GetX(LeftEnd), input.GetY(LeftEnd), input.GetZ(LeftEnd), LengthOfRow, 1, 1);
				}
				else
				{
					component.voi += VOI<UPIXELS>(input.GetX(LeftEnd), input.GetY(LeftEnd), input.GetZ(LeftEnd), LengthOfRow, 1, 1);
				}

				// Enlarge the volume of component
				component.volume += LengthOfRow;
			}
		}

		component.border = input.OnBorder(component.voi.offset) || input.OnBorder(component.voi.LastPos());

		component.voxel_value = input.GetVoxel(seed_index);

		return component;
	} // FloodFill

	template <class VOXEL> void FloodFill(Image3d<VOXEL> &img, VOXEL color, size_t x, size_t y, size_t z, const Neighbourhood &neib)
	{
		FloodFill(img, color, img.GetIndex(x, y, z), neib);
	} // FloodFill

	template <class VOXEL> void FloodFill(Image3d<VOXEL> &img, VOXEL color, size_t seed_index, const Neighbourhood &neib)
	{
		std::stack<size_t> buf;
		VOXEL *ptr = img.GetVoxelAddr(seed_index);
		size_t max_x = img.GetSizeX() - 1;
		size_t seed_x = img.GetX(seed_index);

		// move the first seed voxel to the rightmost voxel
		while ((seed_x < max_x) && (*ptr == *(ptr + 1)))
		{
			++ptr;
			++seed_index;
			++seed_x;
		}

		buf.push(seed_index);
		while (!buf.empty())
		{
			FillRow(buf, img, img, color, neib);
		}
	} // FloodFill

	//--------------------------------------------------------------------------
	// Explicit instantiations
	//--------------------------------------------------------------------------
	template I3D_ALGO_EXPORT ComponentInfo<bool, bool> FloodFill(const Image3d<bool> &in, Image3d<bool> &out, bool color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<int, bool> FloodFill(const Image3d<bool> &in, Image3d<int> &out, int color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<GRAY8, bool> FloodFill(const Image3d<bool> &in, Image3d<GRAY8> &out, GRAY8 color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<GRAY8, GRAY16> FloodFill(const Image3d<GRAY16> &in, Image3d<byte> &out, GRAY8 color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<GRAY8, GRAY8> FloodFill(const Image3d<GRAY8> &in, Image3d<GRAY8> &out, GRAY8 color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<GRAY16, bool> FloodFill(const Image3d<bool> &in, Image3d<GRAY16> &out, GRAY16 color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<GRAY16, GRAY16> FloodFill(const Image3d<GRAY16> &in, Image3d<GRAY16> &out, GRAY16 color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<size_t, bool> FloodFill(const Image3d<bool> &in, Image3d<size_t> &out, size_t color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<size_t, GRAY8> FloodFill(const Image3d<GRAY8> &in, Image3d<size_t> &out, size_t color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<size_t, GRAY16> FloodFill(const Image3d<GRAY16> &in, Image3d<size_t> &out, size_t color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<size_t, float> FloodFill(const Image3d<float> &in, Image3d<size_t> &out, size_t color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
		
	template I3D_ALGO_EXPORT ComponentInfo<bool, bool> FloodFill(const Image3d<bool> &in, Image3d<bool> &out, bool color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<int, bool> FloodFill(const Image3d<bool> &in, Image3d<int> &out, int color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<GRAY8, bool> FloodFill(const Image3d<bool> &in, Image3d<GRAY8> &out, GRAY8 color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<GRAY8, GRAY8> FloodFill(const Image3d<GRAY8> &in, Image3d<GRAY8> &out, GRAY8 color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<GRAY8, GRAY16> FloodFill(const Image3d<GRAY16> &in, Image3d<GRAY8> &out, GRAY8 color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<GRAY16, bool> FloodFill(const Image3d<bool> &in, Image3d<GRAY16> &out, GRAY16 color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<GRAY16, GRAY16> FloodFill(const Image3d<GRAY16> &in, Image3d<GRAY16> &out, GRAY16 color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<GRAY16, GRAY8> FloodFill(const Image3d<GRAY8> &in, Image3d<GRAY16> &out, GRAY16 color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<size_t, bool> FloodFill(const Image3d<bool> &in, Image3d<size_t> &out, size_t color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<size_t, GRAY8> FloodFill(const Image3d<GRAY8> &in, Image3d<size_t> &out, size_t color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<size_t, GRAY16> FloodFill(const Image3d<GRAY16> &in, Image3d<size_t> &out, size_t color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT ComponentInfo<size_t, float> FloodFill(const Image3d<float> &in, Image3d<size_t> &out, size_t color, size_t seed_index, const Neighbourhood &neib);

	template I3D_ALGO_EXPORT void FloodFill(Image3d<bool> &img, bool color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT void FloodFill(Image3d<GRAY8> &img, GRAY8 color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT void FloodFill(Image3d<GRAY16> &img, GRAY16 color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT void FloodFill(Image3d<RGB> &img, RGB color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT void FloodFill(Image3d<RGB16> &img, RGB16 color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT void FloodFill(Image3d<size_t> &img, size_t color, size_t x, size_t y, size_t z, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT void FloodFill(Image3d<float> &img, float color, size_t x, size_t y, size_t z, const Neighbourhood &neib);

	template I3D_ALGO_EXPORT void FloodFill(Image3d<bool> &img, bool color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT void FloodFill(Image3d<GRAY8> &img, GRAY8 color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT void FloodFill(Image3d<GRAY16> &img, GRAY16 color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT void FloodFill(Image3d<RGB> &img, RGB color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT void FloodFill(Image3d<RGB16> &img, RGB16 color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT void FloodFill(Image3d<size_t> &img, size_t color, size_t seed_index, const Neighbourhood &neib);
	template I3D_ALGO_EXPORT void FloodFill(Image3d<float> &img, float color, size_t seed_index, const Neighbourhood &neib);

	/**************************************************************************\

	*

	*						Other functions

	*

	\**************************************************************************/
	//--------------------------------------------------------------------------
	// Negative
	//--------------------------------------------------------------------------
	template <class VOXEL> void Negative(Image3d<VOXEL> &img)
	{
		size_t i = 0;
		while (i < img.GetImageSize())
		{
			img.SetVoxel(i, ColorInversion(img.GetVoxel(i)));
			i++;
		}
	} // Negative

	//--------------------------------------------------------------------------
	// Explicit instantiations
	//--------------------------------------------------------------------------
	template I3D_ALGO_EXPORT void Negative(Image3d<bool> &img);
	template I3D_ALGO_EXPORT void Negative(Image3d<GRAY8> &img);
	template I3D_ALGO_EXPORT void Negative(Image3d<GRAY16> &img);

} // i3d namespace
