/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * lab_reg_nha_domain_initializer.cc
 *
 * Initializer of the labelled region-based version of the Nilsson-Heyden algorithm
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006 - 2011
 */

#include "lab_reg_nha_domain_initializer.h"

namespace i3d
{
	namespace fls
	{
		/////////////////////////////////////////////////////////////////
		//                                                             //
		//            Class Lab_Reg_NHA_DomainInitializer              //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		void Lab_Reg_NHA_DomainInitializer::Initialize(std::vector<Type::Size> &initial_interface)
		{
			// index of the starting grid point 
			Type::Size index = m_grid.m_sx + 1;

			// implicit function at the starting grid point
			int *ptr = m_grid.m_implicit_function.ptr(index);

			// label at the starting grid point
			Type::LabVoxel *label = m_labels.ptr(index);

			// the first grid point in given labeled image
			const Type::LabVoxel *mask = m_mask;

			// the first grid point in given domain
			const Type::BinVoxel *domain = m_domain;

			// mask as well as domain size in the x axis
			Type::Size sx = m_grid.m_sx - 2;

			// mask as well as domain size in the y axis
			Type::Size sy = m_grid.m_sy - 2;

			// go through the grid and look for border grid points in given mask with respect to given domain

			// check the first row
			for (Type::Size x = 0; x < sx; ++x)
			{
				// check whether a grid point is in given domain
				if (*domain)
				{
					// copy the label
					*label = *mask;

					// check whether a grid point has a foreground label
					if (*mask != 0)
					{
						*ptr = 0;
						initial_interface.push_back(index);

						// add a grid point to the foreground
						m_obj_manager -> AddForeground(index);
					}
					else
					{
						*ptr = 1;

						// add a grid point to the background
						m_obj_manager -> AddBackground(index);
					}
				}

				// move to the next grid point
				++index; ++ptr; ++label; ++mask; ++domain;
			}

			// move to the next row
			index += 2; ptr += 2;

			// check the remaining rows except the last one
			for (Type::Size y = 2; y < sy; ++y)
			{
				// check the first grid point in the row
				if (*domain)
				{
					// copy the label
					*label = *mask;

					if (*mask != 0)
					{
						*ptr = 0;
						initial_interface.push_back(index);

						// add a grid point to the foreground
						m_obj_manager -> AddForeground(index);
					}
					else
					{
						*ptr = 1;

						// add a grid point to the background
						m_obj_manager -> AddBackground(index);
					}
				}

				// move to the next grid point
				++index; ++ptr; ++label; ++mask; ++domain;

				// chech the remaining grid points in the row except the last one
				for (Type::Size x = 2; x < sx; ++x)
				{
					if (*domain)
					{
						// copy the label
						*label = *mask;

						if (*mask != 0)
						{
							// check whether all four neighbours are in given domain as well as in given mask
							if (*(domain - sx) && (*(mask - sx) == *mask) && *(domain - 1) && (*(mask - 1) == *mask) && 
								*(domain + 1) && (*(mask + 1) == *mask) && *(domain + sx) && (*(mask + sx) == *mask))
							{
								*ptr = -1;
							}
							else
							{
								*ptr = 0;
								initial_interface.push_back(index);
							}

							// add a grid point to the foreground
							m_obj_manager -> AddForeground(index);
						}
						else
						{
							*ptr = 1;

							// add a grid point to the background
							m_obj_manager -> AddBackground(index);
						}
					}

					// move to the next grid point
					++index; ++ptr; ++label; ++mask; ++domain;
				}

				// check the last grid point in the row
				if (*domain)
				{
					// copy the label
					*label = *mask;

					if (*mask != 0)
					{
						*ptr = 0;
						initial_interface.push_back(index);

						// add a grid point to the foreground
						m_obj_manager -> AddForeground(index);
					}
					else
					{
						*ptr = 1;

						// add a grid point to the background
						m_obj_manager -> AddBackground(index);
					}
				}

				// move to the next row
				index += 3; ptr += 3; label += 3; ++mask; ++domain;
			}

			// check the last row
			for (Type::Size x = 0; x < sx; ++x)
			{
				// check whether a grid point is in given domain
				if (*domain)
				{
					// copy the label
					*label = *mask;

					// check whether a grid point has a foreground label
					if (*mask != 0)
					{
						*ptr = 0;
						initial_interface.push_back(index);

						// add a grid point to the foreground
						m_obj_manager -> AddForeground(index);
					}
					else
					{
						*ptr = 1;

						// add a grid point to the background
						m_obj_manager -> AddBackground(index);
					}
				}

				// move to the next grid point
				++index; ++ptr; ++label; ++mask; ++domain;
			}
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//           Class Lab_Reg_NHA_DomainInitializer3D             //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		void Lab_Reg_NHA_DomainInitializer3D::Initialize(std::vector<Type::Size> &initial_interface)
		{
			// index of the starting grid point 
			Type::Size index = m_grid.m_slice + m_grid.m_sx + 1;

			// implicit function at the starting grid point
			int *ptr = m_grid.m_implicit_function.ptr(index);

			// label at the starting grid point
			Type::LabVoxel *label = m_labels.ptr(index);

			// shift to the next slice
			Type::Size shift = 2 * m_grid.m_sx;

			// the first grid point in given mask
			const Type::LabVoxel *mask = m_mask;

			// the first grid point in given domain
			const Type::BinVoxel *domain = m_domain;

			// mask as well as domain size in the x axis
			Type::Size sx = m_grid.m_sx - 2;

			// mask as well as domain size in the y axis
			Type::Size sy = m_grid.m_sy - 2;

			// mask as well as domain size in the z axis
			Type::Size sz = m_grid.m_sz - 2;

			// mask as well as domain slice size in the x and y axes
			Type::Size slice = sx * sy;

			// go through the grid and look for border grid points in given mask

			// check the first slice
			for (Type::Size y = 0; y < sy; ++y)
			{
				for (Type::Size x = 0; x < sx; ++x)
				{
					// check whether a grid point is in given domain
					if (*domain)
					{
						// copy the label
						*label = *mask;

						// check whether a grid point is has a foreground label
						if (*mask != 0)
						{
							*ptr = 0;
							initial_interface.push_back(index);

							// add a grid point to the foreground
							m_obj_manager -> AddForeground(index);
						}
						else
						{
							*ptr = 1;

							// add a grid point to the background
							m_obj_manager -> AddBackground(index);
						}
					}

					// move to the next grid point
					++index; ++ptr; ++label; ++mask; ++domain;
				}

				// move to the next row
				index += 2; ptr += 2; label += 2;
			}

			// move to the next slice
			index += shift; ptr += shift; label += shift;

			// check the remaining slices except the last one
			for (Type::Size z = 2; z < sz; ++z)
			{
				// check the first row
				for (Type::Size x = 0; x < sx; ++x)
				{
					// check whether a grid point is in given domain
					if (*domain)
					{
						// copy the label
						*label = *mask;

						// check whether a grid point has a foreground label
						if (*mask != 0)
						{
							*ptr = 0;
							initial_interface.push_back(index);

							// add a grid point to the foreground
							m_obj_manager -> AddForeground(index);
						}
						else
						{
							*ptr = 1;

							// add a grid point to the background
							m_obj_manager -> AddBackground(index);
						}
					}

					// move to the next grid point
					++index; ++ptr; ++label; ++mask; ++domain;
				}

				// move to the next row
				index += 2; ptr += 2; label += 2;

				// check the remaining rows except the last one
				for (Type::Size y = 2; y < sy; ++y)
				{
					// check the first grid point in the row
					if (*domain)
					{
						// copy the label
						*label = *mask;

						if (*mask != 0)
						{
							*ptr = 0;
							initial_interface.push_back(index);

							// add a grid point to the foreground
							m_obj_manager -> AddForeground(index);
						}
						else
						{
							*ptr = 1;

							// add a grid point to the background
							m_obj_manager -> AddBackground(index);
						}
					}

					// move to the next grid point
					++index; ++ptr; ++label; ++mask; ++domain;

					// chech the remaining grid points in the row except the last one
					for (Type::Size x = 2; x < sx; ++x)
					{
						if (*domain)
						{
							// copy the label
							*label = *mask;

							if (*mask != 0)
							{
								// check whether all six neighbours are in given domain as well as in given mask
								if (*(domain - slice) && (*(mask - slice) == *mask) && *(domain - sx) && (*(mask - sx) == *mask) && 
									*(domain - 1) && (*(mask - 1) == *mask) && *(domain + 1) && (*(mask + 1) == *mask) && 
									*(domain + sx) && (*(mask + sx) == *mask) && *(domain + slice) && (*(mask + slice) == *mask))
								{
									*ptr = -1;
								}
								else
								{
									*ptr = 0;
									initial_interface.push_back(index);
								}

								// add a grid point to the foreground
								m_obj_manager -> AddForeground(index);
							}
							else
							{
								*ptr = 1;

								// add a grid point to the background
								m_obj_manager -> AddBackground(index);
							}
						}

						// move to the next grid point
						++index; ++ptr; ++label; ++mask; ++domain;
					}

					// check the last grid point in the row
					if (*domain)
					{
						// copy the label
						*label = *mask;

						if (*mask != 0)
						{
							*ptr = 0;
							initial_interface.push_back(index);

							// add a grid point to the foreground
							m_obj_manager -> AddForeground(index);
						}
						else
						{
							*ptr = 1;

							// add a grid point to the background
							m_obj_manager -> AddBackground(index);
						}
					}

					// move to the next row
					index += 3; ptr += 3; label += 3; ++mask; ++domain;
				}

				// check the last row
				for (Type::Size x = 0; x < sx; ++x)
				{
					// check whether a grid point is in given domain
					if (*domain)
					{
						// copy the label
						*label = *mask;				

						// check whether a grid point has a foreground label
						if (*mask != 0)
						{
							*ptr = 0;
							initial_interface.push_back(index);

							// add a grid point to the foreground
							m_obj_manager -> AddForeground(index);
						}
						else
						{
							*ptr = 1;

							// add a grid point to the background
							m_obj_manager -> AddBackground(index);
						}
					}

					// move to the next grid point
					++index; ++ptr; ++label; ++mask; ++domain;
				}

				// move to the next slice
				index += shift + 2; ptr += shift + 2; label += shift + 2;
			}

			// check the last slice
			for (Type::Size y = 0; y < sy; ++y)
			{
				for (Type::Size x = 0; x < sx; ++x)
				{
					// check whether a grid point is in given domain
					if (*domain)
					{
						// copy the label
						*label = *mask;				

						// check whether a grid point is has a foreground label
						if (*mask != 0)
						{
							*ptr = 0;
							initial_interface.push_back(index);

							// add a grid point to the foreground
							m_obj_manager -> AddForeground(index);
						}
						else
						{
							*ptr = 1;

							// add a grid point to the background
							m_obj_manager -> AddBackground(index);
						}
					}

					// move to the next grid point
					++index; ++ptr; ++label; ++mask; ++domain;
				}

				// move to the next row
				index += 2; ptr += 2; label += 2;
			}
		}

		/**********************************************************************/

	} // namespace fls 

} // namespace i3d

